/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#ifndef __architecture_CountedPool_h__
#define __architecture_CountedPool_h__

#define FE_CP_TSS	(FE_TSS && FE_OS!=FE_WIN32 && FE_OS!=FE_OSX)

namespace fe
{

/**************************************************************************//**
	@brief Pool of Counted elements

	@ingroup architecture

	The templated type needs to provide a reset(void) method
	to facilitate reuse of objects (when they are only referenced
	by this container).

	The class is designed to be declared mutable and as such still be
	functional as a member of a const object.
	Objects of this class are non-locking thread-safe by the use of
	thread specific pointers.
	Each thread accesses an independant pool.

	If thread specific pointers are not deemed available or reliable,
	this object may devolve to simple new() allocation.
*//***************************************************************************/
template<class T>
class CountedPool
{
	public:
				CountedPool(void)
#if FE_CP_TSS
					:m_tssArray(&cleanup)
#endif
				{
#if !FE_CP_TSS
					feLog("CountedPool TSS not used\n");
#endif
				}

		sp<T>	get(void);

	private:

#if FE_CP_TSS
static	void	cleanup(std::vector< sp<T> >* a_pArray);

		boost::thread_specific_ptr< std::vector< sp<T> > >	m_tssArray;
#endif
};

#if FE_CP_TSS
template<class T>
void CountedPool<T>::cleanup(std::vector< sp<T> >* a_pArray)
{
	delete a_pArray;
}
#endif

template<class T>
sp<T> CountedPool<T>::get(void)
{
#if FE_CP_TSS
	std::vector< sp<T> >* pArray;
	if(!(pArray=m_tssArray.get()))
	{
		pArray=new std::vector< sp<T> >();
		m_tssArray.reset(pArray);
		feLog("CountedPool<T>::get new vector\n");
	}
	std::vector< sp<T> >& array= *pArray;

	//* find an unused slot
	const U32 size=array.size();
	for(U32 m=0;m<size;m++)
	{
		FEASSERT(array[m].isValid());
		FEASSERT(array[m]->count()>0);
//		feLog("CountedPool::get %p %d/%d %d\n",
//				this,m,size,array[m]->count());
		if(array[m]->count()==1)
		{
			array[m]->reset();
			return array[m];
		}
	}

	//* else make a new slot
	array.resize(size+1);
	array[size]=new T();
//	feLog("CountedPool::get %p NEW %d %d\n",
//			this,size,array[size]->count());
	return array[size];
#else
	return sp<T>(new T);
#endif
}

} /* namespace */

#endif /* __architecture_CountedPool_h__ */


