/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#ifndef __draw_DrawCommon_h__
#define __draw_DrawCommon_h__

namespace fe
{

#define FE_DC_CIRCLE_SIDES			64	//* tweak
#define FE_DC_CYLINDER_MAXSLICES	16	//* tweak

#define FE_DC_BUFFER_VERTICES		100	//* reusable space vs new()
#define FE_DC_ICOSAHEDRON_VERTICES	26	//* don't tweak, includes repeats

// TODO SAFEGUARD all drawing operations

/**************************************************************************//**
	@brief Drawing functions not specific to the graphics language

	@ingroup draw

	Using setDrawMode(NULL) activates the default mode,
	which you can get with drawMode() and change.
	The default mode is also the initial mode.
*//***************************************************************************/
class FE_DL_EXPORT DrawCommon: public Initialize<DrawCommon>,
		public ObjectSafe<DrawCommon>, virtual public DrawI
	{
	public:
						DrawCommon(void);
		void			initialize(void);
virtual					~DrawCommon(void);

virtual void			setDrawMode(sp<DrawMode> spMode);
virtual sp<DrawMode>	drawMode(void) const
						{	SAFEGUARD;
							return m_spDrawMode; }
virtual void			pushDrawMode(sp<DrawMode> mode);
virtual sp<DrawMode>	popDrawMode(void);

virtual void			setView(sp<ViewI> spViewI)
						{	SAFEGUARD;
							m_spViewI=spViewI.isValid()?
									spViewI: m_spDefaultView; }
virtual sp<ViewI>		view(void) const
						{	SAFEGUARD;
							return m_spViewI; }

virtual void			draw(cp<DrawableI> cpDrawableI,const Color* color);

virtual void			drawBox(const Box3& box,const Color& color);

virtual void			drawCircle(const SpatialTransform &transform,
								const SpatialVector *scale,const Color &color);

virtual void			drawSphere(const SpatialTransform &transform,
								const SpatialVector *scale,const Color &color);

virtual void			drawCylinder(const SpatialTransform& transform,
								const SpatialVector *scale,Real baseScale,
								const Color& color,U32 slices);

virtual void			drawCylinder(const SpatialVector& location1,
								const SpatialVector& location2,
								Real radius1,Real radius2,
								const Color& color,U32 slices);

virtual void			drawCylinders(const SpatialTransform* transform,
								const SpatialVector* scale,
								const Real *baseScale,
								const U32* slices,U32 cylinders,
								BWORD multicolor,const Color* color);

virtual	void			drawArc(const SpatialTransform &transform,
								const SpatialVector *scale,
								Real startangle,Real endangle,
								const Color &color);

virtual void			drawTransformedPoints(
								const SpatialTransform &transform,
								const SpatialVector *scale,
								const SpatialVector *vertex,
								const SpatialVector *normal,U32 vertices,
								BWORD multicolor,const Color *color);

virtual void			drawTransformedTriangles(
								const SpatialTransform &transform,
								const SpatialVector *scale,
								const SpatialVector *vertex,
								const SpatialVector *normal,
								const Vector2f *texture,U32 vertices,
								StripMode strip,BWORD multicolor,
								const Color *color);

virtual void			drawTransformedLines(
								const SpatialTransform &transform,
								const SpatialVector *scale,
								const SpatialVector *vertex,U32 vertices,
								StripMode strip,BWORD multicolor,
								const Color *color);

virtual	void			drawAxes(Real scale);
virtual	void			drawTransformedAxes(const SpatialTransform &transform,
								Real scale);

virtual	void			drawMarker(Real radius,const Color& color);
virtual	void			drawTransformedMarker(
								const SpatialTransform &transform,
								Real radius,const Color& color);

	protected:

		DrawMode::DrawStyle drawStyle(void) const
					{	DrawMode::DrawStyle style=m_spDrawMode->drawStyle();
						return style==DrawMode::e_defaultStyle?
								(isDirect()? DrawMode::e_edgedSolid:
								DrawMode::e_outline): style; }

		sp<ImageI>	textureImage(void) const
					{	return m_spDrawMode->textureImage(); }
		Real		pointSize(void) const
					{	return m_spDrawMode->pointSize(); }
		Real		lineWidth(void) const
					{	return m_spDrawMode->lineWidth(); }
		BWORD		antialias(void) const
					{	return m_spDrawMode->antialias(); }
		BWORD		backfaceCulling(void) const
					{	return m_spDrawMode->backfaceCulling(); }
		BWORD		twoSidedLighting(void) const
					{	return m_spDrawMode->twoSidedLighting(); }


	private:

		void			drawOrthogonalCircles(
								const SpatialTransform &transform,
								const SpatialVector *scale,const Color &color);
		void			drawIcosahedron(const SpatialTransform &transform,
								const SpatialVector *scale,const Color &color);

		void			createCircle(void);
		void			createIcosahedron(void);
		void			createCylinders(void);
		void			destroyCylinders(void);

		SpatialVector	m_circleVertex[FE_DC_CIRCLE_SIDES+2];
		SpatialVector	m_circleNormal[FE_DC_CIRCLE_SIDES+2];

		SpatialVector	m_transformBuffer[FE_DC_BUFFER_VERTICES];
		SpatialVector	m_normalBuffer[FE_DC_BUFFER_VERTICES];

		SpatialVector	m_icosahedronVertex[FE_DC_ICOSAHEDRON_VERTICES];
		SpatialVector	m_icosahedronNormal[FE_DC_ICOSAHEDRON_VERTICES];
		Vector2f		m_icosahedronTexture[FE_DC_ICOSAHEDRON_VERTICES];

		SpatialVector*	m_pCylinderVertex[FE_DC_CYLINDER_MAXSLICES-2];
		SpatialVector*	m_pCylinderNormal[FE_DC_CYLINDER_MAXSLICES-2];
		Vector2f*		m_pCylinderTexture[FE_DC_CYLINDER_MAXSLICES-2];

		sp<DrawMode>	m_spDefaultMode;
		sp<DrawMode>	m_spDrawMode;
		sp<ViewI>		m_spDefaultView;
		sp<ViewI>		m_spViewI;

		std::vector< sp<DrawMode> >	m_drawStack;

	};

} // namespace

#endif /* __draw_DrawCommon_h__ */
