/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "solve.pmh"

namespace fe
{

BlockPCG::BlockPCG(void)
{
	m_tol = 1.0e-6;
	m_maxIterations = 1000;
	m_n = 0;
}

BlockPCG::~BlockPCG(void)
{
}

Real BlockPCG::inner(const std::vector<SpatialVector> &a_left, const std::vector<SpatialVector> &a_right)
{
	Real s = 0.0;
	unsigned int n = a_left.size();
	for(unsigned int i = 0; i < n; i++)
	{
		s += dot(a_left[i], a_right[i]);
	}
	return s;
}

void BlockPCG::residual(std::vector<SpatialVector> &a_r, sp<SparseMatrix3x3> a_A, const std::vector<SpatialVector> &a_x, const std::vector<SpatialVector> &a_b)
{
	unsigned int n = a_b.size();

	a_A->multiply(a_r, a_x);

	for(unsigned int i = 0; i < n; i++)
	{
		a_r[i] = a_b[i] - a_r[i];
	}
}

void BlockPCG::setSize(unsigned long a_size)
{
	if(a_size != m_n)
	{
		m_n = a_size;
		m_r.resize(m_n);
		m_q.resize(m_n);
		m_d.resize(m_n);
		m_s.resize(m_n);
		m_tmp0.resize(m_n);
		m_tmp1.resize(m_n);
	}
}

void BlockPCG::applyFilters(std::vector<SpatialVector> &a_r, const std::vector<FilterConstraint> &a_filters)
{
	for(unsigned int i = 0; i < a_filters.size(); i++)
	{
		unsigned long index = a_filters[i].index();
		a_r[index] = fe::multiply(a_filters[i].filter(), a_r[index]);
	}
}

const std::vector<SpatialVector> &BlockPCG::residual(void) const
{
	return m_r;
}

bool BlockPCG::solve(std::vector<SpatialVector> &a_x, const sp<SparseMatrix3x3> a_A, const std::vector<SpatialVector> &a_b, const Preconditioner &a_P, const std::vector<FilterConstraint> &a_filters)
{
	bool converged = false;

	setSize(a_b.size());

	Real delta0, deltaNew, deltaOld, deltaStop;

	unsigned int modn = 50;

for(unsigned int i = 0; i < a_b.size(); i++)
{
	fprintf(stderr, "B %f %f %f\n", a_b[i][0], a_b[i][1], a_b[i][2]);
}
	m_tmp0 = a_b;
	applyFilters(m_tmp0, a_filters);
for(unsigned int i = 0; i < m_tmp0.size(); i++)
{
	fprintf(stderr, "m_tmp0 %f %f %f\n", m_tmp0[i][0], m_tmp0[i][1], m_tmp0[i][2]);
}
	a_P.apply(m_tmp1, m_tmp0);
for(unsigned int i = 0; i < m_tmp1.size(); i++)
{
	fprintf(stderr, "m_tmp1 %f %f %f\n", m_tmp1[i][0], m_tmp1[i][1], m_tmp1[i][2]);
}
	delta0 = inner(m_tmp0, m_tmp1);

	residual(m_r, a_A, a_x, a_b);
	applyFilters(m_r, a_filters);

	a_P.apply(m_d, m_r);
	applyFilters(m_d, a_filters);

	deltaNew = inner(m_r,m_d);
	deltaStop = m_tol * m_tol * delta0;

fprintf(stderr, "PCG %d %f %f\n", m_maxIterations, deltaNew, deltaStop);

	unsigned int iter = 0;
	while((iter < m_maxIterations) && (deltaNew > deltaStop))
	{
		a_A->multiply(m_q, m_d);
		applyFilters(m_q, a_filters);

		Real alpha = deltaNew / inner(m_d, m_q);

		for(unsigned int i = 0; i < m_n; i++)
		{
			a_x[i] += alpha * m_d[i];
		}

		if(!(iter % modn))
		{
			residual(m_r, a_A, a_x, a_b);
			applyFilters(m_r, a_filters);
		}
		else
		{
			for(unsigned int i = 0; i < m_n; i++)
			{
				m_r[i] -= alpha * m_q[i];
			}
		}

		a_P.apply(m_s, m_r);

		deltaOld = deltaNew;

		deltaNew = inner(m_r, m_s);

		Real beta = deltaNew/deltaOld;

		for(unsigned int i = 0; i < m_n; i++)
		{
			m_d[i] = m_s[i] + beta*m_d[i];
		}

		applyFilters(m_d, a_filters);

		iter++;
	}

	if(iter < m_maxIterations)
	{
		converged = true;
	}
	residual(m_r, a_A, a_x, a_b);

fprintf(stderr, "CONVERGED IN %d ITERATIONS\n", iter);

	return converged;
}

BlockDiagonalPreconditioner::BlockDiagonalPreconditioner(void)
{
}

BlockDiagonalPreconditioner::~BlockDiagonalPreconditioner(void)
{
}

std::vector<SpatialMatrix> &BlockDiagonalPreconditioner::diagonal(void)
{
	return m_diagonal;
}

void BlockDiagonalPreconditioner::apply(std::vector<SpatialVector> &a_r, const std::vector<SpatialVector> &a_x) const
{
	unsigned int n = a_r.size();
	for(unsigned int i = 0; i < n; i++)
	{
		multiply(a_r[i], m_diagonal[i], a_x[i]);
//fprintf(stderr, "apply P %f %f %f\n", a_x[i][0], a_x[i][1], a_x[i][2]);
//fprintf(stderr, "apply P %f %f %f\n", a_r[i][0], a_r[i][1], a_r[i][2]);
	}
}

FilterConstraint::FilterConstraint(void)
{
}

FilterConstraint::~FilterConstraint(void)
{
}

void FilterConstraint::setIndex(unsigned long a_index)
{
	m_index = a_index;
}

const SpatialMatrix &FilterConstraint::filter(void) const
{
	return m_filter;
}

unsigned long FilterConstraint::index(void) const
{
	return m_index;
}

void FilterConstraint::makePointConstraint(void)
{
	setAll(m_filter, 0.0);
}

void FilterConstraint::makeLineConstraint(const SpatialVector &a_direction)
{
	setAll(m_filter, 0.0);
	outerProduct(m_filter, a_direction, a_direction);
}

void FilterConstraint::makePlaneConstraint(const SpatialVector &a_normal)
{
	setIdentity(m_filter);
	outerProduct(m_filter, a_normal, a_normal);
}


} /* namespace */

