/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#ifndef __solve_SparseMatrix3x3_h__
#define __solve_SparseMatrix3x3_h__

namespace fe
{

class SparseMatrix3x3 : public Component
{
	public:
virtual	~SparseMatrix3x3(void) {}
virtual	void multiply(std::vector<SpatialVector> &a_b,
			const std::vector<SpatialVector> &a_x) const					= 0;
virtual	void scale(Real a_scale)											= 0;
};

/// Vector Map Rows (vector of row maps)
class FullVMR : public SparseMatrix3x3
{
	public:
		FullVMR(void);
virtual	~FullVMR(void);

virtual	void multiply(std::vector<SpatialVector> &a_b,
			const std::vector<SpatialVector> &a_x) const;
virtual	void scale(Real a_scale);

		typedef std::map<unsigned int, SpatialMatrix> t_row;

		void			clear(void);
		void			setRows(unsigned int a_count);
		unsigned int	rows(void);
		t_row			&row(unsigned int a_index);

		SpatialMatrix	&block(unsigned int a_i, unsigned int a_j);

	protected:
		std::vector<t_row>	m_rows;

};

/// Upper Triangular Vector Map Rows (vector of row maps)
class UpperTriangularVMR : public FullVMR
{
	public:
		UpperTriangularVMR(void) {}
virtual	~UpperTriangularVMR(void) {}

virtual	void multiply(std::vector<SpatialVector> &a_b,
			const std::vector<SpatialVector> &a_x) const;
	private:

};

class FullBCRS : public SparseMatrix3x3
{
	public:
		FullBCRS(void);
virtual	~FullBCRS(void);


virtual	void multiply(std::vector<SpatialVector> &a_b,
			const std::vector<SpatialVector> &a_x) const;
virtual	void scale(Real a_scale);

		void				clear(void);
		void				startrow(void);
		void				done(void);
		SpatialMatrix		&insert(unsigned int a_columnIndex,
								const SpatialMatrix &a_block);

	protected:
		std::vector<SpatialMatrix>			m_blocks;
		std::vector<unsigned int>			m_colind;
		std::vector<unsigned int>			m_rowptr;
		unsigned int						m_r;
};

class UpperTriangularBCRS : public FullBCRS
{
	public:
		UpperTriangularBCRS(void){}
virtual	~UpperTriangularBCRS(void){}

virtual	void multiply(std::vector<SpatialVector> &a_b,
			const std::vector<SpatialVector> &a_x) const;

};


inline FullBCRS::FullBCRS(void)
{
	m_r = 0;
}

inline FullBCRS::~FullBCRS(void)
{
}

inline void FullBCRS::clear(void)
{
	m_blocks.clear();
	m_colind.clear();
	m_rowptr.clear();
	m_r = 0;
}

inline void FullBCRS::startrow(void)
{
	m_rowptr.push_back(m_r);
}

inline void FullBCRS::done(void)
{
	m_rowptr.push_back(m_r);
}

inline SpatialMatrix &FullBCRS::insert(unsigned int a_columnIndex, const SpatialMatrix &a_block)
{
fprintf(stdout, "insert %u %u\n", m_r, a_columnIndex);
	m_blocks.push_back(a_block);
	m_colind.push_back(a_columnIndex);
	m_r++;
	return m_blocks.back();
}

inline void FullBCRS::multiply(std::vector<SpatialVector> &a_b, const std::vector<SpatialVector> &a_x) const
{
	unsigned int n = m_rowptr.size() - 1;
	for(unsigned int i = 0; i < n; i++)
	{
		a_b[i] = zeroVector;
		for(unsigned int k = m_rowptr[i]; k < m_rowptr[i+1]; k++)
		{
			a_b[i] += fe::multiply(m_blocks[k], a_x[m_colind[k]]);
		}
	}
}

inline void FullBCRS::scale(Real a_scale)
{
	unsigned int n = m_blocks.size();
	for(unsigned int i = 0; i < n; i++)
	{
		m_blocks[i] *= a_scale;
	}
}

inline void UpperTriangularBCRS::multiply(std::vector<SpatialVector> &a_b, const std::vector<SpatialVector> &a_x) const
{
	unsigned int n = m_rowptr.size() - 1;
	for(unsigned int i = 0; i < n; i++)
	{
		a_b[i] = zeroVector;
	}
	for(unsigned int i = 0; i < n; i++)
	{
		unsigned int k = m_rowptr[i];
		a_b[i] += fe::multiply(m_blocks[k], a_x[i]);
		for(k=k+1; k < m_rowptr[i+1]; k++)
		{
			a_b[i] += fe::multiply(m_blocks[k], a_x[m_colind[k]]);
			a_b[m_colind[k]] += fe::transposeMultiply(m_blocks[k], a_x[i]);
		}
	}
}

inline FullVMR::FullVMR(void)
{
}

inline FullVMR::~FullVMR(void)
{
}

inline void UpperTriangularVMR::multiply(std::vector<SpatialVector> &a_b, const std::vector<SpatialVector> &a_x) const
{
	unsigned int n = a_b.size();
	for(unsigned int i = 0; i < n; i++)
	{
		a_b[i] = zeroVector;
	}

	unsigned int r,c;
	n = m_rows.size();
	for(unsigned int i = 0; i < n; i++)
	{
		t_row::const_iterator i_row = m_rows[i].begin();
		r = i;
		const SpatialMatrix &block = (*i_row).second;
		assert((*i_row).first == r);
		a_b[r] += fe::multiply(block, a_x[r]);
		i_row++;

		for(;i_row != m_rows[i].end(); i_row++)
		{
			const SpatialMatrix &block = (*i_row).second;
			c = (*i_row).first;

			a_b[r] += fe::multiply(block, a_x[c]);
			a_b[c] += fe::transposeMultiply(block, a_x[r]);
		}
	}
}

inline void FullVMR::multiply(std::vector<SpatialVector> &a_b, const std::vector<SpatialVector> &a_x) const
{
	unsigned int r,c;
	unsigned int n = m_rows.size();
	for(unsigned int i = 0; i < n; i++)
	{
		a_b[i] = zeroVector;
		r = i;
		for(t_row::const_iterator i_row = m_rows[i].begin();
			i_row != m_rows[i].end(); i_row++)
		{
			const SpatialMatrix &block = (*i_row).second;
			c = (*i_row).first;
			a_b[r] += fe::multiply(block, a_x[c]);
		}
	}
}

inline void FullVMR::scale(Real a_scale)
{
	unsigned int n = m_rows.size();
	for(unsigned int i = 0; i < n; i++)
	{
		for(t_row::iterator i_row = m_rows[i].begin();
			i_row != m_rows[i].end(); i_row++)
		{
			SpatialMatrix &block = (*i_row).second;
			fe::multiply(block, a_scale);
		}
	}
}

inline void FullVMR::clear(void)
{
	m_rows.clear();
}

inline void FullVMR::setRows(unsigned int a_count)
{
	m_rows.resize(a_count);
}

inline unsigned int FullVMR::rows(void)
{
	return m_rows.size();
}

inline UpperTriangularVMR::t_row &FullVMR::row(unsigned int a_index)
{
	return m_rows[a_index];
}

inline SpatialMatrix &FullVMR::block(unsigned int a_i, unsigned int a_j)
{
	assert(a_i < m_rows.size());
	t_row::iterator i_row = m_rows[a_i].find(a_j);
	if(i_row != m_rows[a_i].end())
	{
		return i_row->second;
	}
	SpatialMatrix &m = m_rows[a_i][a_j];
	setAll(m, 0.0);
	return m;
}


} /* namespace */

#endif /* __solve_SparseMatrix3x3_h__ */

