/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#ifndef __thread_JobQueue_h__
#define __thread_JobQueue_h__

#define FE_JQ_LOCK_DEBUG	FALSE

namespace fe
{

/**************************************************************************//**
	@brief Queue of tasks

	@ingroup thread
*//***************************************************************************/
template<typename T>
class JobQueue: public Handled< JobQueue<T> >
{
	public:
				JobQueue(void):
					m_workers(0),
					m_barrierCount(0),
					m_barrierMet(FALSE)
				{	Counted::setName("JobQueue"); }

		U32		workers(void) const			{ return m_workers; }

		void	post(T job)
				{
#if FE_JQ_LOCK_DEBUG
					feLog("JobQueue::post %d LOCK\n",job);
#endif
					boost::mutex::scoped_lock lock(m_jobMonitor);
//					feLog("JobQueue::post %d\n",job);
					m_posted.push(job);
					m_jobAvailable.notify_one();
				}
		void	post(T first_job,T last_job)
				{
#if FE_JQ_LOCK_DEBUG
					feLog("JobQueue::post %d %d LOCK\n",first_job,last_job);
#endif
					boost::mutex::scoped_lock lock(m_jobMonitor);
//					feLog("JobQueue::post %d\n",job);
					for(T job=first_job;job<=last_job;job++)
					{
						m_posted.push(job);
						m_jobAvailable.notify_one();
					}
				}

		BWORD	take(T& job)
				{
//					feLog("JobQueue::take\n");
#if FE_JQ_LOCK_DEBUG
					feLog("JobQueue::take %d LOCK\n",job);
#endif
					boost::mutex::scoped_lock lock(m_jobMonitor);
//					feLog("JobQueue::take locked\n");
					if(m_posted.empty())
					{
						return FALSE;
					}
					job=m_posted.front();
					m_posted.pop();
					return TRUE;
				}
		void	waitForJob(T& job)
				{
//					feLog("JobQueue::waitForJob\n");
#if FE_JQ_LOCK_DEBUG
					feLog("JobQueue::waitForJob %d LOCK\n",job);
#endif
					boost::mutex::scoped_lock lock(m_jobMonitor);
//					feLog("JobQueue::waitForJob locked\n");
					while(m_posted.empty())
					{
//						feLog("JobQueue::waitForJob empty\n");
						m_jobAvailable.wait(lock);
					}
					FEASSERT(!m_posted.empty());
					job=m_posted.front();
//					feLog("JobQueue::waitForJob job=%d\n",job);
					m_posted.pop();
				}

		void	deliver(T job)
				{
#if FE_JQ_LOCK_DEBUG
					feLog("JobQueue::deliver %d LOCK\n",job);
#endif
					boost::mutex::scoped_lock lock(m_workerMonitor);
//					feLog("JobQueue::deliver %d\n",job);
					m_delivered.push(job);
					m_workerDelivered.notify_one();
				}

		BWORD	acceptDelivery(T& job)
				{
#if FE_JQ_LOCK_DEBUG
					feLog("JobQueue::acceptDelivery %d LOCK\n",job);
#endif
					boost::mutex::scoped_lock lock(m_workerMonitor);
//					feLog("JobQueue::acceptDelivery\n");
					if(m_delivered.empty())
					{
						return FALSE;
					}
					job=m_delivered.front();
					m_delivered.pop();
					return TRUE;
				}
		void	waitForDelivery(T& job)
				{
#if FE_JQ_LOCK_DEBUG
					feLog("JobQueue::waitForDelivery %d LOCK\n",job);
#endif
					boost::mutex::scoped_lock lock(m_workerMonitor);
//					feLog("JobQueue::waitForDelivery\n");
					while(m_delivered.empty())
					{
//						feLog("JobQueue::waitForDelivery empty\n");
						m_workerDelivered.wait(lock);
					}
					FEASSERT(!m_delivered.empty());
					job=m_delivered.front();
					m_delivered.pop();
				}

				/**	@brief waits until all threads call sync

					Returns true to exactly one of the threads. */
		BWORD	synchronize(void)
				{
#if FE_JQ_LOCK_DEBUG
					feLog("JobQueue::synchronize LOCK\n");
#endif
					boost::mutex::scoped_lock lock(m_barrierMonitor);
//					feLog("JobQueue::sync %d\n",m_barrierCount);
					while(m_barrierMet)
					{
//						feLog("JobQueue::sync previous %d/%d\n",
//								m_barrierCount,m_workers);
						m_barrierHit.wait(lock);
					}
//					feLog("JobQueue::sync %d++\n",m_barrierCount);
					m_barrierCount++;
					while(!m_barrierMet && m_barrierCount<m_workers)
					{
//						feLog("JobQueue::sync wait %d/%d\n",
//								m_barrierCount,m_workers);
						m_barrierHit.wait(lock);
					}
					BWORD result=(m_barrierCount==m_workers);
					m_barrierCount--;
					m_barrierMet=(m_barrierCount>0);
					m_barrierHit.notify_all();
//					feLog("JobQueue::sync complete\n");
					return result;
				}

	protected:
		U32						m_workers;

	private:
		boost::mutex			m_jobMonitor;
		boost::mutex			m_workerMonitor;
		boost::condition		m_jobAvailable;
		boost::condition		m_workerDelivered;

		std::queue<T>			m_posted;
		std::queue<T>			m_delivered;

		boost::mutex			m_barrierMonitor;
		boost::condition		m_barrierHit;
		U32						m_barrierCount;
		U32						m_barrierMet;
};

} // namespace

#endif // __thread_JobQueue_h__
