/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "vegetation.pmh"

#define FE_TREE_MT_SOLVE			TRUE
#define FE_TREE_MT_KINE				TRUE
#define FE_TREE_MT_DEBUG			FALSE
#define FE_TREE_TRACE				FALSE
#define FE_TREE_EXPLICIT			FALSE	// zero the Jacobians
#define FE_TREE_PRECONDITION		TRUE	// precondition here vs solver
#define FE_TREE_PREMULTIPLY_MASS	TRUE	// force symmetry
#define FE_TREE_SOLVE_OFF_MAIN		FALSE	// use 2 other threads for solvers
#define FE_TREE_MT_THREADS			3		// FE_TREE_SOLVE_OFF_MAIN? 2: 1;
#define FE_TREE_VISIBLE_MIN_RADIUS	0.001f

#define	FE_TREE_SOLVER_THRESHOLD	1e-4
#define	FE_TREE_DAMP_REGION			1.2f	// vel drain, >=1 up to perhaps 2

#define	FE_TREE_MT_ACTION_QUIT		0x00010000
#define	FE_TREE_MT_ACTION_SOLVE		0x00020000
#define	FE_TREE_MT_ACTION_KINE		0x00040000
#define	FE_TREE_MT_ACTION_MASK		0x7fff0000
#define	FE_TREE_MT_INDEX_MASK		0x0000ffff

namespace fe
{
void Tree::initialize(void)
{
	m_boundLevel= -1;
	m_sticks=1;
	m_lastDeltaT=0.0f;
	m_pLeafPoints=NULL;
	m_pLeafNormals=NULL;

	set(m_location);
	set(m_effectorForce);
	setStick(0,&m_root);
	m_solver[0].setThreshold(FE_TREE_SOLVER_THRESHOLD);
	m_solver[1].setThreshold(FE_TREE_SOLVER_THRESHOLD);
//	m_solver[0].setPreconditioning(
//			ConjugateGradient<MatrixN,VectorN>::e_diagonal_A);
//	m_solver[1].setPreconditioning(
//			ConjugateGradient<MatrixN,VectorN>::e_diagonal_A);

	//* start helper threads
	m_spGang=new Gang<SolverWorker,I32>();
	m_spGang->start(FE_TREE_MT_THREADS,sp<Counted>(this));

	m_effectorForceArray.resize(FE_TREE_MT_THREADS+1,SpatialVector());

	m_bundles=FE_TREE_MT_THREADS+1;
	m_stickTransformBundles.resize(m_bundles);
	m_stickScaleBundles.resize(m_bundles);
	m_stickBaseScaleBundles.resize(m_bundles);
	m_stickSliceBundles.resize(m_bundles);
	m_bundleSize.resize(m_bundles);

	for(U32 m=0;m<m_bundles;m++)
	{
		m_stickTransformBundles[m]=NULL;
		m_stickScaleBundles[m]=NULL;
		m_stickBaseScaleBundles[m]=NULL;
		m_stickSliceBundles[m]=NULL;
		m_bundleSize[m]=0;
	}

	//* initialize profiling
	m_spProfiler=new Profiler("Tree Simulation");
	m_spProfileReset=new Profiler::Profile(m_spProfiler,"Reset");
	m_spProfilePopulate=new Profiler::Profile(m_spProfiler,"Populate");
	m_spProfileSolve=new Profiler::Profile(m_spProfiler,"Solve");
	m_spProfileSolveWait=new Profiler::Profile(m_spProfiler,"Solve Wait");
	for(U32 m=0;m<2;m++)
	{
		String name;
		name.sPrintf("%d Setup",m);
		m_spProfileSetup[m]=new Profiler::Profile(m_spProfiler,name);
		name.sPrintf("%d Solver",m);
		m_spProfileSolver[m]=new Profiler::Profile(m_spProfiler,name);
	}
	m_spProfileThreadWait=new Profiler::Profile(m_spProfiler,"Thread Wait");
	m_spProfileKine=new Profiler::Profile(m_spProfiler,"Kine");
	m_spProfileKineWait=new Profiler::Profile(m_spProfiler,"Kine Wait");
}

Tree::~Tree(void)
{
	feLog("%s:\n%s\n",m_spProfiler->name().c_str(),
			m_spProfiler->report().c_str());

	//* finish both threads
	for(U32 m=0;m<FE_TREE_MT_THREADS;m++)
	{
		m_spGang->post(FE_TREE_MT_ACTION_QUIT);
	}
	U32 count=FE_TREE_MT_THREADS;
	while(count)
	{
		I32 job;
		m_spGang->waitForDelivery(job);
		if(job==FE_TREE_MT_ACTION_QUIT)
		{
			count--;
		}
	}

	for(U32 m=0;m<m_bundles;m++)
	{
		delete[] m_stickSliceBundles[m];
		delete[] m_stickBaseScaleBundles[m];
		delete[] m_stickScaleBundles[m];
		delete[] m_stickTransformBundles[m];
	}

	delete[] m_pLeafNormals;
	delete[] m_pLeafPoints;
}

void Tree::Stick::grow(sp<Tree>& rspTree,U32& stick,
		const U32 segment,const U32 level,
		const F32 a_fullLength,const F32 along,
		const F32 baseRotate,const F32 rotate)
{
//	feLog("a_fullLength %.6G\n",a_fullLength);
	rspTree->bindLevel(level);
	const TreeSeed& seedRV=rspTree->m_seedRV;
	const StickLevel& levelRV=rspTree->m_levelRV;

	const F32 stiffPower=2.5f;

	const F32 stiffness=seedRV.stiffness();
	const F32 stillness=seedRV.stillness();
	const F32 density=seedRV.density();
	const U32 levels=seedRV.levels();

	const U32 segments=levelRV.segments();
	const U32 sideshoots=levelRV.sideshoots();
	const F32 bare=levelRV.bare();
	const F32 valleybase=levelRV.valleybase();
	const F32 valleytip=levelRV.valleytip();
	const F32 peakalong=levelRV.peakalong();
	const F32 powerbase=levelRV.powerbase();
	const F32 powertip=levelRV.powertip();
	const F32 forkrotate=levelRV.forkrotate()*degToRad;
	const F32 forktrench=levelRV.forktrench()*degToRad;
	const I32 forks=levelRV.forks();
	const F32 shootdown=levelRV.shootdown()*degToRad;
	const F32 shootramp=levelRV.shootramp()*degToRad;
	const F32 shootreact=levelRV.shootreact();
	const F32 shootdrain=levelRV.shootdrain();
	const F32 centerdeath=levelRV.centerdeath();
	const F32 sidedeath=levelRV.sidedeath();
	const F32 curvedown=levelRV.curvedown()*degToRad;
	const F32 areaRatio=levelRV.areaRatio();
	const F32 taper=levelRV.taper();
	const F32 fullLength=(a_fullLength>0.0)?
			a_fullLength: (1.0f-along)*levelRV.length();

	if(m_length<=0.0f)
	{
		m_length=fullLength/F32(segments);
		m_radius1=sqrtf(areaRatio/pi*fullLength);
/*
		m_radius1=sqrtf(areaRatio/pi*fullLength)*
				(1.0f-taper*segment/F32(segments));
*/
	}
	m_level=level;
	m_resolution=(level>0)? 3: 6;
	m_radius2=m_radius1*(1.0f-taper*m_length/fullLength);
	m_spring=stiffness*powf(m_radius1,stiffPower);
	m_drag=stillness*powf(m_radius1,stiffPower);
	const F32 volume=m_length*pi*m_radius1*m_radius1;
	m_mass=density*volume;
	BWORD fork=(forks==1 && !segment);
	if(segments<3)
	{
		fork=(I32(segment)<forks);
	}
	else if(forks>1)
	{
		const U32 remainder=segment*(forks-1)%(segments-2);
		fork=!remainder || remainder < (segment-1)*(forks-1)%(segments-2);
	}

#if FALSE
	feLog("%2d lv %d/%d sg %d/%d sp %.3G st %.3G"
			" rest=%.3G,%.3G l=%.3G r=%.3G m=%.3G f=%d rot=%.3G,%.3G\n",
			stick,level,levels,segment,segments,m_spring,m_drag,
			m_rest[0],m_rest[1],m_length,m_radius1,m_mass,fork,
			baseRotate,rotate);
#endif

	if(FE_INVALID_SCALAR(m_radius1))
	{
		feX("Tree::Stick::grow","invalid radius");
	}

	if(segment>=segments || level>=levels)
	{
		return;
	}

	const BWORD no_center=(segment+1==segments) ||
			(randomReal(0.0f,1.0f)<centerdeath);
	I32 sides=0;
	if(level+1<levels && fork && segment/F32(segments)>=bare)
	{
		for(U32 m=0;m<sideshoots;m++)
		{
			if(randomReal(0.0f,1.0f)>=sidedeath)
			{
				sides++;
			}
		}
	}

	const F32 childAlong=segment/F32(segments);
	const F32 shapeScale=shapeFunction(
			(childAlong-bare)/(1.0f-bare),valleybase,
			valleytip,peakalong,powerbase,powertip);
//	feLog("lv %d sh %.6G al %.6G vb %.6G vt %.6G pe %.6G pb %.6G pt %.6G\n",
//			level,shapeScale,childAlong,valleybase,
//			valleytip,peakalong,powerbase,powertip);

	const F32 totalArea=pi*m_radius2*m_radius2;
	F32 remainingArea=totalArea;
	for(I32 m=sides;m>=no_center;m--)
	{
		const U32 nextSegment=m? 0: segment+1;
		const F32 nextAlong=m? childAlong: along;
		const F32 sideRotate=m? pi+2.0f*pi*(m-1)/F32(sides): 0.0f;
		const F32 nextTrench=m? (forktrench*((segment%2)*2-1.0f)): 0.0;
		const F32 nextRotate=rotate+(fork? forkrotate: 0.0f)+
				sideRotate+nextTrench;
		const F32 nextBaseRotate=m? nextRotate: baseRotate;

		const I32 nextLevel=level+(m!=0);
		rspTree->bindLevel(nextLevel);

		const F32 nextMaxLength=fullLength*(m? levelRV.length(): 1.0);

		if(levelRV.form()=="Leaf")
		{
			U32 leafIndex=rspTree->newLeaf();
			attachLeaf(leafIndex);
			Tree::Leaf* pLeaf=
					dynamic_cast<Tree::Leaf*>(rspTree->leaf(leafIndex));

			const F32 alongY=m/F32(sides);
			const F32 radial=m_radius1+alongY*(m_radius2-m_radius1)+
					nextMaxLength*levelRV.bare();

			set(pLeaf->m_normal,sinf(nextRotate),0.5f,cosf(nextRotate));
			//* HACK should we normalize it?

			pLeaf->m_center=radial*pLeaf->m_normal;
			pLeaf->m_center[1]=m_length*alongY;

			continue;
		}

		const F32 nextFullLength=(m)? nextMaxLength*shapeScale: fullLength;

		Stick* pStick=new Stick();
		pStick->m_index=stick++;
		rspTree->setStick(pStick->m_index,pStick);
		attach(pStick);

		if(level==0 && nextLevel==1)
		{
			rspTree->addKine(pStick);
		}

		const U32 nextSegments=levelRV.segments();
		const F32 nextAreaRatio=levelRV.areaRatio();
		F32 nextArea=remainingArea;

		pStick->m_length=nextFullLength/F32(nextSegments);
		if(m>no_center)
		{
			pStick->m_radius1=sqrtf(nextAreaRatio/pi*nextFullLength);
			const F32 maxRadius=m_radius2*0.4f;
			if(pStick->m_radius1>maxRadius)
			{
				pStick->m_radius1=maxRadius;
			}
			nextArea=pi*pStick->m_radius1*pStick->m_radius1;
			remainingArea-=nextArea*(no_center? 1.0f: shootdrain);
		}
		else
		{
			pStick->m_radius1=sqrtf(remainingArea/pi);
			m_radius2=pStick->m_radius1;
		}

		if(fork)
		{
			pStick->m_rest= -(shootdown+nextAlong*shootramp)*
					(m? 1.0: shootreact)*
					(1.0f-nextArea/totalArea)*
					Vector2(sinf(nextRotate),cosf(nextRotate));
		}
		else
		{
			set(pStick->m_rest);
		}
		if(segments>1)
		{
			pStick->m_rest-=curvedown/F32(segments-1)*
					Vector2(sinf(nextBaseRotate),cosf(nextBaseRotate));
		}

		pStick->grow(rspTree,stick,nextSegment,nextLevel,
				nextFullLength,nextAlong,nextBaseRotate,nextRotate);
	}
}

F32 Tree::shapeFunction(const F32 along,const F32 valleybase,
		const F32 valleytip,const F32 peakalong,
		const F32 powerbase,const F32 powertip)
{
	if(along>=peakalong)
	{
		return valleytip+(1.0f-valleytip)*
				powf((1.0-along)/(1.0f-peakalong),powertip);
	}
	return valleybase+(1.0f-valleybase)*
			powf(along/peakalong,powerbase);
}

void Tree::bindLevel(const U32 level)
{
	if(m_boundLevel==I32(level))
	{
		return;
	}
	m_boundLevel=I32(level);

	sp<RecordGroup> spLevelGroup=m_seedRV.levelgroup();
	sp<RecordArray> spRA;

	for(RecordGroup::iterator it=spLevelGroup->begin();
			it!=spLevelGroup->end();it++)
	{
		spRA= *it;
		m_levelRV.bind(spRA);
		while(m_levelRV.step())
		{
			if(m_levelRV.level() == I32(level))
			{
				return;
			}
		}
	}

	// TODO some intentional default vs simply last one read
}

void Tree::generate(const Record seed)
{
	//* first random very near zero
	randomReal(0.0f,1.0f);

	m_seedRV.bind(seed);
	feLog("Tree::generate levels=%d stiffness=%.6G density=%.6G\n",
			m_seedRV.levels(),m_seedRV.stiffness(),m_seedRV.density());

	sp<Tree> spTree(this);
	m_root.grow(spTree,m_sticks,0,0,-1.0f,0.0f,0.0f,0.0f);

	U32 leaves=m_leafArray.size();
	feLog("Tree::generate %d sticks %d leaves\n",m_sticks,leaves);

	delete[] m_pLeafNormals;
	delete[] m_pLeafPoints;
	m_pLeafPoints=new SpatialVector[leaves];
	m_pLeafNormals=new SpatialVector[leaves];

	for(U32 m=0;m<m_bundles;m++)
	{
		delete[] m_stickSliceBundles[m];
		delete[] m_stickBaseScaleBundles[m];
		delete[] m_stickScaleBundles[m];
		delete[] m_stickTransformBundles[m];
		m_stickTransformBundles[m]=new SpatialTransform[m_sticks];
		m_stickScaleBundles[m]=new SpatialVector[m_sticks];
		m_stickBaseScaleBundles[m]=new Real[m_sticks];
		m_stickSliceBundles[m]=new U32[m_sticks];
		m_bundleSize[m]=0;
	}

	m_spProfileKineLimbArray.resize(m_kineArray.size());
	for(U32 m=0;m<m_spProfileKineLimbArray.size();m++)
	{
		String name;
		name.sPrintf("%d Kine Limb",m);
		m_spProfileKineLimbArray[m]=new Profiler::Profile(m_spProfiler,name);
	}

	m_im.reset(m_sticks);
	m_mass.reset(m_sticks);
	m_him.reset(m_sticks);

	for(U32 m=0;m<2;m++)
	{
		m_tempVector1[m].reset(m_sticks);
		m_tempVector2[m].reset(m_sticks);
		m_b[m].reset(m_sticks);
		m_deltaV[m].reset(m_sticks);

//		m_A2[m].reset(m_sticks,m_sticks);
		m_tempMatrix[m].reset(m_sticks,m_sticks);
		m_A[m].reset(m_sticks,m_sticks);
		m_dfdx[m].reset(m_sticks,m_sticks);
//		m_dfdx2[m].reset(m_sticks,m_sticks);
//		m_hdfdx[m].reset(m_sticks,m_sticks);
		m_imdfdx[m].reset(m_sticks,m_sticks);
		m_imdfdv[m].reset(m_sticks,m_sticks);
	}

	VectorN zero[2];
	zero[0].reset(m_sticks);
	zero[1].reset(m_sticks);
	set(zero[0]);
	set(zero[1]);
	m_root.forward_kine_recursive(0,spTree,0.0f,zero,zero,zero,
			m_effectorForce,FALSE);

//	FEASSERT(I32(m_sticks)==(1<<levels)-1);

	set(m_im);

	MatrixN dfdv[2];
	for(U32 m=0;m<2;m++)
	{
		dfdv[m].reset(m_sticks,m_sticks);
		set(dfdv[m]);

		set(m_dfdx[m]);
	}

	m_root.populate_static(spTree,m_dfdx,dfdv,m_im);
#if FE_TREE_EXPLICIT
	m_dfdx[0].clear();
	m_dfdx[1].clear();
	dfdv[0].clear();
	dfdv[1].clear();
#endif

	for(U32 m=0;m<2;m++)
	{
		premultiplyDiagonal(m_imdfdx[m],m_im,m_dfdx[m]);
		premultiplyDiagonal(m_imdfdv[m],m_im,dfdv[m]);
	}
}

void Tree::update(const Real deltaT)
{
	m_spProfiler->begin();

	// HACK
	F32 new_deltaT=deltaT;//0.001f;

	if(new_deltaT<=0.0f)
	{
		return;
	}
	else if(m_lastDeltaT!=new_deltaT)
	{
		feLog("Tree::update dt changed %.6G -> %.6G (%.6G FPS)\n",
				m_lastDeltaT,new_deltaT,1.0f/new_deltaT);

		m_lastDeltaT=new_deltaT;

		MatrixN id(m_sticks,m_sticks);
		setIdentity(id);

		const F32& h=m_lastDeltaT;

		m_him=h*m_im;
		for(U32 m=0;m<m_im.size();m++)
		{
			m_mass[m]=1.0f/m_im[m];
		}

		for(U32 m=0;m<2;m++)
		{
#if FE_TREE_PREMULTIPLY_MASS
			m_tempMatrix[m]=id-h*m_imdfdv[m]-(h*h)*m_imdfdx[m];
			premultiplyInverseDiagonal(m_A[m],m_im,m_tempMatrix[m]);

//			feLog("%d premultiply mass on A from\n%s\n",
//					m,c_print(m_tempMatrix[m]));
//			feLog("to\n%s\n",c_print(m_A[m]));
//			feLog("im\n%s\n",c_print(m_im));
#else
			m_A[m]=id-h*m_imdfdv[m]-(h*h)*m_imdfdx[m];
#endif
#if FE_TREE_PRECONDITION
			m_invSqrtA[m].reset(width(m_A[m]));
			for(U32 k=0;k<m_A[m].rows();k++)
			{
				m_invSqrtA[m][k]=1.0f/sqrt(m_A[m](k,k));
			}

			premultiplyDiagonal(m_tempMatrix[m],m_invSqrtA[m],m_A[m]);
			postmultiplyDiagonal(m_A[m],m_tempMatrix[m],m_invSqrtA[m]);

//			feLog("invSqrtA\n%s\n",c_print(m_invSqrtA[m]));
//			feLog("mid\n%s\n",c_print(m_tempMatrix[m]));
//			feLog("A'\n%s\n",c_print(m_A[m]));
#endif
//			m_hdfdx[m]=h*m_dfdx[m];
		}
	}

	m_spProfileReset->start();

	for(U32 m=0;m<2;m++)
	{
		m_force[m].reset(m_sticks);
		m_velocity[m].reset(m_sticks);
		m_addPosition[m].reset(m_sticks);
		m_addVelocity[m].reset(m_sticks);
//		m_dfdx2[m].clear();
	}

	m_spProfilePopulate->replace(m_spProfileReset);
	sp<Tree> spTree(this);
	m_root.populate_dynamic(spTree,m_force,m_velocity,
			m_addPosition,m_addVelocity,FALSE);
#if FE_TREE_EXPLICIT
	m_dfdx2[0].clear();
	m_dfdx2[1].clear();
#endif

	m_spProfileSolve->replace(m_spProfilePopulate);

#if FE_TREE_MT_SOLVE
	runSolvers();
#else
	for(U32 m=0;m<2;m++)
	{
		solve(m);
	}
#endif

	m_spProfileKine->replace(m_spProfileSolve);

	set(m_effectorForce);
	for(U32 m=0;m<m_effectorForceArray.size();m++)
	{
		set(m_effectorForceArray[m]);
	}

	for(U32 m=0;m<m_bundles;m++)
	{
		m_bundleSize[m]=0;
	}

#if FE_TREE_MT_KINE
	runKine();
#else
	m_root.forward_kine_recursive(0,spTree,m_lastDeltaT,m_deltaV,
			m_addPosition,m_addVelocity,m_effectorForce,FALSE);
#endif

	for(U32 m=0;m<m_effectorForceArray.size();m++)
	{
		m_effectorForce+=m_effectorForceArray[m];
	}

	m_spProfileKine->finish();
	m_spProfiler->end();
}

void Tree::solve(U32 m)
{
	m_spProfileSetup[m]->start();

/*
	m_A2=m_A[m]-h*premultiplyDiagonal(m_tempMatrix,m_him,m_dfdx2[m]);
	VectorN b=m_him*(m_force[m]+(m_hdfdx[m]+h*m_dfdx2[m])*m_velocity[m]+
			(m_dfdx[m]+m_dfdx2[m])*m_addPosition[m]);
*/

/*
	const F32& h=m_lastDeltaT;
	premultiplyDiagonal(m_A2[m],m_him,m_dfdx2[m]);
	m_A2[m]*= -h;
	m_A2[m]+=m_A[m];
	m_tempMatrix[m]=m_dfdx2[m];
	m_tempMatrix[m]*=h;
	m_tempMatrix[m]+=m_hdfdx[m];
	m_tempMatrix[m].transform(m_velocity[m],m_tempVector1[m]);
	m_tempVector1[m]+=m_force[m];
	m_tempMatrix[m]=m_dfdx[m];
	m_tempMatrix[m]+=m_dfdx2[m];
	m_tempMatrix[m].transform(m_addPosition[m],m_tempVector2[m]);
	m_tempVector1[m]+=m_tempVector2[m];
	m_him.transform(m_tempVector1[m],m_b[m]);
*/

//	m_A2[m]=m_A[m];

//	m_hdfdx[m].transform(m_velocity[m],m_tempVector1[m]);
	m_dfdx[m].transform(m_velocity[m],m_tempVector1[m]);
	m_tempVector1[m]*=m_lastDeltaT;

	m_tempVector1[m]+=m_force[m];
	m_dfdx[m].transform(m_addPosition[m],m_tempVector2[m]);
	m_tempVector1[m]+=m_tempVector2[m];

#if FE_TREE_PREMULTIPLY_MASS
//	m_him.transform(m_tempVector1[m],m_tempVector2[m]);
	componentMultiply(m_tempVector2[m],m_him,m_tempVector1[m]);

//	premultiplyInverseDiagonal(m_b[m],m_im,m_tempVector2[m]);
	componentMultiply(m_b[m],m_mass,m_tempVector2[m]);

//	feLog("%d premultiply M on b from\n%s\n",m,c_print(m_tempVector2[m]));
//	feLog("to\n%s\n",c_print(m_b[m]));
#else
	m_him.transform(m_tempVector1[m],m_b[m]);
#endif
#if FE_TREE_PRECONDITION
//	m_tempVector1[m]=m_b[m];
//	premultiplyDiagonal(m_b[m],m_invSqrtA[m],m_tempVector1[m]);
	componentMultiply(m_b[m],m_invSqrtA[m],m_b[m]);

//	feLog("b=<%s>\n",c_print(m_tempVector1[m]));
//	feLog("b`=<%s>\n",c_print(m_b[m]));
#endif

	m_spProfileSolver[m]->replace(m_spProfileSetup[m]);

	//* A2*dV=b solve for dV
//	m_solver[m].solve(m_deltaV[m],m_A2[m],m_b[m]);
	m_solver[m].solve(m_deltaV[m],m_A[m],m_b[m]);

	m_spProfileSolver[m]->finish();

#if FE_TREE_PRECONDITION
//	m_tempVector1[m]=m_deltaV[m];
//	premultiplyDiagonal(m_deltaV[m],m_invSqrtA[m],m_tempVector1[m]);
	componentMultiply(m_deltaV[m],m_invSqrtA[m],m_deltaV[m]);

//	feLog("dV'=<%s>\n",c_print(m_tempVector1[m]));
//	feLog("dV=<%s>\n",c_print(m_deltaV[m]));
#endif
}

void Tree::Stick::populate_static(sp<Tree>& rspTree,
		MatrixN* dfdx,MatrixN* dfdv,VectorN& invMass)
{
	if(m_pParent)
	{
		U32 parent=m_pParent->m_index;

		dfdx[0](parent,parent)-=m_spring;
		dfdx[1](parent,parent)-=m_spring;

		dfdx[0](m_index,parent)+=m_spring;
		dfdx[1](m_index,parent)+=m_spring;

		dfdx[0](parent,m_index)+=m_spring;
		dfdx[1](parent,m_index)+=m_spring;

		if(m_drag != 0.0f)
		{
			dfdv[0](parent,parent)-=m_drag;
			dfdv[1](parent,parent)-=m_drag;

			dfdv[0](m_index,parent)+=m_drag;
			dfdv[1](m_index,parent)+=m_drag;

			dfdv[0](parent,m_index)+=m_drag;
			dfdv[1](parent,m_index)+=m_drag;
		}
	}

	dfdx[0](m_index,m_index)-=m_spring;
	dfdx[1](m_index,m_index)-=m_spring;

	if(m_drag != 0.0f)
	{
		dfdv[0](m_index,m_index)-=m_drag;
		dfdv[1](m_index,m_index)-=m_drag;
	}

	invMass[m_index]=1.0f/m_mass;

	Stick* pStick;
	List<Stick*>::Iterator iterator(m_children);
	while((pStick= *iterator++)!=NULL)
	{
		pStick->populate_static(rspTree,dfdx,dfdv,invMass);
	}
}

void Tree::Stick::populate_dynamic(sp<Tree>& rspTree,VectorN* force,
		VectorN* velocity,VectorN* addPosition,VectorN* addVelocity,
		bool rekine)
{
	const F32 gravity=9.81f;

	SpatialVector externalForce=1e3f*rspTree->m_windDirection;	//* tweak

	//* HACK needs to be facing area
	const F32 wind2=magnitudeSquared(rspTree->m_windDirection);
	if(wind2>0.0f)
	{
		externalForce*=m_radius1/sqrtf(wind2)*
				fabs(dot(m_span,rspTree->m_windDirection));
	}

	externalForce[2]-=gravity*m_mass;

	//* in case parent just moved with rekine
	if(rekine && m_pParent)
	{
		m_base=m_pParent->m_base+m_pParent->m_span;
	}

	const Vector2 angle=m_rest+m_position;

#if FALSE
	for(U32 m=0;m<2;m++)
	{
		feLog("%d %d %.6G+%.6G=%.6G (%.6G)\n",
				m_index,m,m_rest[m],m_position[m],angle[m],angle[m]/degToRad);
	}
#endif

	const F32 radius_exaggerate=1.2f;	//* tweak
	SpatialVector effector=rspTree->m_effector;
	const F32 radius=radius_exaggerate*rspTree->m_effector[3];

	//* point line-segment distance
	const SpatialVector toEffect=effector-m_base;
	F32 along=dot(toEffect,m_span)/(m_length*m_length);
	SpatialVector effect;
	if(along<0.0f)
	{
		along=0.0f;
		m_contact=m_base;
		effect=m_base-effector;
	}
	else if(along>1.0f)
	{
		along=1.0f;
		m_contact=m_base+m_span;
		effect=m_base+m_span-effector;
	}
	else
	{
		SpatialVector projection=m_base+along*m_span;
		m_contact=projection;
		effect=projection-effector;
	}

	set(m_intensity);

	const F32 radius_along=m_radius1+(m_radius2-m_radius1)*along;
	const F32 radius_sum=radius+radius_along;
	const F32 damp_radius=FE_TREE_DAMP_REGION*radius_sum;

	F32 distance2=magnitudeSquared(effect);
	if(radius>0.0f && distance2>0.0f && distance2<(radius_sum*radius_sum))
	{
		const F32 distance=sqrtf(distance2)-radius_along;
		const F32 correction=distance>0.0f? 1.0f-distance/radius: 1.0f;
		const SpatialVector correctiveEffect=correction*unit(effect);

		m_contact2=effector;

		const U32 passes=4;
		for(U32 pass=0;pass<passes;pass++)
		{
			for(U32 m=0;m<2;m++)
			{
				SpatialVector residual=correctiveEffect-
						m_intensity[!m]*m_dir[!m];
				m_intensity[m]=dot(residual,m_dir[m]);
			}
		}

		Vector2 response=unit(m_intensity);

		for(U32 m=0;m<2;m++)
		{
#if FALSE
			//* raw penalty force
			const F32 push=0.0f;
			force[m][m_index]+=push*m_intensity[m];
#endif

			//* for Baraff's Position Alteration
			//* TODO non-arbitrary scale
#if FALSE
			const F32 restitution=0.0f;
			addVelocity[m][m_index]= -(1.0+restitution)*m_velocity[m];
#else
			const F32 tweak=4.0f;
			addPosition[m][m_index]= tweak*along*m_intensity[m];
#endif

#if TRUE
			// deplete kinetic energy
			m_velocity[m]=0.0f;
			addVelocity[m][m_index]=0.0f;
#else
			if((m_velocity[m]>0.0f) != (response[m]>0.0f))
			{
				addVelocity[m][m_index]= -m_velocity[m]*fabs(response[m]);
			}
			else
			{
				addVelocity[m][m_index]=0.0f;
			}
#endif

#if FALSE
			m_lastAddPosition[m]*=0.9f;
			if(m_lastAddPosition[m]<0.0f &&
					addPosition[m][m_index]>m_lastAddPosition[m])
			{
				addPosition[m][m_index]=m_lastAddPosition[m];
			}
			if(m_lastAddPosition[m]>0.0f &&
					addPosition[m][m_index]<m_lastAddPosition[m])
			{
				addPosition[m][m_index]=m_lastAddPosition[m];
			}
			m_lastAddPosition[m]=addPosition[m][m_index];
#endif
		}
		rekine=TRUE;
	}
	else
	{
		if(distance2>0.0f && distance2<(damp_radius*damp_radius))
		{
			// deplete kinetic energy
			const F32 distance=sqrtf(distance2)-radius_along;
			const F32 damp_factor=(distance-radius)/(damp_radius-radius);

			m_velocity*=damp_factor;
		}

		set(m_contact);
		set(m_contact2);

#if FALSE
		m_lastAddPosition*=0.9f;
		for(U32 m=0;m<2;m++)
		{
			addPosition[m][m_index]=m_lastAddPosition[m];
		}
#endif
	}

	Vector2 parent_position;
	Vector2 parent_velocity;
	set(parent_position);
	set(parent_velocity);
	if(m_pParent)
	{
		U32 parent=m_pParent->m_index;

		parent_position=m_pParent->m_rest+m_pParent->m_position;
		parent_velocity=m_pParent->m_velocity;

		force[0][parent]+=m_spring*(m_position[0]-parent_position[0]);
		force[1][parent]+=m_spring*(m_position[1]-parent_position[1]);
	}

	//* NOTE gravity from local node only
	//* TODO accumulate axial child gravity during kinematics
	Vector2 addForce;
#if FALSE
//	const SpatialVector downforce(0.0f,-gm*(1.0f-fabs(unit(m_span)[1])),0.0f);
	const SpatialVector downforce(0.0f,-gm,0.0f);
	SpatialVector unitdir[2];
	unitdir[0]=unit(m_dir[0]);
	unitdir[1]=unit(m_dir[1]);

	set(addForce);
	const U32 passes=4;
	for(U32 pass=0;pass<passes;pass++)
	{
		for(U32 m=0;m<2;m++)
		{
			SpatialVector residual=downforce-addForce[!m]*unitdir[!m];
			addForce[m]=dot(residual,unitdir[m]);
		}
	}
#else
	const SpatialVector unitSpan=m_span/m_length;
	SpatialVector dir=unit(m_dir[0]);
	addForce[0]=dot(dir,externalForce);
	dir=unit(m_dir[1]);
	addForce[1]=dot(dir,externalForce);

	const F32 mag2=magnitudeSquared(addForce);
	const F32 max=fabs(dot(externalForce,unitSpan));
	if(mag2>max)
	{
		addForce*=max/sqrtf(mag2);
	}
#endif

	force[0][m_index]+=addForce[0];
	force[1][m_index]+=addForce[1];

#if FALSE
	// NOTE currently messes up ball
	m_intensity[0]= 0.001f*addForce[0];
	m_intensity[1]= 0.001f*addForce[1];
#endif

	force[0][m_index]-=m_spring*(m_position[0]-parent_position[0]);
	force[1][m_index]-=m_spring*(m_position[1]-parent_position[1]);

	if(m_drag != 0.0f)
	{
		force[0][m_index]-=m_drag*(m_velocity[0]-parent_velocity[0]);
		force[1][m_index]-=m_drag*(m_velocity[1]-parent_velocity[1]);
	}

	velocity[0][m_index]=m_velocity[0];
	velocity[1][m_index]=m_velocity[1];

	if(rekine)
	{
		forward_kine(0,rspTree,0.0f,NULL,addPosition,addVelocity,
				rspTree->m_effectorForce);
	}

	Stick* pStick;
	List<Stick*>::Iterator iterator(m_children);
	while((pStick= *iterator++)!=NULL)
	{
		pStick->populate_dynamic(rspTree,force,velocity,
				addPosition,addVelocity,rekine);
	}
}

void Tree::Stick::forward_kine(U32 thread,sp<Tree>& rspTree,
		const F32 deltaT,const VectorN* deltaV,
		const VectorN* addPosition,const VectorN* addVelocity,
		SpatialVector& rEffectorForce)
{
	const BWORD transform_leaves=(deltaV!=NULL);
	Vector2 parentAngle;
	SpatialQuaternion parentRotation;
	if(m_pParent)
	{
		m_base=m_pParent->m_base+m_pParent->m_span;
		parentAngle=m_pParent->m_rest+m_pParent->m_position;
		parentRotation=m_pParent->m_rotation;
	}
	else
	{
		m_base=rspTree->m_location;
		set(parentAngle);
		set(parentRotation);
	}

#if FE_TREE_TRACE
	feLog("%d lev %d p %s v %s ap %.6G %.6G av %.6G %.6G",
			m_index,m_level,c_print(m_position),c_print(m_velocity),
			addPosition[0][m_index],addPosition[1][m_index],
			addVelocity[0][m_index],addVelocity[1][m_index]);
#endif

	if(deltaV)
	{
		m_velocity[0]+=deltaV[0][m_index];
		m_velocity[1]+=deltaV[1][m_index];

#if FE_TREE_TRACE
		feLog(" dv %.6G %.6G",
				deltaV[0][m_index],deltaV[1][m_index]);
#endif
	}
#if FE_TREE_TRACE
	feLog("\n");
#endif

	m_velocity[0]+=addVelocity[0][m_index];
	m_velocity[1]+=addVelocity[1][m_index];

	m_position[0]+=deltaT*m_velocity[0]+addPosition[0][m_index];
	m_position[1]+=deltaT*m_velocity[1]+addPosition[1][m_index];

	if(FE_INVALID_SCALAR(m_velocity[0]) || FE_INVALID_SCALAR(m_velocity[1]))
	{
		feX("Tree::Stick::forward_kine","invalid m_velocity");
	}

	Vector2 angle=m_rest+m_position-parentAngle;
#if FALSE
	F32 mag_angle=magnitude(angle);
	if(mag_angle>pi)
	{
		const Vector2 unitangle=unit(angle);
		mag_angle=2*pi-mag_angle;
		angle= -mag_angle*unitangle;
		m_position=angle-m_rest;
	}
#endif

//	SpatialVector yVector(0.0f,m_length);

	SpatialVector span;
	for(U32 pass=0;pass<3;pass++)
	{
		const F32 delta=0.01f;
		const F32 invDelta=(1.0f/delta);

		const F32 angle0=angle[0]+(pass==1? delta: 0.0f);
		const F32 angle1=angle[1]+(pass==2? delta: 0.0f);

#if FALSE
		//* reverse projection - intersection of planes
		const SpatialVector normalx(cosf(angle0),-sinf(angle0),0.0f);
		const SpatialVector normalz(0.0f,-sinf(angle1),cosf(angle1));
		const SpatialVector direction=unit(cross(normalz,normalx));
		SpatialQuaternion local_rotation;
		// TODO special set
		set(local_rotation,Vector3f(0.0f,1.0f,0.0f),direction);
//		feLog("\n%x pass %d angle %.6G %.6G\n",this,pass,angle0,angle1);
//		feLog("normalx %s\n",c_print(normalx));
//		feLog("normalz %s\n",c_print(normalz));
//		feLog("direction %s\n",c_print(direction));
//		feLog("local_rotation %s\n",c_print(local_rotation));
		const SpatialQuaternion rotation=parentRotation*local_rotation;
//		feLog("rotation %s\n",c_print(rotation));
#endif

#if TRUE
		//* hypotenuse angle with proportional axis
		const F32 phi=sqrtf(angle0*angle0+angle1*angle1);
		SpatialVector axis=SpatialVector(-angle1,angle0,0.0f);
		if(phi>1e-6f)
		{
			axis*=1.0f/phi;
		}
//		F32 phi2=0.5f*(fabs(angle0)+fabs(angle1));
		const SpatialQuaternion rotation=
				parentRotation*SpatialQuaternion(phi,axis);

//		feLog("%d %.6G %.6G %.6G  %s  %s  %s\n",
//				pass,angle0,angle1,phi,c_print(axis),
//				c_print(rotation),c_print(span));
#endif

#if FALSE
		//* universal joint (like a drive shaft)
		const SpatialQuaternion rotationx=SpatialQuaternion(angle0,e_xAxis);
		const SpatialQuaternion rotationz=SpatialQuaternion(angle1,e_zAxis);
		const SpatialQuaternion rotation=parentRotation*rotationx*rotationz;
#endif

		rotateZVector(rotation,m_length,span);
		if(!pass)
		{
			m_span=span;
			m_rotation=rotation;
			m_transform=m_rotation;
			setTranslation(m_transform,m_base);
		}
		else
		{
			m_dir[pass-1]=(span-m_span)*invDelta;
		}
	}

	// push back on effector
	const F32 tweak=8.0f;
	rEffectorForce -= tweak*(m_dir[0]*m_intensity[0]+
			m_dir[1]*m_intensity[1]);

	if(transform_leaves)
	{
		const U32 leafIndexCount=m_leafIndexArray.size();
		for(U32 leafLookup=0;leafLookup<leafIndexCount;leafLookup++)
		{
			const U32 leafIndex=m_leafIndexArray[leafLookup];
			const Tree::Leaf& rLeaf=rspTree->m_leafArray[leafIndex];
			transformVector(m_transform,rLeaf.center(),
					rspTree->m_pLeafPoints[leafIndex]);
			rotateVector(m_transform,rLeaf.normal(),
					rspTree->m_pLeafNormals[leafIndex]);
		}
	}

	// TODO try skipping during rekine
	if(m_radius1>FE_TREE_VISIBLE_MIN_RADIUS)
	{
		U32& index=rspTree->m_bundleSize[thread];

		//* TODO consider, maybe cylinders should point in Y

#if FALSE
		//* rotate object's Z axis to Y axis
		const SpatialTransform& transform=m_transform;
		SpatialTransform& cylTransform=
				rspTree->m_stickTransformBundles[thread][index];
		cylTransform=transform;
//		cylTransform.up()= -transform.direction();
//		cylTransform.direction()= transform.up();
#else
		rspTree->m_stickTransformBundles[thread][index]=m_transform;
#endif

		set(rspTree->m_stickScaleBundles[thread][index],
				m_radius2,m_radius2,m_length);
		rspTree->m_stickBaseScaleBundles[thread][index]=m_radius1/m_radius2;
		rspTree->m_stickSliceBundles[thread][index]=m_resolution;

		index++;
	}
}

void Tree::Stick::forward_kine_recursive(U32 thread,sp<Tree>& rspTree,
		const F32 deltaT,const VectorN* deltaV,
		const VectorN* addPosition,const VectorN* addVelocity,
		SpatialVector& rEffectorForce,BWORD only_zero)
{
	forward_kine(thread,rspTree,deltaT,deltaV,addPosition,addVelocity,
			rEffectorForce);

	Stick* pStick;
	List<Stick*>::Iterator iterator(m_children);
	while((pStick= *iterator++)!=NULL)
	{
		if(!only_zero || pStick->m_level==0)
		{
			pStick->forward_kine_recursive(thread,rspTree,deltaT,deltaV,
					addPosition,addVelocity,rEffectorForce,only_zero);
		}
	}
}

void Tree::forward_kine_thread(U32 thread,I32 index)
{
#if FE_TREE_MT_DEBUG
	feLog("[33mTree::forward_kine_thread %d start %p %p[0m\n",index,
			m_kineArray[index],m_spProfileKineLimbArray[index].raw());
#endif
	m_spProfileKineLimbArray[index]->start();
	sp<Tree> spTree(this);
	m_kineArray[index]->forward_kine_recursive(thread,spTree,
			m_lastDeltaT,m_deltaV,
			m_addPosition,m_addVelocity,m_effectorForceArray[thread],FALSE);
	m_spProfileKineLimbArray[index]->finish();
}

void Tree::runSolvers(void)
{
#if FE_TREE_MT_DEBUG
	feLog("[33mTree::runSolvers start[0m\n");
#endif

#if FE_TREE_SOLVE_OFF_MAIN
	m_spGang->post(FE_TREE_MT_ACTION_SOLVE|0,FE_TREE_MT_ACTION_SOLVE|1);
#else
	m_spGang->post(FE_TREE_MT_ACTION_SOLVE|1);
	solve(0);
#endif

#if FE_TREE_MT_DEBUG
	feLog("[33mTree::runSolvers posted[0m\n");
#endif

	U32 jobs=FE_TREE_SOLVE_OFF_MAIN? 2: 1;
	while(jobs)
	{
		I32 job=0;
		m_spProfileSolveWait->start();
		m_spGang->waitForDelivery(job);
		m_spProfileSolveWait->finish();
		jobs--;

#if FE_TREE_MT_DEBUG
		feLog("[33mTree::runSolvers accepted %10p[0m\n",job);
#endif
	}

#if FE_TREE_MT_DEBUG
	feLog("[33mTree::runSolvers done[0m\n");
#endif
}

void Tree::runKine(void)
{
#if FE_TREE_MT_DEBUG
	feLog("[33mTree::runKine start[0m\n");
#endif

	sp<Tree> spTree(this);
	m_root.forward_kine_recursive(0,spTree,m_lastDeltaT,m_deltaV,
			m_addPosition,m_addVelocity,m_effectorForceArray[0],TRUE);

	U32 jobs=m_kineArray.size();
	m_spGang->post(FE_TREE_MT_ACTION_KINE|0,FE_TREE_MT_ACTION_KINE|(jobs-1));

	I32 job=0;
	while(m_spGang->take(job))
	{
#if FE_TREE_MT_DEBUG
		feLog("[33mTree::runKine take     %10p[0m\n",job);
#endif
		forward_kine_thread(0,job&FE_TREE_MT_INDEX_MASK);
#if FE_TREE_MT_DEBUG
		feLog("[33mTree::runKine deliver  %10p[0m\n",job);
#endif
		m_spGang->deliver(job);
	}
#if FE_TREE_MT_DEBUG
		feLog("[33mTree::runKine all taken[0m\n");
#endif
	while(jobs)
	{
		m_spProfileKineWait->start();
		m_spGang->waitForDelivery(job);
		m_spProfileKineWait->finish();
		jobs--;

#if FE_TREE_MT_DEBUG
		feLog("[33mTree::runKine accepted %10p[0m\n",job);
#endif
	}

#if FE_TREE_MT_DEBUG
	feLog("[33mTree::runKine done[0m\n");
#endif
}

void Tree::SolverWorker::operator()(void)
{
	const U32 thread=m_id+1;

#if FE_TREE_MT_DEBUG
	feLog("[32mTree::SolverWorker::operator()"
			" %p Thread %d starting[0m\n",this,thread);

	BWORD wait_message=TRUE;
#endif

	BWORD started=FALSE;

	I32 job=0;
	I32 index=0;
	while(TRUE)
	{
		if(started)
		{
			try
			{
				m_hpTree->m_spProfileThreadWait->start();
			}
			catch(Exception &e)
			{
				if(e.getResult()!=e_invalidHandle) { throw e; }
				feLog("[33mTree::SolverWorker::operator()"
						" m_hpTree invalid[0m\n\n");
			}
		}

		m_hpJobQueue->waitForJob(job);
		if(job==FE_TREE_MT_ACTION_QUIT)
		{
#if FE_TREE_MT_DEBUG
			feLog("[32mTree::SolverWorker::operator()"
					" %p Thread %d Job %10p break[0m\n",this,thread,job);
#endif
			m_hpJobQueue->deliver(job);
			break;
		}

		if(started)
		{
			m_hpTree->m_spProfileThreadWait->finish();
		}

#if FE_TREE_MT_DEBUG
		feLog("[32mTree::SolverWorker::operator()"
				" %p Thread %d Job %10p[0m\n",
				this,thread,job);
#endif

		started=TRUE;

		index=job&FE_TREE_MT_INDEX_MASK;

		if(job&FE_TREE_MT_ACTION_SOLVE)
		{
			m_hpTree->solve(index);
		}
		else if(job&FE_TREE_MT_ACTION_KINE)
		{
			m_hpTree->forward_kine_thread(thread,index);
		}

#if FE_TREE_MT_DEBUG
		feLog("[32mTree::SolverWorker::operator()"
				" %p Thread %d Job %10p finished[0m\n",
				this,thread,job);
		wait_message=TRUE;
#endif

		m_hpJobQueue->deliver(job);
	}
#if FE_TREE_MT_DEBUG
	feLog("[32mTree::SolverWorker::operator()"
			" %p Thread %d leaving[0m\n",this,thread);
#endif
}


} // namespace
