/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#ifndef __core_Initialized_h__
#define __core_Initialized_h__

namespace fe
{

/**************************************************************************//**
	@brief Base class providing collective initialization (post-constructor)

	@ingroup core

	This mechanism allows you to supply multiple initialization functions
	in a class hierarchy, all of which are called.  This operation is
	similar to construction and is intended to supply objects with the ability
	to do initialization that is not safe or possible during construction.

	This mechanism works for multiple inheritance and virtual inheritance.
	In order to follow the same call order as the constructors, any base
	classes that may contain initialize functions should precede the
	Initialize template in the inheritance specification.

	To participate, any derived class @em T supplying an @em initialize method
	needs to also derive from Initialize <T>.

	@code
class A: virtual public Initialized, public Initialize<A>
{
	public:
virtual			~A(void)			{}
		void	initialize(void)	{ feLog("initialize A\n");
};

class B: public A, public Initialize<B>
{
	public:
virtual			~B(void)			{}
		void	initialize(void)	{ feLog("initialize B\n");
};

int main(int argc,char** argv,char** env)
{
	B b;
	b.initialize();		// just calls B::initialize
	b.initializeAll();	// calls A::initialize and B::initialize
}
	@endcode
*//***************************************************************************/
class Initialized
{
	public:
				Initialized(void) { m_active = true; }
virtual			~Initialized(void)											{}

		void	initializeAll(void)
				{	if(!m_active) return;
					int m,size=m_initializers.size();
					for(m=0;m<size;m++)
						m_initializers[m](this); }

		void	setActive(bool a_active) { m_active = a_active; }

	protected:

typedef void	(*InitializeFunction)(Initialized*);

		void	addInitializer(InitializeFunction function)
				{	m_initializers.push_back(function); }

	private:

		std::vector<InitializeFunction>	m_initializers;
		bool							m_active;
};

/**************************************************************************//**
	@brief Per-class participation in the Initialized <> mechanism
*//***************************************************************************/
template<class T>
class Initialize: virtual public Initialized
{
	public:
				Initialize(void)
				{	addInitializer(&Initialize<T>::static_initialize); }

	private:

static	void	static_initialize(Initialized* obj)
				{	dynamic_cast<T*>(obj)->initialize(); }
};

} // namespace

#endif // __core_Initialized_h__
