/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "core.pmh"

#define	FE_TRACK_TRACE			FALSE

#define	FE_TRACK_EXCEPTION		TRUE
#define	FE_TRACK_SUPPRESSIONS	TRUE
#define	FE_ASSERT_NAMES			FALSE	// turn on to find unnamed entries

namespace fe
{

Tracker::~Tracker(void)
{
}

U32 Tracker::totalCount(void) const
{
	SAFEGUARD;
	U32 sum=0;

	std::map<void*,Track>::const_iterator it=m_trackMap.begin();
	while(it!=m_trackMap.end())
	{
		if(!it->second.m_suppress)
		{
			sum+=it->second.m_count;
		}
		it++;
	}

	return sum;
}

void Tracker::suppress(void *pPtr)
{
#if FE_TRACK_SUPPRESSIONS
	SAFEGUARD;
	std::map<void*,Track>::iterator it=m_trackMap.find(pPtr);
	if(it==m_trackMap.end())
	{
		change(false,pPtr,"",0);
		it=m_trackMap.find(pPtr);
	}
	if(it!=m_trackMap.end())
		it->second.m_suppress=true;
#endif
}

void Tracker::change(BWORD release,void *pPtr,String name,U32 newcount)
{
	SAFEGUARD;
	std::map<void*,Track>::iterator it=m_trackMap.find(pPtr);
	if(it==m_trackMap.end())
	{
#if FE_TRACK_TRACE==FALSE
		if(newcount>2)
#endif
			feLog("%s Tracker %p new%s%2d %s\n",m_name.c_str(),pPtr,
					release? "--": "++",newcount,name.c_str());
		if(!release || newcount)
		{
			Track& track=m_trackMap[pPtr];
			track.m_name=name;
			track.m_count=newcount;
		}
	}
	else
	{
#if FE_TRACK_TRACE==FALSE
		if(it->second.m_count && newcount &&
				it->second.m_count+(release? -1: 1) != newcount)
#endif
			feLog("%s Tracker %p %2d %s%2d %s\n",m_name.c_str(),pPtr,
					it->second.m_count,
					release? "--": "++",newcount,name.c_str());
#if FE_TRACK_EXCEPTION
		if(it->second.m_count && newcount &&
				it->second.m_count+(release? -1: 1) != newcount)
			feX("Tracker::change","mistrack (thread collision?)");
#endif
		it->second.m_name=name;
		it->second.m_count=newcount;
		if(release && !newcount)
			m_trackMap.erase(it);

#if FE_ASSERT_NAMES
	//* Please supply a name for your Counted object!
	if(release)
		FEASSERT(!name.empty() && name!="Counted");
#endif
	}
}

void Tracker::registerRegion(void* pPtr,void* pT,U32 size,String name)
{
	SAFEGUARD;
	Track& track=m_trackMap[pPtr];
	track.m_base=pT;
	track.m_size=size;
	track.m_name=name;
}

void Tracker::deregisterRegion(void* pPtr)
{
	SAFEGUARD;
	std::map<void*,Track>::iterator it=m_trackMap.find(pPtr);
	if(it!=m_trackMap.end())
		m_trackMap.erase(it);
}

void Tracker::changeReference(BWORD remove,void* pPtr,void* pReference)
{
	SAFEGUARD;
	std::map<void*,Track>::iterator it=m_trackMap.find(pPtr);
	if(it==m_trackMap.end())
		return;

	if(remove)
		it->second.m_references.remove(pReference);
	else
		it->second.m_references.append(pReference);
}

String Tracker::report(void) const
{
	SAFEGUARD;
	if(m_trackMap.empty())
		return "Tracker: all clear\n";

	String message="Tracker's outstanding objects:\n";

	U32 suppressed=0;
	std::map<void*,Track>::const_iterator it=m_trackMap.begin();
	while(it!=m_trackMap.end())
	{
		U32& referenceFound=
				*const_cast<U32*>(&it->second.m_referencesFound);
		referenceFound=0;

		if(it->second.m_suppress)
		{
			if(it->second.m_count)
				suppressed++;
			it++;
			continue;
		}
		message.cat("%08p ",it->first);
		if(it->second.m_references.size()!=it->second.m_count)
			message.cat("%d/",it->second.m_references.size());
		message.cat("%d %s",it->second.m_count,it->second.m_name.c_str());
		if(it->second.m_size)
			message.cat(" (%p+%d bytes)",it->second.m_base,it->second.m_size);
		message.cat("\n");
		void* pReference;
		List<void*>::Iterator iRef(
				(List<void*>&)(it->second.m_references));
		iRef.toHead();
		while((pReference= *iRef++)!=NULL)
		{
			BWORD found=false;

			std::map<void*,Track>::const_iterator it2=m_trackMap.begin();
			while(it2!=m_trackMap.end())
			{
				if( pReference==it2->first ||
						(pReference>it2->second.m_base &&
						(UWORD)pReference<
						(UWORD)(it2->second.m_base)+
						(UWORD)(it2->second.m_size)))
				{
					if(!found)
						message.cat("  %p",pReference);
					message.cat(" -> %p+%d %s",
							it2->second.m_base,
							it2->second.m_size,it2->second.m_name.c_str());
					found=true;
					referenceFound++;
				}
				it2++;
			}
			if(found)
				message.cat("\n");
#if TRUE
			else
				message.cat("  %p\n",pReference);
#endif
		}

		it++;
	}
	message.cat("\nSuppressions: %d\n",suppressed);

	message.cat("\nHoldouts:\n");
	it=m_trackMap.begin();
	while(it!=m_trackMap.end())
	{
		message.cat(reportAt(it,0));
		it++;
	}

	return message;
}

String Tracker::reportAt(std::map<void*,Track>::const_iterator it,
		U32 depth) const
{
	SAFEGUARD;
	if(it->second.m_suppress ||
			(!depth && it->second.m_referencesFound==1))
		return "";

	String message;
	BWORD found=false;
	std::map<void*,Track>::const_iterator it2=m_trackMap.begin();
	while(it2!=m_trackMap.end())
	{
		if(it2->second.m_suppress)
		{
			it2++;
			continue;
		}

		void* pReference;
		List<void*>::Iterator iRef(
				(List<void*>&)(it2->second.m_references));
		iRef.toHead();
		while((pReference= *iRef++)!=NULL)

		if( pReference==it->first ||
				(pReference>it->second.m_base &&
				(UWORD)pReference<
				(UWORD)(it->second.m_base)+
				(UWORD)(it->second.m_size)))
		{
			if(!depth && !found)
			{
				message.cat("%d %s\n",it->second.m_count,
						it->second.m_name.c_str());
			}
			message.cat("%*c%d %s\n",4*(depth+1),' ',it2->second.m_count,
					it2->second.m_name.c_str());
			found=true;

			if(it2->second.m_referencesFound<2)
			{
				message.cat(reportAt(it2,depth+1));
			}
		}
		it2++;
	}

	return message;
}

} /* namespace */
