/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "core.pmh"

namespace fe
{

#define FEVARIANT_WARNINGS				true
#define FEVARIANT_DEBUG				false


//* static
bool BaseVariantType::baseRegister(TypeMaster *master,U32 owner,
			BaseType *pType,BaseType **ppStaticType,FastExMutex &rmType,
			String *argName)
{
	String localName;

	if(!master) { delete pType; return 0; }

	if(argName)
	{
		localName= *argName;
	}
	else
	{
		localName.SPrintf("%p",(UWORD)ppStaticType);

		feLog("AutoRegister %s\n",localName.c_str());
	}

	{ Guard gType(rmType);

		if(!*ppStaticType)
		{
			pType->setOwner(owner);
		}
		else
		{
			delete pType;
			pType = *ppStaticType;
		}

		pType->addName(localName);
		*ppStaticType = master->registerType(pType);
	}

	return (pType == *ppStaticType);
}


Variant::~Variant(void)
{
	if(m_pBase)
		delete m_pBase;
}


void Variant::operator=(UWORD ignored)
{
	FEASSERT(ignored==0);
	if(m_pBase)
		delete m_pBase;
	m_pBase=NULL;
}


void Variant::operator=(const Variant &vs)
{
	FEASSERT( (!m_pBase && !vs.m_pBase) ||
									m_pBase!=vs.m_pBase );
	if(m_pBase)
		delete m_pBase;
	if(vs.m_pBase)
		m_pBase=vs.m_pBase->clone();
	else
		m_pBase=NULL;
}


#if FEVARIANT_CLONING
// create new m_pBase of same type and specified size
void *Variant::malloc(UWORD size)
{
	if(!m_pBase)
		return NULL;

	BaseVariantType *old_holder=m_pBase;

	void *buffer=allocate(size);
	m_pBase=old_holder->clone(buffer);

	delete old_holder;

	return buffer;
}


Variant Variant::getOffsetElement(UWORD offset)
{
	Variant indexed;
	indexed.m_pBase=m_pBase->subClone(offset);
	return indexed;
}
#endif


Variant *Variant::pointTo(Variant *vs)
{
	if(m_pBase)
		delete m_pBase;
	if(!vs)
		return vs;
	m_pBase=new VariantType<Variant>(vs);
//	fprintf(stderr,"holder id %u\n", m_pBase->GetClassID());
	return vs;
}


Variant::Result Variant::baseCopyPointerTo(void **tptr,VariantID staticT) const
{
	if(!tptr)
		return invalidPointer;

	VariantID voidT=FE_VTYPE(void);

#if FEVARIANT_DEBUG
	feLog("Variant copy type %p to %p (\"%s\" to \"%s\")\n",
			m_pBase->GetClassID(),staticT,
			ms_pTypeMaster->lookupID(m_pBase->getClassID())->nameDump().c_str(),
			ms_pTypeMaster->lookupID(staticT)->nameDump().c_str() );
#endif
	if(!m_pBase)
	{
#if FEVARIANT_DEBUG
		feLog("Variant::copyPointerTo() m_pBase==NULL\n");
#endif
		return notInitialized;
	}
	//* NOTE if either is (void *), ignore check
	if(staticT==voidT || m_pBase->getClassID()==voidT ||
							m_pBase->getClassID()==staticT)
	{
		*tptr=m_pBase->getRawPointer();
		return variantCast;
	}
#if FEVARIANT_WARNINGS
	else
		feLog("Variant can't copy type %p into %p (\"%s\" to \"%s\")\n",
			m_pBase->getClassID(),staticT,
			ms_pTypeMaster->lookupID(m_pBase->getClassID())->nameDump().c_str(),
			ms_pTypeMaster->lookupID(staticT)->nameDump().c_str() );
#endif
	return cannotFind;
}


Variant::Result Variant::convertCreate(Variant &variant)
{
	void *temp=variant.getRawPointer();
	Variant::Result result=baseConvertCreate(temp,variant.getClassID());

	return result;
}


Variant::Result Variant::convertInto(Variant &variant)
{
	void *temp=NULL;
	Variant::Result result=baseConvertCreate(temp,variant.getClassID());
	if(result == variantCast || result == variantCopy)
	{
		variant.baseType()->copy(variant.getRawPointer(),temp);
	}

	if(result==variantCopy)
		variant.baseType()->deleteInstance(temp);

	return result;
}


Variant::Result Variant::baseConvertCreate(void *&tptr,VariantID staticT) const
{
	VariantID voidT=FE_VTYPE(void);

/*
	feLog("holder=%p T=%p void=%p I32=%p F32=%p\n",
			m_pBase->GetClassID(),staticT,voidT,
			FE_VTYPE(I32),FE_VTYPE(F32));
*/

	if(!m_pBase)
		return notInitialized;

	//* if either is (void *), ignore check
	if(staticT==voidT || m_pBase->getClassID()==voidT ||
							m_pBase->getClassID() == staticT)
	{
		tptr=m_pBase->getRawPointer();
		return variantCast;
	}
	else if(ms_pTypeMaster)
	{
		TypeConverter converter=ms_pTypeMaster->lookupConverter(
											m_pBase->getClassID(),staticT);

		if(converter)
		{
#if FEVARIANT_DEBUG
	feLog("Variant converting type %p to %p (\"%s\" to \"%s\")\n",
			m_pBase->getClassID(),staticT,
			ms_pTypeMaster->lookupID(m_pBase->getClassID())->nameDump().c_str(),
			ms_pTypeMaster->lookupID(staticT)->nameDump().c_str() );
#endif

			return converter(m_pBase->getRawPointer(),tptr);
		}
#if FEVARIANT_WARNINGS
	feLog("Variant cannot convert type %p to %p (\"%s\" to \"%s\")\n",
			m_pBase->getClassID(),staticT,
			ms_pTypeMaster->lookupID(m_pBase->getClassID())->nameDump().c_str(),
			ms_pTypeMaster->lookupID(staticT)->nameDump().c_str() );
#endif
	}

	return cannotFind;
}

} /* namespace */
