/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#ifndef __core_Variant_h__
#define __core_Variant_h__

namespace fe
{

#ifndef SUPPORT_VARIANT
typedef	U32			VariantID;
#endif

#define FEOWNER_DEBUG				false	//* verbosity

//* register on construction if not done manually (may not cross DLL's)
#define FEVARIANT_AUTOREGISTER		false

#define	FEVARIANT_COPY				FE_OK
#define	FEVARIANT_CAST				FE_true


#define FE_VTYPE(T)					VariantType< T >::getStaticID()
#define FE_ISTYPE(T,V)				(V.isType(FE_VTYPE(T)))


template<class T> class VariantType;


/**************************************************************************//**
*//***************************************************************************/
class BaseVariantType
{
	public:
							/// Get the variant ID for the specific type.
		VariantID			getClassID(void) const	{ return m_classid; }

							/** Create another handle to the object and
								return a generic handle to it. */
		BaseVariantType		*clone(void)	{ return newVariantType(m_tptr); }

							/** (internal) Return the BaseType associated
								with the specific type. */
virtual	BaseType			*baseType(void)									=0;

							/// (internal) Access the contained object.
		void				*getRawPointer(void)	{ return (void *)m_tptr; }

							/// (internal) Destroy the contained object.
virtual	void				deletePointer(void)								=0;

	protected:
							/** (internal) Non-template registration method
								called from VariantType<T>::Register(). */
static	bool				baseRegister(TypeMaster *master,U32 owner,
								BaseType *pType,BaseType **ppStaticType,
								FastExMutex &rmType,String *name=NULL);

							/// (internal) Set the variant type identifier.
		void				setClassID(VariantID set)	{ m_classid=set; }

							/** (internal) Create a new instance of the derived
								VariantType<T>. */
virtual	BaseVariantType		*newVariantType(void *newptr)					=0;

		void				*m_tptr;

	private:
		VariantID			m_classid;
};


#define	FE_MAKESTRING(T)	#T
/**************************************************************************//**
*//***************************************************************************/
template<class T>
class VariantType: public BaseVariantType
{
	public:
					/// (internal) Constructed by Variant.
					VariantType(void *initval)
					{
#if FEVARIANT_AUTOREGISTER
						if(!ms_pType)
						{
							String name=typeid(T).name();
							baseRegister(Variant::getTypeMaster(),0,
									new Type<T>,&ms_pType,ms_mType,&name);
						}
#endif

#if FE_CODEGEN<=FE_DEBUG
						if(!ms_pType)
							FELOG(">>> WARNING: type \"%s\" not registered\n",
									FE_MAKESTRING(T));
#endif
						FEASSERT(ms_pType);
						setClassID(getStaticID());
						m_tptr=initval;
					}

					//* see BaseVariantType
virtual	void		deletePointer(void)
					{	delete (T*)m_tptr;
						m_tptr=NULL; }

					//* see BaseVariantType
virtual	BaseType	*baseType(void)
					{	FEASSERT(ms_pType);
						return ms_pType; }

					/**	Register this templated type by name with the given
						typemaster.  The optional owner argument is a unique
						dynamic library identifier
						(0 for the base application). */
static	bool		registerType(String name,TypeMaster *master, U32 owner)
					{
						return baseRegister(master,owner,
									new Type<T>,&ms_pType,ms_mType,&name);
					}

					/// (internal)
static	VariantID	getStaticID(void)
					{
#if FE_CODEGEN<=FE_DEBUG
						if(!ms_pType)
							FELOG(">>> WARNING: type \"%s\" not registered\n",
									FE_MAKESTRING(T));
#endif
						FEASSERT(ms_pType);
						return ms_pType->id(); }

					/** (internal) Return the BaseType associated with the
						specific type. */
static	BaseType	*staticBaseType(void)
					{	FEASSERT(ms_pType);
						return ms_pType; }


	protected:
							//* see BaseVariantType
virtual	BaseVariantType	*newVariantType(void *newptr)
							{	return new VariantType<T>(newptr); }

	private:
static	BaseType				*ms_pType;
static	FastExMutex				ms_mType;

};


/**************************************************************************//**
*//***************************************************************************/
class Variant
{
	public:
		enum
		{
			variantCast,
			variantCopy,
			notInitialized,
			invalidPointer,
			cannotFind
		}	Result;
	public:
					/** General constructor.  The optional argument allows for
						assignment to NULL.  A non-zero argument may assert. */
					Variant(UWORD ignored=0)
					{
						FEASSERT(ignored==0);
						m_pBase=NULL;
					};

					/// Copy constructor
					Variant(const Variant &vs)
					{
						m_pBase=NULL;
						operator=(vs);
					};

virtual				~Variant(void);

					/** Operator specifically intended for assignment to NULL.
						A non-zero integer argument may assert. */
		void		operator=(UWORD ignored);

					/// Deep copy.
		void		operator=(const Variant &vs);

					/// (internal) Globally hook in a typemaster.
static	void		setTypeMaster(TypeMaster *pTypeMaster)
					{	ms_pTypeMaster=pTypeMaster; }

					/// Access the attched typemaster.
static	TypeMaster	*getTypeMaster(void)
					{	return ms_pTypeMaster; }

					//* Returns true if argument is of the same type
//		bool		IsPointerToTypeOf(const Variant &vs);

					/// Return true if id matches the type of this variant.
		bool		isType(VariantID id)
					{	return id==m_pBase->getClassID();	};

					/// Return true if this variant is a NULL pointer.
		bool		isEmpty(void)
					{	return m_pBase==NULL;	};

					/// Return the class id of the contained pointer.
		VariantID	getClassID(void)
					{	return m_pBase? m_pBase->getClassID(): 0; };

					/// delete the object pointed to as whatever type it is
		void		deletePointer(void)
					{	if(m_pBase)
							m_pBase->deletePointer();
						operator=(0L); }

					/** Return a handle to the contained pointer's type
						in the non-template base form. */
		BaseType	*baseType(void)
					{	return m_pBase? m_pBase->baseType(): NULL; }

					/// Access the stored pointer, bypassing all type checking.
		void		*getRawPointer(void)
					{	return m_pBase? m_pBase->getRawPointer(): NULL; }

					/** Make this variant contain a pointer to a variant.
						The normal copy constructor and copy operator methods
						won't work since they will assume you are trying to
						copy what is <em>in</em> the given variant. */
		Variant		*pointTo(Variant *vs);

					/** Copy the data in this variant's pointed-to object into
						the given variant's object.  If the types
						are not exactly compatible, the list of converters
						is searched.  If an appropriate converter is found,
						it is called.  Otherwise, a failure is returned.

						Returns FEVARIANT_CAST for an exact or castable match,
						FEVARIANT_COPY for a successful conversion,
						FE_E_NOT_INITIALIZED if this variant is empty,
						FE_E_INVALID_POINTER for an empty variant argument,
						or FE_E_CANNOT_FIND for a type mismatch. */
		Variant::Result			convertCreate(Variant &variant);

					//			`template member functions'

					/** Construct on arbitrary pointer and retain the type
						of that pointer. */
					template<class T>
					Variant(T *tptr)
					{	m_pBase= tptr? new VariantType<T>(tptr): NULL; }

					/** Set the state using an arbitrary pointer and
						retain the type of that pointer. */
					template<class T>
		T			*operator=(T *tptr)
					{
						delete m_pBase;
						m_pBase= tptr? new VariantType<T>(tptr): NULL;

						return tptr;
					}

					/** Copy the contained pointer into the argument if
						they are exactly compatible types.  No conversions
						are made.

						Returns FEVARIANT_CAST for an exact match,
						FE_E_NOT_INITIALIZED if this variant is empty,
						FE_E_INVALID_POINTER for a NULL pointer argument,
						or FE_E_CANNOT_FIND for a type mismatch. */
					template<class T>
		Variant::Result		copyPointerTo(T **tptr) const
					{	return baseCopyPointerTo((void**)tptr,FE_VTYPE(T)); }

					//* TODO conversion operators
					/** Copy the data in this variant's pointed-to object into
						the object pointed to by the object.  If the types
						are not exactly compatible, the list of converters
						is searched.  If an appropriate converter is found,
						it is called.  Otherwise, a failure is returned.
						The type T that this function is templated on must
						have an method T::operator=(T).

						If a conversion occurs, the argument will be replaced
						by a new object of the given type or a compatible 
						derived type.  It is the caller's responsibility
						to delete the object.  If an exact match occurs,
						it will point to the existing object and probably
						shouldn't be deleted by the caller.

						Returns FEVARIANT_CAST for an exact or castable match,
						FEVARIANT_COPY for a successful conversion,
						FE_E_NOT_INITIALIZED if this variant is empty,
						or FE_E_CANNOT_FIND for a type mismatch. */
					template<class T>
		Variant::Result		convertCreate(T *&tptr) const
					{	return baseConvertCreate((void*&)tptr,FE_VTYPE(T)); }

					template<class T>
		Variant::Result		convertInto(T *tptr) const
					{	return convertInto(Variant(tptr)); }

		Variant::Result		convertInto(Variant &variant);

	private:

		Variant::Result		baseCopyPointerTo(void **tptr,
							VariantID staticT) const;
		Variant::Result		baseConvertCreate(void *&tptr,
							VariantID staticT) const;

		BaseVariantType	*m_pBase;
static	TypeMaster		*ms_pTypeMaster;
};


/**************************************************************************//**
*//***************************************************************************/
template<class T> T dereference(Variant &variant)
{
	T *pointer=NULL;
	variant.copyPointerTo(&pointer);

//	printf("T pointer=%p id=%p\n",pointer,variant.GetClassID());

	if(pointer)
		return *pointer;

	return (T)0;
}


/**************************************************************************//**
*//***************************************************************************/
template<class T> T cast(Variant &variant)
{
	T pointer=NULL;
	variant.copyPointerTo(&pointer);

//	printf("T* pointer=%p id=%p\n",pointer,variant.GetClassID());

	if(pointer)
		return pointer;

//	T *ppointer=NULL;
//	variant.CopyPointerTo(&ppointer);
//	printf("T** ppointer=%p\n",ppointer);
//	return ppointer? *ppointer: NULL;

	return NULL;
}


/**************************************************************************//**
*//***************************************************************************/
template<class T>
class Temp
{
	public:
			Temp(T value)		{ m_value=value; }
		T	*operator &(void)	{ return &m_value; }

	private:
			T	m_value;
};


inline void VariantType<void>::deletePointer(void)
{	deallocate(m_tptr);
	m_tptr=NULL; }


/***	Stroustrup C++ 2nd Ed r.9.4 pg548
		"Static members of a global class have external linkage."
		however,
		Stroustrup C++ 2nd Ed r.3.3 pg484
		"There must be exactly one definition for each function, object,
		class and enumerator used in a program."
		This compiles under g++ and VC6, so does this link into one definition
		even after it is included in every file?
		The same seems to apply to enumerators found in headers. */

template<class T> BaseType *VariantType<T>::ms_pType = NULL;

template<class T> FastExMutex VariantType<T>::ms_mType;

} /* namespace */


#endif /* __core_Variant_h__ */
