/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#ifndef __data_Accessor_h__
#define __data_Accessor_h__

#if FE_CODEGEN<=FE_DEBUG
#define FE_ATTRIBUTE_DEBUG
#define FE_ACCESSOR_DEBUG
#endif

namespace fe
{

/** @brief Type inspecific Accessor.

	An accessor is a type of functor that is bound to a particular attribute.
	It is useful for attribute specific operations, most notably accessing a
	attribute within a record.  An accessor can also be used to check if a
	attribute exists in a Record, RecordArray, Bag, or Layout.  An accessor is
	also a convienient way to setup attributes in a Scope.
	*/
class FE_DL_EXPORT BaseAccessor
{
	public:
		class FE_DL_PUBLIC Exception : public fe::Exception
		{
			public:
				Exception(const char *a_message) : fe::Exception(a_message) {}
		};

		BaseAccessor(void);
		~BaseAccessor(void);
		/**	initialize functions setup the accessor but do not setup anything
			with the Scope. */
		void			initialize(	sp<Scope> scope,
									const String &attribute);
		/**	initialize functions setup the accessor but do not setup anything
			with the Scope. */
		void			initialize(	Scope *pScope,
									const String &attribute);
		/**	initialize functions setup the accessor but do not setup anything
			with the Scope. */
		void			initialize(		sp<Scope> scope,
										sp<Attribute> spAttribute);
		/** setup functions setup the accessor and also setup the Scope. */
		void			setup(		sp<Scope> spScope,
									const String &attribute);
		/** setup functions setup the accessor and also setup the Scope. */
		void			setup(		sp<Scope> scope,
									const String &attribute,
									const String &attributetype);

		sp<Scope>		scope(void) const;

		bool			operator==(const BaseAccessor &other) const;
		bool			operator<(const BaseAccessor &other) const;

		BaseAccessor	&operator=(const BaseAccessor &other);
						operator UWORD() const;
		UWORD			index(void) const;
		/** Return true if @em record has attribute */
		bool			check(const Record& r) const;
		/** Return true if @em record has attribute */
		bool			check(const WeakRecord r) const;
		/** Return true if @em record has attribute */
		bool			check(Record *pR) const;
		/** Return true if @em spRA has attribute */
		bool			check(sp<RecordArray>& rspRA) const;
		/** Return true if @em rRA has attribute */
		bool			check(RecordArray &rRA) const;
		/** Return true if @em spL has attribute */
		bool			check(sp<Layout>& rspL) const;
		/** Return true if @em L has attribute */
		bool			check(Layout &rL) const;

		/**	Throw exception on failure */
		void			check(const Record& r, const char *message) const;
		/**	Throw exception on failure */
		void			check(const WeakRecord r, const char *message) const;
		/**	Throw exception on failure */
		void			check(Record *pR, const char *message) const;
		/**	Throw exception on failure */
		void			check(sp<RecordArray>& rspRA,const char *message) const;
		/**	Throw exception on failure */
		void			check(RecordArray &rRA, const char *message) const;
		/**	Throw exception on failure */
		void			check(sp<Layout>& rspL, const char *message) const;
		/**	Throw exception on failure */
		void			check(Layout &rL, const char *message) const;


		/** Return the attribute this accessor is for */
		sp<Attribute>	attribute(void) const;
		const String	&name(void) const;

	protected:
		UWORD				m_index;
		hp<Scope>			m_hpScope;
		TypeInfo			m_typeInfo;
};


/** @brief The main data access class for the data system.

	@ingroup data
	*/
template <class T>
class FE_DL_EXPORT Accessor : public BaseAccessor
{
	public:
		Accessor(void);
		Accessor(sp<Scope> scope, const String &attribute);
		~Accessor(void);

		Accessor<T>	&operator=(const Accessor<T> &other);
		/** Return the attribute.  Checking for existence is not done */
		T			&operator()(const Record& r);
		/** Return the attribute.  Checking for existence is not done */
const	T			operator()(const Record& r) const;
		/** Return the attribute.  Checking for existence is not done */
		T			&operator()(const WeakRecord r);
		/** Return the attribute.  Checking for existence is not done */
const	T			operator()(const WeakRecord r) const;
		/** Return the attribute.  Checking for existence is not done */
		T			&operator()(Record *record);
		/** Return the attribute.  Checking for existence is not done */
		T			&operator()(sp<RecordArray>& rspRA, UWORD index);
		/** Return the attribute.  Checking for existence is not done */
const	T			&operator()(sp<RecordArray>& rspRA, UWORD index) const;
		/** Return the attribute.  Checking for existence is not done */
		T			&operator()(hp<RecordArray>& rhpRA, UWORD index);

		/** Return the attribute.  Checking for existence is done.
			If the attribute does not exist then an Exception is thrown */
		T			&operator()(const Record& r, const char *message);
		/** Return the attribute.  Checking for existence is done.
			If the attribute does not exist then an Exception is thrown */
const	T			operator()(const Record& r, const char *message) const;
		/** Return the attribute.  Checking for existence is done.
			If the attribute does not exist then an Exception is thrown */
		T			&operator()(const WeakRecord r, const char *message);
		/** Return the attribute.  Checking for existence is done.
			If the attribute does not exist then an Exception is thrown */
const	T			operator()(const WeakRecord r, const char *message) const;
		/** Return the attribute.  Checking for existence is done.
			If the attribute does not exist then an Exception is thrown */
		T			&operator()(Record *record, const char *message);


#if 0
		/** Return the attribute.  Checking for existence is done.
			If the attribute does not exist then the second argument is
			substituted */
		T			&operator()(const Record record, T &data);
		/** Return the attribute.  Checking for existence is done.
			If the attribute does not exist then the second argument is
			called */
		void		get(T &a_data, const Record a_record);
		/** Return the attribute.  Checking for existence is done.
			If the attribute does not exist then the second argument is
			substituted */
		T			&operator()(Record *record, T &data);
		/** Return the attribute.  Checking for existence is done.
			If the attribute does not exist then an Exception is thrown */
#endif
		T			&get(const Record record, const char *message = NULL);
		/** Return the attribute.  Checking for existence is done.
			If the attribute does not exist then an Exception is thrown */
		T			&get(Record *record, const char *message = NULL);
		/** Return true if the attribute exists in @em record.  Return the
			attribute data in @data */
		bool		queryAttribute(const Record& r, T &data);
		/** Return true if the attribute exists in @em record.  Return the
			attribute data in @data */
		bool		queryAttribute(const WeakRecord r, T &data);
		/** Return true if the attribute exists in @em layout.  Return the
			attribute data in @data.  Assumes the state block @em is of 
			the corresponding layout */
		bool		queryAttribute(sp<Layout>& rspL, void *sb, T *&data);
		/** Return true if the attribute exists in @em layout.  Return the
			attribute data in @data.  Assumes the state block @em is of 
			the corresponding layout */
		bool		queryAttribute(hp<Layout>& rhpL, void *sb, T *&data);
		/** Return a pointer to the attribute if it exists in @em record.
			Otherwise return NULL */
		T			*queryAttribute(const Record& r);
		/** Return a pointer to the attribute if it exists in @em record.
			Otherwise return NULL */
		T			*queryAttribute(const WeakRecord r);
		/** Return a pointer to the attribute if it exists in @em record.
			Otherwise return NULL */
		T			*queryAttribute(Record *pR);
		/** Return a pointer to the attribute if it exists in @em record
			indexed into the RecordArray.
			Otherwise return NULL */
		T			*queryAttribute(sp<RecordArray>& rspRA, UWORD index);
		void		registerDefault(T *pT);

	private:
		T			*m_pT;
};

template <>
class FE_DL_EXPORT Accessor<void> : public BaseAccessor
{
	public:
		Accessor(void)
		{
			m_typeInfo = TypeInfo(typeid(void));
		}
		Accessor(sp<Scope> scope, const String &attribute)
		{
			m_typeInfo = TypeInfo(typeid(void));
			setup(scope, attribute);
		}
		~Accessor(void)
		{
		}
};


/*****************************************************************************
IMPLEMENTATION
*****************************************************************************/
template <class T>
inline T &accessAttributeByOffset(Offset o, const WeakRecord r)
{
	return *((T *)((char *)(r.data()) + o));
}

inline bool checkAttribute(UWORD i, const Record& r)
{
	return (r.layout()->offsetTable()[i] != offsetNone);
}

inline bool checkAttribute(UWORD i, const WeakRecord r)
{
	return (r.layout()->offsetTable()[i] != offsetNone);
}

inline bool checkAttribute(UWORD i, Layout &rL)
{
#ifdef FE_ACCESSOR_DEBUG
	if(!rL.locked())
	{
		feX(e_notInitialized,
			"checkAttribute",
			"Layout not locked");
	}
#endif
	return (rL.offsetTable()[i] != offsetNone);
}

inline bool checkAttribute(UWORD i, sp<Layout>& rspL)
{
#ifdef FE_ACCESSOR_DEBUG
	if(!rspL->locked())
	{
		feX(e_notInitialized,
			"checkAttribute",
			"Layout not locked");
	}
#endif
	return (rspL->offsetTable()[i] != offsetNone);
}

inline bool checkAttribute(String n, const WeakRecord r)
{
	register UWORD index;
	if(!(r.layout()->scope()->findAttribute(n,index).isValid()))
	{
		return false;
	}
	return checkAttribute(index, r);
}

inline bool checkAttribute(String n, Layout &rL)
{
	register UWORD index;
	if(!(rL.scope()->findAttribute(n,index).isValid()))
	{
		return false;
	}
	return checkAttribute(index, rL);
}

inline bool checkAttribute(String n, sp<Layout> spL)
{
	register UWORD index;
	if(!(spL->scope()->findAttribute(n,index).isValid()))
	{
		return false;
	}
	return checkAttribute(index, spL);
}

template <class T>
inline T &accessAttribute(UWORD i, const Record& r)
{
#ifdef FE_ACCESSOR_DEBUG
	if(!checkAttribute(i,r))
	{
		feX("fe::accessAttribute","attempt to access invalid attribute");
	}
#endif
#if FE_CODEGEN<=FE_DEBUG
	return *((T *)((char *)(r.data()) + r.layout()->offsetTable()[i]));
#else
	return *((T *)((char *)(r.data()) + r.rawLayout()->offsetTable()[i]));
#endif
}

template <class T>
inline T &accessAttribute(UWORD i, const WeakRecord r)
{
#ifdef FE_ACCESSOR_DEBUG
	if(!checkAttribute(i,r))
	{
		feX("fe::accessAttribute","attempt to access invalid attribute");
	}
#endif
#if FE_CODEGEN<=FE_DEBUG
	return *((T *)((char *)(r.data()) + r.layout()->offsetTable()[i]));
#else
	return *((T *)((char *)(r.data()) + r.rawLayout()->offsetTable()[i]));
#endif
}

template <class T>
inline T &accessAttribute(UWORD i, sp<Layout> spL, void *sb)
{
#ifdef FE_ACCESSOR_DEBUG
	if(!spL->locked())
	{
		feX(e_notInitialized,
			"checkAttribute",
			"Layout not locked");
	}
#endif
	return *((T *)((char *)(sb) + spL->offsetTable()[i]));
}

template <class T>
inline T &accessAttribute(UWORD i, Layout *pL, void *sb)
{
#ifdef FE_ACCESSOR_DEBUG
	if(!pL->locked())
	{
		feX(e_notInitialized,
			"checkAttribute",
			"Layout not locked");
	}
#endif
	return *((T *)((char *)(sb) + pL->offsetTable()[i]));
}

template <class T>
inline T &accessAttribute(String n, const WeakRecord r)
{
	register UWORD index;
	r.layout()->scope()->findAttribute(n,index);
	return *((T *)((char *)(r.data()) + r.layout()->offsetTable()[index]));
}

template <class T>
inline T &accessAttribute(String n, const Record& r)
{
	register UWORD index;
	r.layout()->scope()->findAttribute(n,index);
	return *((T *)((char *)(r.data()) + r.layout()->offsetTable()[index]));
}

inline bool BaseAccessor::operator==(const BaseAccessor &other) const
{
	if(m_index != other.m_index) { return false; }
	if(!(m_hpScope == other.m_hpScope)) { return false; }
	return true;
}

inline bool BaseAccessor::operator<(const BaseAccessor &other) const
{
	return (m_index < other.m_index);
}


inline BaseAccessor::operator UWORD() const
{
	return m_index;
}

inline UWORD BaseAccessor::index(void) const
{
	return m_index;
}

#ifdef FE_ATTRIBUTE_DEBUG
#define FE_SCOPE_CHECK(s)													\
		if(!m_hpScope.isValid())											\
		{																	\
			feX("FE_SCOPE_CHECK",											\
			"Accessor for \'%s\' is not initialized",						\
					s->attribute(m_index)->name().c_str());					\
		}																	\
		if(s != m_hpScope)													\
		{																	\
			feX("FE_SCOPE_CHECK",											\
			"Accessor for \'%s\' is not bound to the same scope as Record",	\
					m_hpScope->attribute(m_index)->name().c_str());			\
		}
#else
#define FE_SCOPE_CHECK(s) /* NOOP */
#endif

inline bool BaseAccessor::check(const Record& record) const
{
	FE_SCOPE_CHECK(record.layout()->scope());
	return checkAttribute(m_index, record);
}

inline bool BaseAccessor::check(const WeakRecord record) const
{
	FE_SCOPE_CHECK(record.layout()->scope());
	return checkAttribute(m_index, record);
}

inline bool BaseAccessor::check(Record *record) const
{
	FE_SCOPE_CHECK(record->layout()->scope());
	return checkAttribute(m_index, *record);
}

inline bool BaseAccessor::check(sp<RecordArray>& rspRA) const
{
	FE_SCOPE_CHECK(rspRA->layout()->scope());
	return checkAttribute(m_index, *(rspRA->layout()));
}

inline bool BaseAccessor::check(RecordArray &rRA) const

{
	FE_SCOPE_CHECK(rRA.layout()->scope());
	return checkAttribute(m_index, *(rRA.layout()));
}

inline bool BaseAccessor::check(sp<Layout>& rspL) const
{
	FE_SCOPE_CHECK(rspL->scope());
	return checkAttribute(m_index, rspL);
}

inline bool BaseAccessor::check(Layout &rL) const
{
	FE_SCOPE_CHECK(rL.scope());
	return checkAttribute(m_index, rL);
}

inline void BaseAccessor::check(const Record& record, const char *message) const
{
	FE_SCOPE_CHECK(record.layout()->scope());
	if(!checkAttribute(m_index, record))
	{
		feX(e_cannotFind,
			(message!=NULL)?message:"",
			"%s not found in layout %s",
			m_hpScope->attribute(m_index)->name().c_str(),
			record.layout()->name().c_str());
	}
}

inline void BaseAccessor::check(const WeakRecord record, const char *message) const
{
	FE_SCOPE_CHECK(record.layout()->scope());
	if(!checkAttribute(m_index, record))
	{
		feX(e_cannotFind,
			(message!=NULL)?message:"",
			"%s not found in layout %s",
			m_hpScope->attribute(m_index)->name().c_str(),
			record.layout()->name().c_str());
	}
}

inline void BaseAccessor::check(Record *record, const char *message) const
{
	FE_SCOPE_CHECK((*record).layout()->scope());
	if(!checkAttribute(m_index, (*record)))
	{
		feX(e_cannotFind,
			(message!=NULL)?message:"",
			"%s not found in layout %s",
			m_hpScope->attribute(m_index)->name().c_str(),
			(*record).layout()->name().c_str());
	}
}

inline void BaseAccessor::check(sp<RecordArray>& rspRA,const char *message) const
{
	FE_SCOPE_CHECK(rspRA->layout()->scope());
	if(!checkAttribute(m_index, (rspRA->layout())))
	{
		feX(e_cannotFind,
			(message!=NULL)?message:"",
			"%s not found in layout %s",
			m_hpScope->attribute(m_index)->name().c_str(),
			rspRA->layout()->name().c_str());
	}
}

inline void BaseAccessor::check(RecordArray &rRA, const char *message) const
{
	FE_SCOPE_CHECK(rRA.layout()->scope());
	if(!checkAttribute(m_index, (rRA.layout())))
	{
		feX(e_cannotFind,
			(message!=NULL)?message:"",
			"%s not found in layout %s",
			m_hpScope->attribute(m_index)->name().c_str(),
			rRA.layout()->name().c_str());
	}
}

inline void BaseAccessor::check(sp<Layout>& rspL, const char *message) const
{
	FE_SCOPE_CHECK(rspL->scope());
	if(!checkAttribute(m_index, rspL))
	{
		feX(e_cannotFind,
			(message!=NULL)?message:"",
			"%s not found in layout %s",
			m_hpScope->attribute(m_index)->name().c_str(),
			rspL->name().c_str());
	}
}

inline void BaseAccessor::check(Layout &rL, const char *message) const
{
	FE_SCOPE_CHECK(rL.scope());
	if(!checkAttribute(m_index, rL))
	{
		feX(e_cannotFind,
			(message!=NULL)?message:"",
			"%s not found in layout %s",
			m_hpScope->attribute(m_index)->name().c_str(),
			rL.name().c_str());
	}
}


template <class T>
inline Accessor<T>::Accessor(void)
{
	m_pT = NULL;
	m_typeInfo = TypeInfo(typeid(T));
}

template <class T>
inline Accessor<T>::Accessor(sp<Scope> scope, const String &attribute)
{
	m_pT = NULL;
	m_typeInfo = TypeInfo(typeid(T));
	setup(scope, attribute);
}

template <class T>
inline Accessor<T>::~Accessor(void)
{
}

#ifdef FE_ATTRIBUTE_DEBUG
#define FE_ATTRIBUTE_CHECK(r)												\
		if(!check(r))														\
		{																	\
			feX("FE_ATTRIBUTE_CHECK",										\
					"invalid access: \'%s\' is not in \'%s\'",				\
					m_hpScope->attribute(m_index)->name().c_str(),			\
					r.layout()->name().c_str());							\
		}
#else
#define FE_ATTRIBUTE_CHECK(r) /* NOOP */
#endif

#ifdef FE_ATTRIBUTE_DEBUG
#define FE_ATTRIBUTE_TYPECHECK(r)											\
	sp<TypeMaster> spTM = m_hpScope->typeMaster();							\
	sp<BaseType> spType = spTM->lookupType(TypeInfo(typeid(T)));			\
	if(spType != m_hpScope->attribute(m_index)->type())						\
	{																		\
		feX("FE_ATTRIBUTE_TYPECHECK",										\
			"accessor on \'%s\' type mismatch \'%s\' is not \'%s\'\n",		\
			m_hpScope->attribute(m_index)->name().c_str(),					\
			spTM->reverseLookup(spType).c_str(),							\
			spTM->reverseLookup(m_hpScope->attribute(m_index)->type()).		\
			c_str());														\
	}
#else
#define FE_ATTRIBUTE_TYPECHECK(r) /* NOOP */
#endif

template <class T>
inline Accessor<T> &Accessor<T>::operator=(const Accessor<T> &other)
{
	if(&other != this)
	{
		m_index = other.m_index;
		m_hpScope = other.m_hpScope;
		m_typeInfo = other.m_typeInfo;
		m_pT = other.m_pT;
	}
	return *this;
}

template <class T>
inline T &Accessor<T>::operator()(const Record& record)
{
	FE_ATTRIBUTE_CHECK(record);
	FE_ATTRIBUTE_TYPECHECK(record);

	return accessAttribute<T>(m_index, record);
}

template <class T>
inline const T Accessor<T>::operator()(const Record& record) const
{
	FE_ATTRIBUTE_CHECK(record);
	FE_ATTRIBUTE_TYPECHECK(record);
	return accessAttribute<T>(m_index, record);
}

template <class T>
inline T &Accessor<T>::operator()(const WeakRecord record)
{
	FE_ATTRIBUTE_CHECK(record);
	FE_ATTRIBUTE_TYPECHECK(record);
	return accessAttribute<T>(m_index, record);
}

template <class T>
inline const T Accessor<T>::operator()(const WeakRecord record) const
{
	FE_ATTRIBUTE_CHECK(record);
	FE_ATTRIBUTE_TYPECHECK(record);
	return accessAttribute<T>(m_index, record);
}

template <class T>
inline T &Accessor<T>::operator()(Record *record)
{
	FE_ATTRIBUTE_CHECK((*record));
	FE_ATTRIBUTE_TYPECHECK(*record);
	return accessAttribute<T>(m_index, *record);
}

#if 0
template <class T>
inline T &Accessor<T>::operator()(Record *record, T &data)
{
	if(!((*record).isValid()))
	{
		return data;
	}
	if(!check(*record))
	{
		return data;
	}
	return accessAttribute<T>(m_index, *record);
}

template <class T>
inline T &Accessor<T>::operator()(const Record record, T &data)
{
	if(!record.isValid())
	{
		return data;
	}
	if(!check(record))
	{
		return data;
	}
	return accessAttribute<T>(m_index, record);
}

template <class T>
inline void Accessor<T>::get(T &a_data, const Record a_record)
{
	if(!a_record.isValid())
	{
		throw BaseAccessor::Exception(attribute()->name().c_str());
	}
	if(!check(a_record))
	{
		throw BaseAccessor::Exception(attribute()->name().c_str());
	}
	a_data = accessAttribute<T>(m_index, a_record);
}
#endif

template <class T>
inline T &Accessor<T>::operator()(const Record& record, const char *message)
{
	if(!record.isValid())
	{
		feX(e_invalidPointer,
			(message!=NULL)?message:"",
			"invalid record in access of %s",
			m_hpScope->attribute(m_index)->name().c_str());
	}
	if(!check(record))
	{
		feX(e_cannotFind,
			(message!=NULL)?message:"",
			"%s not found in layout %s",
			m_hpScope->attribute(m_index)->name().c_str(),
			record.layout()->name().c_str());
	}
	return accessAttribute<T>(m_index, record);
}

template <class T>
inline const T Accessor<T>::operator()(const Record& record, const char *message) const
{
	if(!record.isValid())
	{
		feX(e_invalidPointer,
			(message!=NULL)?message:"",
			"invalid record in access of %s",
			m_hpScope->attribute(m_index)->name().c_str());
	}
	if(!check(record))
	{
		feX(e_cannotFind,
			(message!=NULL)?message:"",
			"%s not found in layout %s",
			m_hpScope->attribute(m_index)->name().c_str(),
			record.layout()->name().c_str());
	}
	return accessAttribute<T>(m_index, record);
}

template <class T>
inline T &Accessor<T>::operator()(const WeakRecord record, const char *message)
{
	if(!record.isValid())
	{
		feX(e_invalidPointer,
			(message!=NULL)?message:"",
			"invalid record in access of %s",
			m_hpScope->attribute(m_index)->name().c_str());
	}
	if(!check(record))
	{
		feX(e_cannotFind,
			(message!=NULL)?message:"",
			"%s not found in layout %s",
			m_hpScope->attribute(m_index)->name().c_str(),
			record.layout()->name().c_str());
	}
	return accessAttribute<T>(m_index, record);
}

template <class T>
inline const T Accessor<T>::operator()(const WeakRecord record, const char *message) const
{
	if(!record.isValid())
	{
		feX(e_invalidPointer,
			(message!=NULL)?message:"",
			"invalid record in access of %s",
			m_hpScope->attribute(m_index)->name().c_str());
	}
	if(!check(record))
	{
		feX(e_cannotFind,
			(message!=NULL)?message:"",
			"%s not found in layout %s",
			m_hpScope->attribute(m_index)->name().c_str(),
			record.layout()->name().c_str());
	}
	return accessAttribute<T>(m_index, record);
}

template <class T>
inline T &Accessor<T>::operator()(Record *record, const char *message)
{
	if(!((*record).isValid()))
	{
		feX(e_invalidPointer,
			(message!=NULL)?message:"",
			"invalid record in access of %s",
			m_hpScope->attribute(m_index)->name().c_str());
	}
	if(!check(*record))
	{
		feX(e_cannotFind,
			(message!=NULL)?message:"",
			"%s not found in layout %s",
			m_hpScope->attribute(m_index)->name().c_str(),
			record->layout()->name().c_str());
	}
	return accessAttribute<T>(m_index, *record);
}


template <class T>
inline T &Accessor<T>::get(Record *record, const char *message)
{
	if(!((*record).isValid()))
	{
		feX(e_invalidPointer,
			(message!=NULL)?message:"",
			"invalid record in access of %s",
			m_hpScope->attribute(m_index)->name().c_str());
	}
	if(!check(*record))
	{
		feX(e_cannotFind,
			(message!=NULL)?message:"",
			"%s not found in layout %s",
			m_hpScope->attribute(m_index)->name().c_str(),
			record->layout()->name().c_str());
	}
	return accessAttribute<T>(m_index, *record);
}

template <class T>
inline T &Accessor<T>::get(const Record record, const char *message)
{
	if(!record.isValid())
	{
		feX(e_invalidPointer,
			(message!=NULL)?message:"",
			"invalid record in access of %s",
			m_hpScope->attribute(m_index)->name().c_str());
	}
	if(!check(record))
	{
		feX(e_cannotFind,
			(message!=NULL)?message:"",
			"%s not found in layout %s",
			m_hpScope->attribute(m_index)->name().c_str(),
			record.layout()->name().c_str());
	}
	return accessAttribute<T>(m_index, record);
}


template <class T>
inline const T &Accessor<T>::operator()(sp<RecordArray>& rspRA, UWORD index) const
{
	FE_ATTRIBUTE_CHECK((*rspRA));
#if FE_CODEGEN<=FE_DEBUG
	return accessAttribute<T>(m_index, rspRA->layout(), rspRA->data(index));
#else
	return accessAttribute<T>(m_index, rspRA->rawLayout(), rspRA->data(index));
#endif
}

template <class T>
inline T &Accessor<T>::operator()(sp<RecordArray>& rspRA, UWORD index)
{
	FE_ATTRIBUTE_CHECK((*rspRA));
#if FE_CODEGEN<=FE_DEBUG
	return accessAttribute<T>(m_index, rspRA->layout(), rspRA->data(index));
#else
	return accessAttribute<T>(m_index, rspRA->rawLayout(), rspRA->data(index));
#endif
}

template <class T>
inline T &Accessor<T>::operator()(hp<RecordArray>& rhpRA, UWORD index)
{
	FE_ATTRIBUTE_CHECK((*rhpRA));
#if FE_CODEGEN<=FE_DEBUG
	return accessAttribute<T>(m_index, rhpRA->layout(), rhpRA->data(index));
#else
	return accessAttribute<T>(m_index, rhpRA->rawLayout(), rhpRA->data(index));
#endif
}

template <class T>
inline bool Accessor<T>::queryAttribute(const Record& record, T &data)
{
	if(!check(record))
	{
		return false;
	}
	data = operator()(record);
	return true;
}

template <class T>
inline bool Accessor<T>::queryAttribute(const WeakRecord record, T &data)
{
	if(!check(record))
	{
		return false;
	}
	data = operator()(record);
	return true;
}

template <class T>
inline bool Accessor<T>::queryAttribute(hp<Layout>& rhpLayout, void *sb,
		T *&data)
{
	if(!checkAttribute(m_index, *rhpLayout.raw()))
	{
		return false;
	}
	data = &(accessAttribute<T>(m_index, rhpLayout.raw(), sb));
	return true;
}

template <class T>
inline bool Accessor<T>::queryAttribute(sp<Layout>& rspLayout, void *sb,
		T *&data)
{
	if(!checkAttribute(m_index, rspLayout))
	{
		return false;
	}
	data = &(accessAttribute<T>(m_index, rspLayout, sb));
	return true;
}

template <class T>
inline T *Accessor<T>::queryAttribute(const Record& record)
{
	if(!check(record))
	{
		return m_pT;
	}
	return &(operator()(record));
}

template <class T>
inline T *Accessor<T>::queryAttribute(const WeakRecord record)
{
	if(!check(record))
	{
		return m_pT;
	}
	return &(operator()(record));
}

template <class T>
inline T *Accessor<T>::queryAttribute(sp<RecordArray>& rspRA, UWORD index)
{
	if(!check(rspRA))
	{
		return m_pT;
	}
	return &(operator()(rspRA,index));
}

template <class T>
inline T *Accessor<T>::queryAttribute(Record *record)
{
	if(!check(record))
	{
		return m_pT;
	}
	return &(operator()(record));
}

template <class T>
inline void Accessor<T>::registerDefault(T *pT)
{
	m_pT = pT;
}

} /* namespace */

#endif /* __data_Accessor_h__ */

