/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "data.pmh"

namespace fe
{

UWORD Layout::ms_unique = 1;

Layout::Layout(void)
{
	feX("fe::Layout::Layout","do not use default Layout constructor");
}

Layout::Layout(sp<Scope> &scope)
{
	m_name.sPrintf("__%d__", ms_unique++);
	constructor(scope);
}

Layout::Layout(sp<Scope> &scope, const String &name)
{
	this->setName(name);
	constructor(scope);
}

Layout::Layout(Scope *pScope, const String &name)
{
	this->setName(name);
	sp<Scope> spScope(pScope);
	constructor(spScope);
}

Layout::Layout(const Layout &other)
{
	copy(other);
}

void Layout::constructor(sp<Scope> &scope)
{
	FEASSERT(scope.isValid());
	m_suppressLock = false;
	m_init = false;
	m_construct = false;
	m_pTable = NULL;
	m_hpScope = scope;
	m_size = 0;
}

Layout::~Layout(void)
{
#if 0
	if(rawOffsetTable())
	{
		delete [] rawOffsetTable();
	}
#endif
}

Layout &Layout::operator=(const Layout &other)
{
	SAFEGUARD;
	return copy(other);
}

Layout &Layout::copy(const Layout &other)
{
	SAFEGUARD;
	if(&other == this) { return *this; }

	m_name = other.name();
	m_init = other.locked();
	m_size = other.size();
	if(m_pTable)
	{
		delete [] m_pTable;
	}
	register UWORD tablesize = sizeof(Offset) * other.attributeCount();
	m_pTable = new Offset[other.attributeCount()];
	memcpy((void *)offsetTable(), (void *)(other.offsetTable()), tablesize);

	return *this;
}

UWORD Layout::attributeCount(void) const
{
	SAFEGUARD;
	return m_hpScope->getAttributeCount();
}

void Layout::populate(const String &attribute_name)
{
	SAFEGUARD;
	m_hpScope->populate(name(), attribute_name);
}

void Layout::populate(const String &attribute_name, const String &attribute_type)
{
	SAFEGUARD;
	m_hpScope->support(attribute_name, attribute_type);
	m_hpScope->populate(name(), attribute_name);
}

void Layout::populate(const BaseAccessor &accessor)
{
	SAFEGUARD;
	if(locked())
	{
		if(!accessor.check(*this))
		{
			sp<Attribute> spAttribute = m_hpScope->attribute((UWORD)accessor);
			feX("Layout::populate",
					"locked layout \'%s\' must have \'%s\' attribute",
					name().c_str(),
					spAttribute->name().c_str());
		}
	}
	else
	{
		sp<Attribute> spAttribute = m_hpScope->attribute((UWORD)accessor);
		m_hpScope->populate(name(), spAttribute->name());
	}
}

void Layout::populate(sp<Attribute> spAttribute)
{
	SAFEGUARD;
	m_hpScope->populate(name(), spAttribute->name());
}

void Layout::populate(sp<Layout> spLayout)
{
	SAFEGUARD;
	m_hpScope->clonePopulate(spLayout->name(), name());
}


void Layout::lock(void)
{
	SAFEGUARD;
	sp<Layout> spLayout(this);
	m_hpScope->lockLayout(spLayout);
}

void Layout::suppressLock(bool suppression)
{
	SAFEGUARD;
	if(locked())
	{
		feX("Layout::suppressLock",
			"attempt to suppress locking on already locked layout \'%s\'",
			name().c_str());
	}
	m_suppressLock = suppression;
}

bool Layout::lockSuppressed(void) const
{
	SAFEGUARD;
	return m_suppressLock;
}

void Layout::initialize(void)
{
	SAFEGUARD;
	if(!scope().isValid())
	{
		feX("fe::Layout::initialize",
			"cannot initialize record layout without scope");
	}
	sp<Layout> rLayout(this);
	scope()->lockLayout(rLayout);
}

void Layout::cacheIndices(void)
{
	SAFEGUARD;
	m_offsetIndex.clear();
	UWORD cnt = scope()->getAttributeCount();
	for(UWORD i = 0; i < cnt ; i++)
	{
		if(m_pTable[i] == offsetNone) { continue; }

		m_offsetIndex.push_back(i);
	}
}

void Layout::constructAttributes(void *stateBlock)
{
	SAFEGUARD;

	if(m_construct)
	{
#if 0
		UWORD cnt = scope()->getAttributeCount();
		for(UWORD i = 0; i < cnt ; i++)
		{
			if(m_pTable[i] == offsetNone) { continue; }

			scope()->attribute(i)->type()->construct(
				(void *)(((char *)stateBlock) + m_pTable[i]));
		}
#endif

		for(UWORD i = 0; i < m_offsetIndex.size(); i++)
		{
			scope()->attribute(m_offsetIndex[i])->type()->construct(
				(void *)(((char *)stateBlock) + m_pTable[m_offsetIndex[i]]));
		}
	}

	for(UWORD i = 0; i < m_offsetIndex.size(); i++)
	{
		Instance &defaultInstance =
			scope()->attribute(m_offsetIndex[i])->defaultInstance();
		if(defaultInstance.data())
		{
			Instance new_instance = Instance(
				(void *)(((char *)stateBlock) + m_pTable[m_offsetIndex[i]]),
				defaultInstance.type(), NULL);

			new_instance.assign(defaultInstance);
		}
	}

}

void Layout::destructAttributes(void *stateBlock)
{
	SAFEGUARD;

	if(m_construct)
	{
#if 0
		UWORD cnt = scope()->getAttributeCount();
		for(UWORD i = 0; i < cnt ; i++)
		{
			if(m_pTable[i] == offsetNone) { continue; }

			scope()->attribute(i)->type()->destruct(
				(void *)(((char *)stateBlock) + m_pTable[i]));
		}
#endif

		for(UWORD i = 0; i < m_offsetIndex.size(); i++)
		{
			scope()->attribute(m_offsetIndex[i])->type()->destruct(
				(void *)(((char *)stateBlock) + m_pTable[m_offsetIndex[i]]));
		}
	}
}


void Layout::peek(Peeker &peeker)
{
	peeker.nl();
	peeker.bold();
	peeker.str().cat("%s:", m_name.c_str());
	peeker.normal();
	peeker.in();

	peeker.nl();
	peeker.str().cat("size: %u", size());

	if(!m_init)
	{
		peeker.nl();
		peeker.red();
		peeker.cat("not yet locked");
	}

	if(m_pTable)
	{
		peeker.nl();
		peeker.cat("offset table:");
		peeker.normal();
		peeker.in();
		peeker.cyan();
		for(UWORD i = 0; i < attributeCount(); i++)
		{
			if(m_pTable[i] == offsetNone)
			{
#if 0
				peeker.nl();
				peeker.cat("NONE");
				peeker.str().cat("\t");
				peeker.cat(m_hpScope->attribute(i)->name());
#endif
			}
			else
			{
				peeker.nl();
				peeker.str().cat("%4u", m_pTable[i]);
				peeker.str().cat("\t");
				peeker.cat(m_hpScope->attribute(i)->name());
			}
		}
		peeker.out();
	}

	peeker.base();

	peeker.out();
}

} /* namespace */


