/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#ifndef __data_Layout_h__
#define __data_Layout_h__

namespace fe
{

const Offset offsetNone=0xFFFFFFFF;

class Scope;
class Record;
class BaseAccessor;


/** @brief Record "type" definition

	@ingroup data

	A Layout is a definition of Attribute layout in Record instances.

	Layout objects can only be created via a Scope.  For typical use
	all operations on Layout objects are also done through Scope methods.

	@todo full data system guide
	*/
class FE_DL_EXPORT Layout: public Component, public Initialize<Layout>,
		public ClassSafe<GlobalHolder>
{
	friend class Scope;
	private:
		Layout(sp<Scope> &scope);
		Layout(sp<Scope> &scope, const String &name);
		Layout(Scope *pScope, const String &name);
		Layout(const Layout &other);
		Layout &operator=(const Layout &other);
	public:
virtual	~Layout(void);

		Layout			&copy(const Layout &other);
		void			initialize(void);

		/** Convienience function to Scope::populate */
		void			populate(const String &attribute_name);
		/** Convienience function to Scope::populate */
		void			populate(const String &attribute_name,
								const String &attribute_type);
		/** Convienience function to Scope::populate.  This particular one
			does a check if the layout is locked and if so verifys the
			exsitence of the attribute. */
		void			populate(const BaseAccessor &accessor);
		/** Convienience function to Scope::populate */
		void			populate(sp<Attribute> spAttribute);
		/** Convienience function to Scope::populate */
		void			populate(sp<Layout> spLayout);

		void			setName(const String &name);
		const String&	name(void) const;
		const String	verboseName(void) const;

		void			setLocked(bool locked);
		const bool		&locked(void) const;
		/** Convienience function to Scope::lockLayout */
		void			lock(void);
		/**	Set lock suppression.  If lock suppression is on (true) then an
			exception is thrown if locking is attempted. */
		void			suppressLock(bool suppression);
		bool			lockSuppressed(void) const;

		void			setConstruct(bool construct);
		const bool		&construct(void) const;

		/// Number of attributes in this layout
		UWORD			attributeCount(void) const;

		void			setSize(const Offset record_size);
		/// Size of an instance of the layout
		Offset			size(void) const;

		void			setOffsetTable(Offset *offset_table);
		const Offset	*offsetTable(void) const;
		const Offset	*rawOffsetTable(void) const;

		hp<Scope>&		scope(void);

		void			constructAttributes(void *stateBlock);
		void			destructAttributes(void *stateBlock);

		void			peek(Peeker &peeker);

	private:
						Layout(void);
		void			constructor(sp<Scope> &scope);

		void			cacheIndices(void);
		void			setOffset(const UWORD index, const Offset offset);

	private:
static	UWORD				ms_unique;
		String				m_name;
		bool				m_init;
		bool				m_construct;
		bool				m_suppressLock;
		Offset				*m_pTable;
		Offset				m_size;
		hp<Scope>			m_hpScope;
		std::vector<UWORD>	m_offsetIndex;

};

inline void Layout::setName(const String &name)
{
	SAFEGUARD;
	m_name = name;
}
inline const String& Layout::name(void) const
{
	SAFEGUARD;
	return m_name;
}
inline const String Layout::verboseName(void) const
{
	return "Layout " + name();
}

inline void Layout::setLocked(bool locked)
{
	SAFEGUARD;
	m_init = locked;
}
inline const bool &Layout::locked(void) const
{
	SAFEGUARD;
	return m_init;
}

inline void Layout::setConstruct(bool construct)
{
	SAFEGUARD;
	m_construct = construct;
}
inline const bool &Layout::construct(void) const
{
	SAFEGUARD;
	return m_construct;
}

#if 1
inline void Layout::setSize(const Offset size)
{
	SAFEGUARD;
	m_size = size;
}
inline Offset Layout::size(void) const
{
	SAFEGUARD;
	return m_size;
}
#endif

inline void Layout::setOffsetTable(Offset *offset_table)
{
	SAFEGUARD;
	m_pTable = offset_table;
}

inline const Offset *Layout::offsetTable(void) const
{
	//SAFEGUARD; // too slow /// @todo remove
#if FE_CODEGEN<=FE_DEBUG
	if(!locked())
	{
		feX("Layout::offsetTable",
			"layout not locked");
	}
#endif
	return m_pTable;
}

inline const Offset *Layout::rawOffsetTable(void) const
{
	//SAFEGUARD; // too slow /// @todo remove
	return m_pTable;
}

#if 1
inline void Layout::setOffset(const UWORD index, const Offset offset)
{
	m_pTable[index] = offset;
}
#endif

inline hp<Scope>& Layout::scope(void)
{
	SAFEGUARD;
	return m_hpScope;
}

} /* namespace */

#endif /* __data_Layout_h__ */

