/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "data.pmh"

namespace fe
{

Record Record::clone(void)
{
	Record r_clone;
	if(!isValid()) { return r_clone; }

	sp<Scope> spScope = m_spLayout->scope();
//	r_clone = spScope->createRecord(m_spLayout);
	r_clone = spScope->produceRecord(m_spLayout->name());

	UWORD cnt = spScope->getAttributeCount();
	for(UWORD i = 0; i < cnt; i++)
	{
		Offset offset = m_spLayout->offsetTable()[i];
		if(offset != offsetNone)
		{
			sp<Attribute> spAttribute=spScope->attribute(i);
			if(spAttribute->isCloneable())
			{
				sp<BaseType> spBT = spAttribute->type();
				spBT->assign(
					(void *)((char *)(r_clone.data()) + offset),
					(void *)((char *)(data()) + offset));
			}
		}
	}

	return r_clone;
}

typedef std::map< sp<RecordGroup>, sp<RecordGroup> >	t_srcrg_dstrg;
typedef std::map< Record, Record >						t_srcr_dstr;

class RecCloner
{
	public:
		RecCloner(sp<TypeMaster> a_typemaster)
		{
			m_spRGType = a_typemaster->lookupType< sp<RecordGroup> >();
			m_spRecordType = a_typemaster->lookupType<Record>();
		}
		Record			clone(Record &r_src);
		sp<RecordGroup> clone(sp<RecordGroup> rg_src);
	private:
		t_srcrg_dstrg		m_rgs;
		t_srcr_dstr			m_rs;
		sp<BaseType>		m_spRGType;
		sp<BaseType>		m_spRecordType;
};

Record RecCloner::clone(Record &r_src)
{
	Record r_clone;
	if(!r_src.isValid()) { return r_clone; }

	sp<Scope> spScope = r_src.layout()->scope();

	t_srcr_dstr::iterator i_r = m_rs.find(r_src);
	if(i_r != m_rs.end())
	{
		r_clone = i_r->second;
		return r_clone;
	}
	else
	{
		r_clone = spScope->createRecord(r_src.layout());
		m_rs[r_src] = r_clone;
	}

	UWORD cnt = spScope->getAttributeCount();
	for(UWORD i = 0; i < cnt; i++)
	{
		Offset offset = r_clone.layout()->offsetTable()[i];
		if(offset != offsetNone)
		{
			sp<Attribute> spAttribute=spScope->attribute(i);
			sp<BaseType> spBT = spAttribute->type();
			if(spBT == m_spRGType)
			{
				accessAttribute< sp<RecordGroup> >(i, r_clone) =
					clone(accessAttribute< sp<RecordGroup> >(i, r_src));
			}
			else if(spBT == m_spRecordType)
			{
				accessAttribute<Record>(i, r_clone) =
					clone(accessAttribute<Record>(i, r_src));
			}
			else if(spAttribute->isCloneable())
			{
				spBT->assign(
					(void *)((char *)(r_clone.data()) + offset),
					(void *)((char *)(r_src.data()) + offset));
			}
		}
	}
	return r_clone;
}

sp<RecordGroup> RecCloner::clone(sp<RecordGroup> rg_src)
{
	sp<RecordGroup> rg_clone;
	if(!rg_src.isValid()) { return rg_clone; }

	t_srcrg_dstrg::iterator i_rg = m_rgs.find(rg_src);
	if(i_rg != m_rgs.end())
	{
		rg_clone = i_rg->second;
		return rg_clone;
	}
	else
	{
		rg_clone = new RecordGroup();
		m_rgs[rg_src] = rg_clone;
	}

	for(RecordGroup::iterator it = rg_src->begin();
		it != rg_src->end(); it++)
	{
		sp<RecordArray> spRA(*it);
		for(int i = 0; i < spRA->length(); i++)
		{
			Record r_src = spRA->getRecord(i);
			rg_clone->add(clone(r_src));
		}
	}
	return rg_clone;
}

Record Record::deepclone(void)
{
	Record r_clone;
	if(!isValid()) { return r_clone; }

	RecCloner cloner(m_spLayout->scope()->typeMaster());

	r_clone = cloner.clone(*this);

	return r_clone;
}

bool Record::extractInstance(Instance &instance, const String &attrName)
{
	UWORD attrIndex;
	FEASSERT(m_spLayout.isValid());
	const sp<Attribute>& rspAttribute =
		m_spLayout->scope()->findAttribute(attrName, attrIndex);
	if(!rspAttribute.isValid())
	{
		return false;
	}

	if(m_spLayout->offsetTable()[attrIndex] == offsetNone)
	{
		return false;
	}

	const sp<BaseType>& rspBT = rspAttribute->type();

	void *pV =
		(void *)((char *)(data())+m_spLayout->offsetTable()[attrIndex]);

	int *pRefCnt = NULL;
	if(m_spLayout->scope()->refCount().check(*this))
	{
		pRefCnt = &(m_spLayout->scope()->refCount()(*this));
	}

	instance.set(pV, rspBT, pRefCnt);

	return true;
}

String RecordInfo::print(void *instance)
{
	if(!instance)
		return String("<error>");

	const Record* pR = (Record*)instance;
	if(!pR->isValid())
		return String("<invalid>");

	String string;
	string.sPrintf("%p %s",pR->data(),pR->layout().isValid()?
			pR->layout()->name().c_str(): "<invalid layout>");
	return string;
}

IWORD RecordInfo::output(std::ostream &ostrm, void *instance, t_serialMode mode)
{
	feX(e_unsupported,
		"RecordInfo::output",
		"record output should be done via fe::Stream");
	return 0;
}

void RecordInfo::input(std::istream &istrm, void *instance, t_serialMode mode)
{
	feX(e_unsupported,
		"RecordInfo::input",
		"record input should be done via fe::Stream");
}

IWORD RecordInfo::iosize(void)
{
	return c_implicit;
}

bool RecordInfo::getConstruct(void)
{
	return true;
}

void RecordInfo::construct(void *instance)
{
	new(instance)Record;
}

void RecordInfo::destruct(void *instance)
{
	((Record *)instance)->~Record();
}

} /* namespace */

