/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#ifndef __data_Record_h__
#define __data_Record_h__

namespace fe
{

#define FE_RECORD_COUNT_DEBUG	FALSE

template <class T>
class Accessor;

/**	@brief Reference to an instance of a Layout

	@ingroup data

	A record is simply a pointer to a Layout and a pointer to a
	state block.  A state block is a block of memory with Attributes laid
	out as specified by a Layout object.
	*/
class FE_DL_EXPORT Record
{
	friend class Scope;
	public:
		Record(I32 ignored=0);
//		Record(void *datablock);
		Record(const Record &other);
virtual	~Record(void);

		Record	&operator=(const Record &r_other);

		bool	operator==(const Record &r_other) const;
		bool	operator!=(const Record &r_other) const;

		/** Return the state block. */
		void		*data(void) const;
		/** Return true if the Record points to a valid state block. */
		bool		isValid(void) const;
		/** Return the Layout. */
		sp<Layout>	layout(void) const;
		/** Set the state block. */
		void		set(void *datablock);

		/** Return a raw pointer to the Layout.  Mainly intended for Accessor
			for speed. */
		Layout		*rawLayout(void) const;

		bool		extractInstance(Instance &instance, const String &attrName);

		Record		clone(void);
		Record		deepclone(void);

	private:
		void		acquire(void);
		void		release(void);

	private:
		sp<Layout>			m_spLayout;
		void				*m_pStateBlock;
};

inline bool operator<(const Record &lhs, const Record &rhs)
{
	return (lhs.data() < rhs.data());
}

class RecordInfo : public BaseType::Info
{
	public:
virtual	String	print(void *instance);
virtual	IWORD	output(std::ostream &ostrm, void *instance, t_serialMode mode);
virtual	void	input(std::istream &istrm, void *instance, t_serialMode mode);
virtual	IWORD	iosize(void);
virtual	bool	getConstruct(void);
virtual	void	construct(void *instance);
virtual	void	destruct(void *instance);
};

inline Record::Record(I32)
{
	m_pStateBlock = NULL;
}

inline Record::Record(const Record &other)
{
	m_spLayout = other.m_spLayout;
	m_pStateBlock = other.m_pStateBlock;
	acquire();
}

/*
inline Record::Record(void *datablock)
{
	m_pStateBlock = datablock;
	acquire();
	m_spLayout = FE_SB_TO_HDR(m_pStateBlock)->m_pStore->getLayout();
}
*/

inline Record::~Record(void)
{
	release();
}

inline Layout *Record::rawLayout(void) const
{
	return m_spLayout.raw();
}

inline Record &Record::operator=(const Record &other)
{
	if(this != &other)
	{
		release();
		m_spLayout = other.m_spLayout;
		m_pStateBlock = other.m_pStateBlock;
		acquire();
	}
	return *this;
}

inline bool Record::operator==(const Record &other) const
{
	return (m_pStateBlock == other.m_pStateBlock);
}

inline bool Record::operator!=(const Record &other) const
{
	return (m_pStateBlock != other.m_pStateBlock);
}

inline void *Record::data(void) const
{
	return m_pStateBlock;
}

inline bool Record::isValid(void) const
{
	return (m_pStateBlock != NULL);
}

inline sp<Layout> Record::layout(void) const
{
	return m_spLayout;
}

inline void Record::set(void *datablock)
{
	release();
	m_pStateBlock = datablock;
	acquire();
	m_spLayout = FE_SB_TO_HDR(m_pStateBlock)->m_pStore->getLayout();
}

inline void Record::release(void)
{
#if FE_RECORD_COUNT_DEBUG
	register UWORD index;
	if(m_spLayout.isValid())
	{
		m_spLayout->scope()->findAttribute(FE_C,index);
		if(m_spLayout->offsetTable()[index] != offsetNone)
		{
			IWORD &count= *((IWORD *)((char *)(data()) +
											layout()->offsetTable()[index]));

			fprintf(stderr,"Release %p %d->%d %s\n",(U32)this,count,count-1,
					m_spLayout->name().c_str());
		}
	}
#endif

	if(m_pStateBlock)
	{
#if FE_COUNTED_TRACK
		Counted::untrackReference(FE_SB_TO_HDR(m_pStateBlock),&m_pStateBlock);
//		Counted::untrackReference(this,this);
//		Counted::deregisterRegion(this);
#endif
		FE_SB_TO_HDR(m_pStateBlock)->m_pStore->releaseSB(m_pStateBlock);
	}
}

inline void Record::acquire(void)
{
#if FE_RECORD_COUNT_DEBUG
	register UWORD index;
	if(m_spLayout.isValid())
	{
		m_spLayout->scope()->findAttribute(FE_C,index);
		if(m_spLayout->offsetTable()[index] != offsetNone)
		{
			IWORD &count= *((IWORD *)((char *)(data()) +
										m_spLayout->offsetTable()[index]));

			fprintf(stderr,"Acquire %p %d->%d %s\n",(U32)this,count,count+1,
					m_spLayout->name().c_str());
		}
	}
#endif

	if(m_pStateBlock)
	{
		FE_SB_TO_HDR(m_pStateBlock)->m_pStore->acquireSB(m_pStateBlock);
#if FE_COUNTED_TRACK
//		Counted::registerRegion(this,sizeof(Record),"Record "+
//				(m_spLayout.isValid()? m_spLayout->name(): "<invalid>"));
//		Counted::trackReference(this,this);
		Counted::trackReference(FE_SB_TO_HDR(m_pStateBlock),&m_pStateBlock);
#endif
	}
}

} /* namespace */

#endif /* __data_Record_h__ */
