/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#ifndef __data_RecordArray_h__
#define __data_RecordArray_h__

// use a hash map to speedup lookups
//#define FE_RA_INDEX

namespace fe
{

struct hash_sb_pointer
{ UWORD operator()(const void *pV) const
	{ return reinterpret_cast<UWORD>(pV); } };

struct eq_sb_pointer
{ bool operator()(void *pV1, void *pV2) const
	{ return pV1 == pV2; } };

typedef ext::hash_map<	void *,
						IWORD,
						hash_sb_pointer,
						eq_sb_pointer>		t_sb_map;

/**	@brief Homogeneous collection of Records

	@ingroup data

	An array of Records, or more precisely and array of state block pointers
	all pointing to state blocks of the same layout.  Therefore only a single
	reference to the layout itself is stored.  Therefore this record array
	object consumes less memory than an array of record objects.
	*/
class FE_DL_EXPORT RecordArray : public Counted
{
	public:
		RecordArray(void);
		RecordArray(sp<Layout> spLayout);
		RecordArray(const RecordArray &other);
virtual	~RecordArray(void);

		RecordArray		&operator=(const RecordArray &other);
		/**	Return the state block at the given index. */
		void			*operator[](IWORD index);
		/**	Return the state block at the given index. */
		void			*data(IWORD index);
		/**	Return the length of the array. */
		IWORD			length(void);
		/** Return the Layout. */
		sp<Layout>		layout(void);
		/** Return a raw pointer to the Layout.  Mainly intended for Accessor
			for speed. */
		Layout			*rawLayout(void) const;
		/**	Add the given record to the array.  If the record array already
			has records and the Layout is different than for @em record
			the record will not be added and false will be returned.
			On success true is returned and if @em index is not NULL the index
			of the added record is return in @em index . */
		bool			add(const Record &record, IWORD *index = NULL);
		bool			add(sp<RecordArray> spRA, IWORD start, IWORD size);
		/** Remove the given record from the array.  On success return
			true and if @em index is provided return the index of the removed
			record.  If the record is not in the array return false. */
		bool			remove(const Record &record, IWORD *index = NULL);
		/** Return the record at the given index. */
		Record			getRecord(IWORD index);
		/// Return a non-persistant record for the given index
		WeakRecord		getWeakRecord(IWORD index);
		/**	Clear the record array. */
		void			clear(void);
		/**	Return true if @em record is in the record array.  Otherwise return
			false. */
		bool			find(const Record &record);

		void			set(WeakRecord &record, IWORD index);

		void			setLayout(const sp<Layout>& rspLayout);

						/// @brief Remove any invalid weak references
		void			prune(void);

						/** @internal

							@brief Choose weak referencing

							This should only be set by RecordGroup at
							construction. */
		void			setWeak(BWORD weak);

						/// @brief Return TRUE if using weak referencing
		BWORD			isWeak(void) const		{ return m_weak; }

						/** Remove record at given index.  The array is kept
							contiguous by moving the last element into the
							the given index after removal.  Therefore, in
							an iteration that involves deletion it is best
							to iterate backwards through the array. */
		void			remove(IWORD index);
	private:
		void			grow(void);
		void			copy(const RecordArray &other);
		void			acquireSB(IWORD index);
		void			releaseSB(IWORD index);

		I32				readSerialNumber(IWORD index) const;

		void			trackBlock(U32 index);
		void			untrackBlock(U32 index);

	private:
		sp<Layout>		m_spLayout;
		void			**m_ppSB;
		IWORD			*m_pSN;
		IWORD			m_allocated;
		IWORD			m_used;
		BWORD			m_weak;
		I32				m_serialOffset;
#ifdef FE_RA_INDEX
		t_sb_map		m_sb_map;
#endif

#if FE_COUNTED_TRACK
	public:
const	String&	name(void) const
				{	return m_spLayout.isValid()?
						m_spLayout->name(): Counted::name(); }
const	String	verboseName(void) const
				{	return "RecordArray " +
							String(m_weak? "(weak) ": "") + name(); }
#endif
};

class PtrRecordArrayInfo : public BaseType::Info
{
	public:
virtual	IWORD	output(std::ostream &ostrm, void *instance, t_serialMode mode);
virtual	void	input(std::istream &istrm, void *instance, t_serialMode mode);
virtual	IWORD	iosize(void);
virtual	bool	getConstruct(void);
virtual	void	construct(void *instance);
virtual	void	destruct(void *instance);
};

inline void RecordArray::setWeak(BWORD weak)
{
	m_weak = weak;
#if FE_COUNTED_TRACK
	//* propagate name change
	sp<RecordArray> spRA(this);
#endif
}

inline void RecordArray::set(WeakRecord &record, IWORD index)
{
	record.set(m_ppSB[index]);
}

inline Layout *RecordArray::rawLayout(void) const
{
	return m_spLayout.raw();
}

inline void *RecordArray::operator[](IWORD index)
{
	return m_ppSB[index];
}

inline IWORD RecordArray::length(void)
{
	return m_used;
}

inline sp<Layout> RecordArray::layout(void)
{
	return m_spLayout;
}

inline Record RecordArray::getRecord(IWORD index)
{
	Record record;
	void* pSB=data(index);
	if(pSB)
	{
		record.set(pSB);
	}
	return record;
}

inline WeakRecord RecordArray::getWeakRecord(IWORD index)
{
	WeakRecord record;
	void* pSB=data(index);
	if(pSB)
	{
		record.set(pSB);
	}
	return record;
}

inline void RecordArray::acquireSB(IWORD index)
{
	if(!m_weak)
	{
		FE_SB_TO_HDR(m_ppSB[index])->m_pStore->acquireSB(m_ppSB[index]);
	}
	trackBlock(index);
}

inline void RecordArray::releaseSB(IWORD index)
{
	untrackBlock(index);
	if(!m_weak)
	{
		FE_SB_TO_HDR(m_ppSB[index])->m_pStore->releaseSB(m_ppSB[index]);
	}
}

inline I32 RecordArray::readSerialNumber(IWORD index) const
{
	return (m_serialOffset>=0 && m_ppSB[index])?
			*((I32 *)((char *)(m_ppSB[index]) + m_serialOffset)): -1;
}

inline void* RecordArray::data(IWORD index)
{
#if FE_CODEGEN<=FE_DEBUG
	if(index >= m_used || index < 0)
	{
		feX("fe::RecordArray::data","index out of range");
	}
#endif
//	if(m_weak)
//	{
//		feLog("RecordArray::data %d SN %d vs %d %s\n",
//				index,m_pSN[index],readSerialNumber(index),name().c_str());
//	}

	if(m_weak && m_pSN[index]>=0 && m_pSN[index]!=readSerialNumber(index))
	{
		feLog("RecordArray::data weak element %d is not valid (%d vs %d)\n",
				index,m_pSN[index],readSerialNumber(index));
#if FE_COUNTED_TRACK
		feLog("  name \"%s\"\n",name().c_str());
#endif

		m_ppSB[index]=NULL;
		return NULL;
	}
	return m_ppSB[index];
}

inline void RecordArray::untrackBlock(U32 index)
{
#if FE_COUNTED_TRACK
	untrackReference(FE_SB_TO_HDR(m_ppSB[index]),this);
#endif
}

inline void RecordArray::trackBlock(U32 index)
{
#if FE_COUNTED_TRACK
	String blockname;
	blockname.sPrintf("block:%d %s",index,name().c_str());
	trackReference(FE_SB_TO_HDR(m_ppSB[index]),this);
#endif
}

};

#endif /* __data_RecordArray_h__ */
