/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "data.pmh"

#include <sstream>

#define FE_COOKBOOK_DEBUG	FALSE

namespace fe
{

sp<RecordCookbook::Recipe> RecordCookbook::lookup(String recipeName)
{
	sp<Recipe> spRecipe=m_recipeMap[recipeName];
	if(!spRecipe.isValid())
	{
		m_recipeMap[recipeName]=
				new Recipe(recipeName,sp<RecordCookbook>(this));
		spRecipe=m_recipeMap[recipeName];
#if FE_COUNTED_TRACK
		spRecipe->trackReference(this);
#endif
	}
	return spRecipe;
}

void RecordCookbook::findLayoutList(String recipeName,
		std::vector<String>& rLayoutList)
{
#if FE_COOKBOOK_DEBUG
	feLog("RecordCookbook::findLayoutList %s\n",recipeName.c_str());
#endif

	sp<Recipe> spRecipe=lookup(recipeName);
	if(spRecipe.isValid())
	{
		spRecipe->findLayoutList(rLayoutList);
	}
}

void RecordCookbook::derive(String recipeName,String parentName)
{
#if FE_COOKBOOK_DEBUG
	feLog("RecordCookbook::derive(%s,%s)\n",recipeName.c_str(),
			parentName.c_str());
#endif

	sp<Recipe> spRecipe=lookup(recipeName);
	if(spRecipe.isValid())
	{
		spRecipe->derive(parentName);
	}

	spRecipe=lookup(parentName);
	if(spRecipe.isValid())
	{
		spRecipe->adopt(recipeName);
	}
}

void RecordCookbook::add(String recipeName,String attributeName,String value)
{
#if FE_COOKBOOK_DEBUG
	feLog("RecordCookbook::add(%s,%s,%s)\n",recipeName.c_str(),
			attributeName.c_str(),value.c_str());
#endif

	sp<Recipe> spRecipe=lookup(recipeName);
	if(spRecipe.isValid())
	{
		spRecipe->add(attributeName,value);
	}
}

void RecordCookbook::cook(String recipeName,Record& rRecord)
{
	cook_internal(recipeName,rRecord,TRUE,NULL);
}

void RecordCookbook::cook_internal(String recipeName,Record& rRecord,
		BWORD try_cache,StringIngredientMap* pNewCacheMap)
{
	if(!rRecord.isValid())
	{
		feLog("RecordCookbook::cook invalid Record\n");
		return;
	}
	sp<Layout> spLayout=rRecord.layout();
	if(!spLayout.isValid())
	{
		feLog("RecordCookbook::cook invalid Layout\n");
		return;
	}

#if FE_COOKBOOK_DEBUG
	feLog("RecordCookbook::cook \"%s\" as \"%s\"\n",
			spLayout->name().c_str(),recipeName.c_str());
#endif

	sp<Recipe> spRecipe=m_recipeMap[recipeName];
	if(spRecipe.isValid())
	{
		spRecipe->cook(rRecord,try_cache,pNewCacheMap);
	}
}

void RecordCookbook::spoil(String recipeName)
{
#if FE_COOKBOOK_DEBUG
	feLog("RecordCookbook::spoil(%s)\n",recipeName.c_str());
#endif

	sp<Recipe> spRecipe=lookup(recipeName);
	if(spRecipe.isValid())
	{
		spRecipe->spoil();
	}
}

void RecordCookbook::Recipe::derive(String parentName)
{
#if FE_COOKBOOK_DEBUG
	feLog("RecordCookbook::Recipe::derive(%s)\n",parentName.c_str());
#endif

	m_prerequisites.push_back(parentName);
	spoil();
}

void RecordCookbook::Recipe::adopt(String childName)
{
#if FE_COOKBOOK_DEBUG
	feLog("RecordCookbook::Recipe::adopt(%s)\n",childName.c_str());
#endif

	m_dependents.push_back(childName);
}

void RecordCookbook::Recipe::spoil(void)
{
#if FE_COOKBOOK_DEBUG
	feLog("RecordCookbook::Recipe::spoil \"%s\"\n",m_name.c_str());
#endif

	m_cacheMap.clear();

	U32 count=m_dependents.size();
	for(U32 m=0;m<count;m++)
	{
		//* TODO cycle protection

		m_hpRecordCookbook->spoil(m_dependents[m]);
	}
}

void RecordCookbook::Recipe::findLayoutList(std::vector<String>& rLayoutList)
{
	if(m_hpRecordCookbook->scope()->lookupLayout(m_name).isValid())
	{
#if FE_COOKBOOK_DEBUG
		feLog("RecordCookbook::Recipe::findLayoutList \"%s\" exists\n",
				m_name.c_str());
#endif
		rLayoutList.push_back(m_name);
		return;
	}
	else
	{
		U32 count=m_prerequisites.size();
		for(U32 m=0;m<count;m++)
		{
			//* TODO cycle protection

#if FE_COOKBOOK_DEBUG
			feLog("RecordCookbook::Recipe::findLayoutList"
					" \"%s\" check \"%s\"\n",
					m_name.c_str(),m_prerequisites[m].c_str());
#endif

			m_hpRecordCookbook->findLayoutList(m_prerequisites[m],rLayoutList);
		}
	}
}

void RecordCookbook::Recipe::add(String attributeName,String value)
{
#if FE_COOKBOOK_DEBUG
	feLog("RecordCookbook::Recipe::add(%s,%s)\n",
			attributeName.c_str(),value.c_str());
#endif

	m_attributeMap[attributeName]=value;
	spoil();
}

void RecordCookbook::Recipe::cook(Record& rRecord,BWORD try_cache,
		StringIngredientMap* pNewCacheMap)
{
#if FE_COOKBOOK_DEBUG
	feLog("RecordCookbook::Recipe::cook\n");
#endif

	if(try_cache && m_cacheMap.size())
	{
		//* use cache
		StringIngredientMap::iterator it;
		for(it=m_cacheMap.begin();it!=m_cacheMap.end();it++)
		{
			Ingredient& value=it->second;

#if FE_COOKBOOK_DEBUG
			String attr_name=it->first;
			feLog("RecordCookbook::Recipe::cook template \"%s\""
					" %s:%p,%p,%p %p\n",m_name.c_str(),attr_name.c_str(),
					value.m_offset,value.m_spBT->size(),
					value.m_block,*value.m_block);
#endif

			BWORD success=value.m_spBT->copy(
					((char*)rRecord.data())+value.m_offset,
					value.m_block);
			if(!success)
			{
				String attr_name=it->first;
				feLog("RecordCookbook::Recipe::cook copy failed \"%s\""
						" %s:%p,%p,%p %p\n",m_name.c_str(),attr_name.c_str(),
						value.m_offset,value.m_spBT->size(),
						value.m_block,*value.m_block);
			}
		}

		return;
	}

	if(!pNewCacheMap)
	{
		pNewCacheMap=&m_cacheMap;
	}

	//* prerequisites recipes
	U32 count=m_prerequisites.size();
	for(U32 m=0;m<count;m++)
	{
		//* TODO cycle protection

#if FE_COOKBOOK_DEBUG
		feLog("RecordCookbook::Recipe::cook prerequisite \"%s\"\n",
				m_prerequisites[m].c_str());
#endif

		m_hpRecordCookbook->cook_internal(m_prerequisites[m],rRecord,
				FALSE,pNewCacheMap);
	}

	//* this recipe
	StringStringMap::iterator it;
	for(it=m_attributeMap.begin();it!=m_attributeMap.end();it++)
	{
		String attr_name=it->first;
		String& value=it->second;


		//* parse like AsciiReader

		UWORD index=0;
		sp<Attribute>& rspAttribute=
				m_hpRecordCookbook->scope()->findAttribute(attr_name,index);
		if(!rspAttribute.isValid())
		{
			feX("RecordCookbook::Recipe::cook",
					"attribute \"%s\" not found in scope for \"%s\"\n",
					attr_name.c_str(),m_name.c_str());
		}

		UWORD offset=rRecord.layout()->offsetTable()[index];

		void *instance = (void *)((char *)(rRecord.data()) + offset);
		sp<BaseType> spBT = rspAttribute->type();

		if(offset==fe::offsetNone)
		{
			feX("RecordCookbook::Recipe::cook",
					"offsetTable[%d] for \"%s\" in \"%s\" invalid\n",
					index,attr_name.c_str(),m_name.c_str());
		}

		if(spBT->getInfo().isValid())
		{
			std::string s(value.c_str());
			std::istringstream istrm(s);
			spBT->getInfo()->input(istrm, instance,BaseType::Info::e_ascii);

			(*pNewCacheMap)[attr_name].set(offset,spBT,(char*)instance);
		}
		else
		{
			feX("RecordCookbook::Recipe::cook","invalid type");
		}
	}
}

void RecordCookbook::Ingredient::set(UWORD offset,sp<BaseType>& rspBT,
		char* data)
{
	if(m_block)
	{
		delete[] m_block;
	}

	m_offset=offset;
	m_spBT=rspBT;
	m_block=new char[rspBT->size()];

	rspBT->construct(m_block);
	BWORD success=rspBT->copy(m_block,data);
	if(!success)
	{
		feLog("RecordCookbook::Ingredient::set copy failed\n");
	}
}

} /* namespace */
