/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "data.pmh"

#define FE_SC_RECORDFACTORY_DEBUG		FALSE

namespace fe
{

Scope::Scope(void)
{
}

void Scope::initialize(void)
{
	if(m_spTypeMaster.isValid())
		return;

	m_hpRegistry=registry();
	if(!m_hpRegistry.isValid())
	{
		feLog("Scope::Scope WARNING registry() failed\n");
		FEASSERT(m_hpRegistry.isValid());
		return;
	}
	sp<Master> spMaster=m_hpRegistry->master();
	if(!spMaster.isValid())
	{
		feLog("Scope::Scope WARNING spRegistry.master() failed\n");
		FEASSERT(spMaster.isValid());
		return;
	}

	construct(spMaster->typeMaster(), spMaster->allocator());

	sp<RecordGroup> &rg_scopes = spMaster->catalog< sp<RecordGroup> >("Scopes");
	if(!rg_scopes.isValid())
	{
		rg_scopes = new RecordGroup();
	}
	sp<Layout> l_component = declare("l_component");
	AsComponent asComponent;
	asComponent.populate(l_component);
	Record r_self = createRecord(l_component);
	asComponent.component(r_self) = this;
	rg_scopes->add(r_self);
}

Scope::Scope(const Master &master)
{
	construct(master.typeMaster(), master.allocator());
}

Scope::Scope(sp<TypeMaster> spTypeMaster, sp<Allocator> allocator)
{
	construct(spTypeMaster, allocator);
}

void assertData(sp<TypeMaster> spTypeMaster)
{
	sp<BaseType> spBT =
		spTypeMaster->assertType<sp<RecordGroup> >("RecordGroup");
	if(!spBT->getInfo().isValid())
	{
		spBT->setInfo(new PtrRecordGroupInfo());
	}
	spBT = spTypeMaster->assertType<sp<RecordGroup> >("group");

	spBT =
		spTypeMaster->assertType<sp<RecordArray> >("RecordArray");
	if(!spBT->getInfo().isValid())
	{
		spBT->setInfo(new PtrRecordArrayInfo());
	}
	spBT = spTypeMaster->assertType<sp<RecordArray> >("array");

	spBT = spTypeMaster->assertType<Record>("record");
	if(!spBT->getInfo().isValid())
	{
		spBT->setInfo(new RecordInfo());
	}

	spBT = spTypeMaster->assertType<WeakRecord>("WeakRecord");
	if(!spBT->getInfo().isValid())
	{
		spBT->setInfo(new WeakRecordInfo());
	}

	spBT = spTypeMaster->assertPtr<Component>("component");
	spBT = spTypeMaster->assertPtr<Layout>("layout");
}

void Scope::construct(sp<TypeMaster> spTypeMaster, sp<Allocator> allocator)
{
	setName("Scope");

	m_spTypeMaster = spTypeMaster;

	if(allocator.isValid())
	{
		m_spAllocator = allocator;
	}
	else
	{
		m_spAllocator = new BasicAllocator();
	}

	m_spRecordCookbook=new RecordCookbook(hp<Scope>(this));

	//m_spTypeMaster->assertPtr<RecordGroup>("RecordGroup");
	assertData(m_spTypeMaster);

	support(FE_USE(":CT"), "integer");
	m_paCount = new Accessor<int>;
	m_paCount->initialize(this, FE_USE(":CT"));
	findAttribute(FE_USE(":CT"))->setSerialize(false);
	findAttribute(FE_USE(":CT"))->setCloneable(false);

	support(FE_USE(":ID"), "integer");
	m_paIDNumber = new Accessor<int>;
	m_paIDNumber->initialize(this, FE_USE(":ID"));

	sp<Attribute> spAttrSN=support(FE_USE(":SN"), "integer");
	spAttrSN->setSerialize(FALSE);
	spAttrSN->setCloneable(FALSE);
	m_paSerialNumber = new Accessor<int>;
	m_paSerialNumber->initialize(this, FE_USE(":SN"));

	m_serialIndex=attributeIndex(m_paSerialNumber->attribute());

	m_looking=0;
	m_serialCount=0;
	m_storeName = "";
}

Scope::Scope(const Scope &other)
{
	FEASSERT(0);
}

Scope::~Scope(void)
{
	shutdown();
}

bool Scope::shutdown(void)
{
	SAFEGUARD;
	m_layouts.clear();
	m_storeMap.clear();
	m_depend.clear();
	m_idNumbered.clear();
	m_freeIDNumbers.clear();
	m_spAllocator = NULL;
	m_spTypeMaster = NULL;
	if(m_paCount)
	{
		delete m_paCount;
		m_paCount = NULL;
	}
	if(m_paIDNumber)
	{
		delete m_paIDNumber;
		m_paIDNumber = NULL;
	}
	if(m_paSerialNumber)
	{
		delete m_paSerialNumber;
		m_paSerialNumber = NULL;
	}
	return true;
}

Scope &Scope::operator=(const Scope &other)
{
	FEASSERT(0);
	return *this;
}

sp<RecordCookbook> Scope::cookbook(void)
{
	return m_spRecordCookbook;
}

void Scope::layoutValidate(void)
{
	SAFEGUARD;
	for(t_depends::iterator it_outer = m_depend.begin();
			it_outer != m_depend.end();
			it_outer++)
	{
		t_depends::iterator it_inner = it_outer;
		for(	it_inner++;
				it_inner != m_depend.end();
				it_inner++)
		{
			if(**it_outer == **it_inner)
			{
				// delete duplicate
				//t_depends::iterator it_remove = it_inner;
				//it_inner--;
				sp<BaseType> inner_t =
					m_spTypeMaster->lookupName((*it_inner)->attributeType());
				feX("Scope::layoutValidate",
					"duplicates should no longer get here [%s]",
					(*it_inner)->attributeName().c_str());
				//m_depend.remove(*it_remove);
			}
			else if((*it_outer)->attributeName() == (*it_inner)->attributeName())
			{
				sp<BaseType> inner_t =
					m_spTypeMaster->lookupName((*it_inner)->attributeType());
				sp<BaseType> outer_t =
					m_spTypeMaster->lookupName((*it_outer)->attributeType());
				if(inner_t != outer_t)
				{
					if(		((*it_outer)->dependFlag() & Depend::e_available)
						&&	((*it_inner)->dependFlag() & Depend::e_available) )
					{
						Peeker peek;
						peek(*this);
						if(!inner_t.isValid())
						{
							feX("fe::Scope::layoutValidate",
								"%s: attribute type not valid: %s\n%s",
								(*it_inner)->attributeName().c_str(),
								(*it_inner)->attributeType().c_str(),
								peek.output().c_str());
						}
						if(!outer_t.isValid())
						{
							feX("fe::Scope::layoutValidate",
								"%s: attribute type not valid: %s\n%s",
								(*it_outer)->attributeName().c_str(),
								(*it_outer)->attributeType().c_str(),
								peek.output().c_str());
						}
						feX("fe::Scope::layoutValidate",
							"attribute type conflict on %s: %s != %s\n%s",
							(*it_outer)->attributeName().c_str(),
							(*it_outer)->attributeType().c_str(),
							(*it_inner)->attributeType().c_str(),
							peek.output().c_str());
					}
				}
			}
		}
	}
}

sp<Attribute> Scope::dynamicAddAttribute(const String &typeName,
		const String &attributeName)
{
#if FE_CODEGEN<=FE_DEBUG
	layoutValidate();
#endif

	SAFEGUARD;
	// if attribute already exists return
	UWORD attributeIndex;
	sp<Attribute> spAttribute = findAttribute(attributeName, attributeIndex);
	if(spAttribute.isValid()) { return spAttribute; }

	UWORD cnt = getAttributeCount();

	spAttribute = addAttribute(typeName, attributeName);

	UWORD newcnt = getAttributeCount();

	FEASSERT(newcnt == cnt+1);

#if 0
	UWORD layoutCount = m_layouts.size();
	for(UWORD index = 0; index < layoutCount; index ++)
	{
		const Offset *lastOffsetTable = m_layouts[index]->rawOffsetTable();
		if(lastOffsetTable)
		{
			m_layouts[index]->setOffsetTable(new Offset[newcnt]);
			for(UWORD i = 0;i < cnt; i++)
			{
				m_layouts[index]->setOffset(i, lastOffsetTable[i]);
			}
			m_layouts[index]->setOffset(cnt, offsetNone);
/*
			feLog("Scope:dynamicAddAttribute %d %s delete[] %p %s\n",
					index,m_layouts[index]->name().c_str(),
					lastOffsetTable,attributeName.c_str());
*/
			delete[] lastOffsetTable;
		}
	}
#endif
	for(unsigned int i = 0; i < m_offsetTables.size(); i++)
	{
		m_offsetTables[i]->resize(newcnt);
	}
	return spAttribute;
}

sp<Attribute> Scope::addAttribute(const String &typeName,
		const String &attributeName)
{
	SAFEGUARD;
	sp<BaseType> spAttributeType;
	FEASSERT(m_spTypeMaster.isValid());
	spAttributeType = m_spTypeMaster->lookupName(typeName);
	if(!spAttributeType.isValid())
	{
		feX("fe::Scope::addAttribute",
			"could not add \'%s\'; attribute type \'%s\' not known",
			attributeName.c_str(), typeName.c_str());
	}
	sp<Attribute> spAttribute(new Attribute(attributeName,spAttributeType,
			m_spTypeMaster));
	m_attributes.push_back(spAttribute);
	return spAttribute;
}

sp<Layout> Scope::declare(const String &name)
{
	SAFEGUARD;
	sp<Layout> spLayout = lookupLayout(name);
	populate(name, FE_USE(":CT"));
	//populate(name, FE_I);
	if(spLayout.isValid())
	{
		return spLayout;
	}
	spLayout = new Layout(this, name);
	m_layouts.push_back(spLayout);
	spLayout->adjoin(sp<Scope>(this));
	return spLayout;
}

bool Scope::lookupStore(sp<Layout> spLayout, sp<StoreI> &spStore)
{
	SAFEGUARD;
	layout_store_map::iterator it = m_storeMap.find(spLayout);
	if(it != m_storeMap.end())
	{
		spStore = it->second;
		return true;
	}

	return false;
}

void Scope::freeIDNumber(sp<Layout> spLayout, void *stateblock)
{
	SAFEGUARD;
	int *pID;
	if(m_paIDNumber &&
		m_paIDNumber->queryAttribute(spLayout, stateblock, pID))
	{
//		feLog("freeIDNumber %p %d\n",stateblock,*pID);
		m_freeIDNumbers.push_back(*pID);
		m_idNumbered[*pID].m_pStateBlock = NULL;
	}

	//* piggyback
	int *pSN;
	if(m_paSerialNumber &&
		m_paSerialNumber->queryAttribute(spLayout, stateblock, pSN))
	{
//		feLog("freeIDNumber SN %p %d\n",stateblock,*pSN);
		*pSN= -1;
	}
}

void Scope::assignIDNumber(const Record &record)
{
	SAFEGUARD;
	int *pID;
	pID = m_paIDNumber->queryAttribute(record);
	if(pID)
	{
		if(m_freeIDNumbers.size() != 0)
		{
			m_idNumbered[m_freeIDNumbers.back()] = record;
			*pID = m_freeIDNumbers.back();
//			feLog("re assignIDNumber %p %d\n",record.data(),*pID);
			m_freeIDNumbers.pop_back();
		}
		else
		{
			int index = m_idNumbered.size();
//			feLog("assignIDNumber %p %d\n",record.data(),index);
			m_idNumbered.resize(index + 1);
			*pID = index;
			WeakRecord weak_record(record);
			m_idNumbered[index] = weak_record;
		}
	}

	//* piggyback
	int *pSN;
	pSN = m_paSerialNumber->queryAttribute(record);
	if(pSN)
	{
		*pSN=m_serialCount++;
//		feLog("assignIDNumber SN %p %d %s\n",record.data(),*pSN,
//				record.layout()->name().c_str());
		if(m_serialCount<0)
		{
			m_serialCount=0;
		}
	}
}

Record Scope::getRecord(IWORD id)
{
	if(id >= (IWORD)(m_idNumbered.size()))
	{
		feX(e_invalidRange,
			"Scope::getRecord",
			"index out of range");
	}
	if(m_idNumbered[id].data() == NULL)
	{
		feX(e_invalidRange,
			"Scope::getRecord",
			"record id not valid");
	}
	Record record;
	record.set(m_idNumbered[id].data());
	return record;
}

void Scope::registerFactory(String name,sp<RecordFactoryI> spRecordFactoryI)
{
	m_factoryMap[name]=spRecordFactoryI;
}

Record Scope::produceRecord(String name)
{
	Record record;

	std::vector<String> layoutList;
	m_spRecordCookbook->findLayoutList(name,layoutList);
//	if(!layoutList.size())
//	{
//		layoutList.push_back(name);
//	}

	if(layoutList.size()>1)
	{
		sp<Layout> spLayout=declare(name);
		for(U32 m=0;m<layoutList.size();m++)
		{
			String prerequisite=layoutList[m];

#if FE_SC_RECORDFACTORY_DEBUG
			feLog("Scope::produceRecord \"%s\" clones \"%s\"\n",
					name.c_str(),prerequisite.c_str());
#endif

			clonePopulate(prerequisite.c_str(),name);
		}
		record=createRecord(spLayout);

		// NOTE once we make the Layout, findLayoutList won't give the old ones
		m_factoryMultiMap[name]=layoutList;
	}
	else if(m_factoryMultiMap.find(name)!=m_factoryMultiMap.end())
	{
		layoutList=m_factoryMultiMap[name];

		sp<Layout> spLayout=lookupLayout(name);
		if(spLayout.isValid())
		{
			record=createRecord(spLayout);
		}
	}

	for(U32 m=0;m<layoutList.size();m++)
	{
		String prerequisite=layoutList[m];

#if FE_SC_RECORDFACTORY_DEBUG
		feLog("Scope::produceRecord \"%s\" requires \"%s\"\n",
				name.c_str(),prerequisite.c_str());
#endif

		sp<RecordFactoryI> spRecordFactoryI=m_factoryMap[prerequisite];
		if(spRecordFactoryI.isValid())
		{
			sp<Scope> spScope(this);
			spRecordFactoryI->bind(spScope);
			spRecordFactoryI->produceRecord(record);
		}
		else if(!record.isValid())
		{
			sp<Layout> spLayout=lookupLayout(prerequisite);
			if(spLayout.isValid())
			{
				record=createRecord(spLayout);
			}
		}
	}

	if(!record.isValid())
	{
		feLog("Scope::produceRecord \"%s\" does not name a"
				" layout, factory, or recipe\n",
				name.c_str());
	}
	else
	{
		m_spRecordCookbook->cook(name,record);
	}
	return record;
}

void Scope::finalizeRecord(Record& rRecord)
{
	String name=rRecord.layout()->name();

	std::vector<String> layoutList;
	m_spRecordCookbook->findLayoutList(name,layoutList);

	if(layoutList.size()>1)
	{
		feX("Scope::finalizeRecord","\"%s\" has %d layouts\n",
				name.c_str(),layoutList.size());
	}
	else if(!layoutList.size())
	{
#if FE_SC_RECORDFACTORY_DEBUG
		feLog("Scope::finalizeRecord","\"%s\" has no layouts\n",
				name.c_str());
#endif
		return;
	}
	else if(m_factoryMultiMap.find(name)!=m_factoryMultiMap.end())
	{
		layoutList=m_factoryMultiMap[name];
	}

	for(U32 m=0;m<layoutList.size();m++)
	{
		String prerequisite=layoutList[m];

#if FE_SC_RECORDFACTORY_DEBUG
		feLog("Scope::finalizeRecord \"%s\" uses \"%s\"\n",
				name.c_str(),prerequisite.c_str());
#endif

		sp<RecordFactoryI> spRecordFactoryI=m_factoryMap[prerequisite];
		if(spRecordFactoryI.isValid())
		{
			sp<Scope> spScope(this);
			spRecordFactoryI->bind(spScope);
			spRecordFactoryI->finalizeRecord(rRecord);
		}
#if FE_SC_RECORDFACTORY_DEBUG
		else
		{
			feLog("Scope::finalizeRecord \"%s\" does not name a factory\n",
					name.c_str());
		}
#endif
	}
}

Record Scope::createRecord(sp<Layout> spLayout)
{
	SAFEGUARD;
	if(!spLayout->locked())
	{
		lockLayout(spLayout);
	}

	sp<StoreI> spStore;
	if(lookupStore(spLayout,spStore))
	{
		Record record;
		record.m_pStateBlock = spStore->createSB();
		record.m_spLayout = spLayout;
		spLayout->constructAttributes(record.data());
		assignIDNumber(record);
#if FE_SCOPE_SUPPORT_WATCHERS
		watch(record);
#endif
		return record;
	}
	else
	{
		feX("fe::Scope::createRecord",
			"attempt to create record of type not of same scope");
	}
}

sp<RecordArray> Scope::createRecordArray(sp<Layout> spLayout,UWORD count)
{
	SAFEGUARD;
	if(!spLayout->locked())
	{
		lockLayout(spLayout);
	}

	sp<RecordArray> spRA(new RecordArray(spLayout));

	// create its own store so that it will delete the memory when necessary
	sp<StoreI> spStore(createStore(spLayout));

	void *pArrayRoot = spStore->createSB(count);
	for(UWORD i = 0;i < count; i++)
	{
		WeakRecord record;
		record.m_hpLayout = spLayout;
		record.m_pStateBlock = pArrayRoot;
		spLayout->constructAttributes(record.data());
		assignIDNumber(record);

		//* NOTE RA add does Store acquire, so release refs done by createSB
		spRA->add(record);
		spStore->releaseSB(record.data());

#if FE_SCOPE_SUPPORT_WATCHERS
		watch(record);
#endif
		pArrayRoot = (void *)(((UWORD)pArrayRoot) + spStore->skipSizeSB());
	}

	return spRA;
}

void Scope::freeRecord(Record record)
{
	SAFEGUARD;
	sp<StoreI> spStore;
	if(lookupStore(record.m_spLayout,spStore))
	{
		if(m_paCount->check(record))
		{
			feX("fe::Scope::freeRecord",
				"attempt to explicitly free reference counted record");
		}
#if FE_SCOPE_SUPPORT_WATCHERS
		unwatch(record);
#endif
		spStore->freeSB(record.data());
	}
	else
	{
		feX("fe::Scope::freeRecord",
			"attempt to free record of type not of same scope");
	}
}

void Scope::lockLayout(sp<Layout> spLayout)
{
	if(spLayout->lockSuppressed())
	{
		feX("fe::Scope::lockLayout",
			"attempt to lock layout \'%s\' with lock suppression on",
			spLayout->name().c_str());
	}
	layoutValidate();

	SAFEGUARD;
	std::vector<String> attributes;

	FEASSERT(spLayout->scope().raw() == this);

	if(spLayout->locked())
	{
		feX("fe::Scope::lockLayout","attempt to lock already locked layout");
	}

	//const Offset *lastOffsetTable = spLayout->rawOffsetTable();

	UWORD cnt = getAttributeCount();

	// setup and clear new table
	if(spLayout->rawOffsetTable() != NULL)
	{
		feX("fe::Scope::lockLayout","offset table not null for lock");
	}

#if 0
	spLayout->setOffsetTable(new Offset[cnt]);
	for(U32 i = 0;i < cnt; i++)
	{
		spLayout->setOffset(i, offsetNone);
	}
#endif
	sp<OffsetTable> newOffsetTable(new OffsetTable());
	newOffsetTable->resize(cnt);
	newOffsetTable->attach(spLayout);

	sp<Depend> spDepend;
	String attributeName;
	t_depends::iterator it_depend;
	std::vector<String>::iterator it_attribute;
	std::vector< std::list<UWORD> > shareTable;
	shareTable.resize(cnt);
	std::map< UWORD, std::pair<Offset,String> > withinMap;
	std::map< UWORD, std::list<UWORD> > withinReverseMap;

	// add attributes based on Depend::e_populate dependencies
	// and build share table
	for(it_depend = m_depend.begin();it_depend != m_depend.end(); it_depend++)
	{
		if(((*it_depend)->dependName() == "")
					||	((*it_depend)->dependName() == spLayout->name()) )
		{
			if((*it_depend)->dependFlag() & Depend::e_populate)
			{
				attributes.push_back((*it_depend)->attributeName());
			}
			if((*it_depend)->dependFlag() & Depend::e_share)
			{
				UWORD idx_1, idx_2;
				sp<Attribute> spT1, spT2;
				spT1 = findAttribute((*it_depend)->shareName(), idx_1);
				spT2 = findAttribute((*it_depend)->attributeName(), idx_2);
				if(spT1.isValid() && spT2.isValid())
				{
					if(spT1->type() != spT2->type())
					{
						std::list<String> names;
						String s1,s2;
						m_spTypeMaster->reverseLookup(spT1->type(), names);
						s1.cat(names," ");
						names.clear();
						m_spTypeMaster->reverseLookup(spT2->type(), names);
						s2.cat(names, " ");
						feX("fe::Scope::lockLayout",
							"share type mismatch %s[%s] != %s[%s]",
							spT1->name().c_str(),s1.c_str(),
							spT2->name().c_str(),s2.c_str() );
					}
					shareTable[idx_1].push_back(idx_2);
					shareTable[idx_2].push_back(idx_1);
				}
			}
			if((*it_depend)->dependFlag() & Depend::e_within)
			{
				UWORD idx_p, idx_c;
				sp<Attribute> spParent, spChild;
				spParent = findAttribute((*it_depend)->shareName(), idx_p);
				spChild = findAttribute((*it_depend)->attributeName(), idx_c);
				if(spParent.isValid() && spChild.isValid())
				{
					std::map< UWORD, std::pair<Offset,String> >::iterator
						i_within = withinMap.find(idx_p);
					if(i_within != withinMap.end())
					{
						std::list<String> names;
						String s1,s2;
						m_spTypeMaster->reverseLookup(spParent->type(), names);
						s1.cat(names," ");
						names.clear();
						m_spTypeMaster->reverseLookup(spChild->type(), names);
						s2.cat(names, " ");
						feX("fe::Scope::lockLayout",
							"cannot next withins %s[%s] %s[%s]",
							spParent->name().c_str(),s1.c_str(),
							spChild->name().c_str(),s2.c_str() );
					}
					withinMap[idx_c] =
						std::pair<Offset, String>
						((*it_depend)->offset(), spParent->name());
					withinReverseMap[idx_p].push_back(idx_c);
				}
			}
		}
	}

	// add attributes based on Depend::e_attribute dependencies
	t_depends depend_list;
	for(it_depend = m_depend.begin();it_depend != m_depend.end(); it_depend++)
	{
		if((*it_depend)->dependFlag() & Depend::e_attribute)
		{
			depend_list.push_back(*it_depend);
			std::sort((*it_depend)->matchAttributes().begin(),
				(*it_depend)->matchAttributes().end());
		}
	}
	bool change = true;
#if 0
	while(change)
	{
		change = false;
		for(it_attribute = attributes.begin();it_attribute != attributes.end();
				it_attribute++)
		{
			for(it_depend = depend_list.begin();
				it_depend != depend_list.end();/* it_depend++ below */)
			{
				if((*it_depend)->matchAttributes()[0] == *it_attribute)
				{
					change = true;
					attributes.push_back((*it_depend)->attributeName());
					sp<Depend> spDep = *it_depend;
					it_depend++;
					depend_list.remove(spDep);
				}
				else
				{
					it_depend++;
				}
			}
		}
	}
#endif
	while(change)
	{
		change = false;
		std::sort(attributes.begin(), attributes.end());
		for(it_depend = depend_list.begin();
				it_depend != depend_list.end();/* it_depend++ below */)
		{
			std::vector<String> &matchattrs = (*it_depend)->matchAttributes();

			if(std::includes<std::vector<String>::iterator,
				std::vector<String>::iterator>(
					attributes.begin(), attributes.end(),
					matchattrs.begin(), matchattrs.end()))
			{
				change = true;
				attributes.push_back((*it_depend)->attributeName());
				sp<Depend> spDep = *it_depend;
				it_depend++;
				depend_list.remove(spDep);
			}
			else
			{
				it_depend++;
			}
		}
	}


	spLayout->setSize(0);
	sp<Attribute> spAttribute;
	UWORD index;

	for(it_attribute = attributes.begin();it_attribute != attributes.end();
			it_attribute++)
	{
		spAttribute = findAttribute(*it_attribute, index);
		if(!spAttribute.isValid())
		{
			feX("fe::Scope::lockLayout",
				"Layout \"%s\" not initialized because "
				"Attribute \"%s\" was not found",
				spLayout->name().c_str(), it_attribute->c_str());
		}
		if(spLayout->rawOffsetTable()[index] == offsetNone)
		{
			// being a child within :) overrides getting added
			std::map< UWORD, std::pair<Offset,String> >::iterator i_within =
				withinMap.find(index);
			if(i_within != withinMap.end())
			{
				bool is_within = false;
				for(std::vector<String>::iterator ja = attributes.begin();
					ja != attributes.end(); ja++)
				{
					if(*ja == i_within->second.second)
					{
						is_within = true;
						break;
					}
				}
				if(is_within)
				{
					continue;
				}
			}

			// check for share
			for(std::list<UWORD>::iterator it_share
					= shareTable[index].begin();
				it_share != shareTable[index].end();
				it_share++)
			{
				UWORD idx = *it_share;
				if(spLayout->rawOffsetTable()[idx] != offsetNone)
				{
					feLog("Scope::lockLayout share %d %p\n",
							index, spLayout->rawOffsetTable()[idx]);
					spLayout->setOffset(index, spLayout->rawOffsetTable()[idx]);
					break;
				}
			}
			// if not shared, add
			if(spLayout->rawOffsetTable()[index] == offsetNone)
			{
				if(!spAttribute->type().isValid())
				{
					feX("fe::Scope::lockLayout",
						"Layout \"%s\" not initialized because "
						"Attribute \"%s\" had no valid type",
						spLayout->name().c_str(), it_attribute->c_str());
				}
				UWORD alignment=0;
				if(spAttribute->type()->getInfo().isValid())
				{
					alignment=spAttribute->type()->getInfo()->alignment();
				}
				if(alignment)
				{
					// TODO: rearrange layout for more efficient alignment
					/*** NOTE SegmentStore actually has 16 bytes already filled
						from the original allocate(), SBHeader + Segment,
						but that has no net skew. */
					UWORD skew=spLayout->size()%alignment;
					if(skew)
					{
						UWORD pad=alignment-skew;

						feLog("Scope::lockLayout padding %s:%s"
								" to %p at %p by %p\n",
								spLayout->name().c_str(),
								spAttribute->name().c_str(),
								alignment,spLayout->size(),pad);

						spLayout->setSize(spLayout->size()+pad);
					}
				}
#if FALSE
				feLog("Scope::lockLayout %s for %s at %p align %p\n",
						spLayout->name().c_str(),
						spAttribute->name().c_str(),
						spLayout->size(),
						alignment);
#endif
				spLayout->setOffset(index, spLayout->size());
				spLayout->setSize(spLayout->size() +
						spAttribute->type()->size());

				// check for children within
				std::map< UWORD, std::list<UWORD> >::iterator i_reverse =
					withinReverseMap.find(index);
				if(i_reverse != withinReverseMap.end())
				{
					for(std::list<UWORD>::iterator i_child =
						i_reverse->second.begin();
						i_child != i_reverse->second.end(); i_child++)
					{
						UWORD idx_c = *i_child;
						spLayout->setOffset(idx_c,
							spLayout->rawOffsetTable()[index] +
							withinMap[idx_c].first);
					}
				}
			}
			if(spAttribute->type()->getConstruct())
			{
				spLayout->setConstruct(true);
			}
		}
	}

	// use a matching/identical offset table if pre-existing one can be found
	// TODO: reference count offset tables and turn this back on
#if 0
	for(unsigned int i_l = 0; i_l < m_layouts.size(); i_l++)
	{
		sp<Layout> spOther = m_layouts[i_l];
		if(spOther == spLayout) { continue; }
		if(!spOther->locked()) { continue; }
		UWORD attr_cnt = getAttributeCount();
		bool match = true;
		for(UWORD i_a = 0; i_a < attr_cnt; i_a++)
		{
			if(spOther->offsetTable()[i_a] != spLayout->rawOffsetTable()[i_a])
			{
				match = false;
				break;
			}
		}
		if(match)
		{
			Offset *pTable = const_cast<Offset *>(spLayout->rawOffsetTable());
			feLog("Scope::lockLayout delete[] %p\n",pTable);
			delete[] pTable;
			pTable = const_cast<Offset *>(spOther->offsetTable());
			spLayout->setOffsetTable(pTable);
			break;
		}
	}
#endif

#if 1
	bool append = true;
	for(unsigned int i_ot = 0; i_ot < m_offsetTables.size(); i_ot++)
	{
		bool match = true;
		UWORD attr_cnt = getAttributeCount();
		for(UWORD i_a = 0; i_a < attr_cnt; i_a++)
		{
			if(newOffsetTable->table()[i_a] != m_offsetTables[i_ot]->table()[i_a])
			{
				match = false;
				break;
			}
		}
		if(match)
		{
			m_offsetTables[i_ot]->attach(spLayout);
			append = false;
			break;
		}

	}
	if(append)
	{
		m_offsetTables.push_back(newOffsetTable);
	}
#else
	m_offsetTables.push_back(newOffsetTable);
#endif

	spLayout->cacheIndices();
	spLayout->setLocked(true);

#if 0
	if(lastOffsetTable)
	{
		feX("fe::Scope::lockLayout",
			"automatic record conversion not implemented");
		delete [] lastOffsetTable;
	}
#endif

	// add a store for this type
	sp<StoreI> spStore;
	if(lookupStore(spLayout,spStore))
	{
		feX("fe::Scope::lockLayout",
			"store for newly locked type already exists");
	}
	spStore = createStore(spLayout);
	m_storeMap[spLayout] = spStore;
	spStore->trackReference(dynamic_cast<Counted*>(this));

}

sp<Layout> Scope::lookupLayout(const String &name)
{
	SAFEGUARD;
	sp<Layout> spLayout;
	for(UWORD i = 0; i < m_layouts.size(); i++)
	{
		if(m_layouts[i]->name() == name)
		{
			spLayout = m_layouts[i];
		}
	}
	if(!m_looking && !spLayout.isValid() && m_factoryMap[name].isValid())
	{
		m_looking=TRUE;

		sp<Scope> spScope(this);
		m_factoryMap[name]->bind(spScope);
		spLayout=m_factoryMap[name]->layout();

		m_looking=FALSE;
	}
	return spLayout;
}

UWORD Scope::getAttributeCount(void)
{
	SAFEGUARD;
	return m_attributes.size();
}

sp<Attribute> Scope::support(const String &mayHave, const String &ofType)
{
	SAFEGUARD;
	sp<Depend> spDepend(new Depend(mayHave));
	spDepend->attributeType() = ofType;
	spDepend->dependFlag() = Depend::e_available;
	return addDepend(spDepend);
}

bool Scope::enforce(const String &ifHas, const String &mustHave)
{
	SAFEGUARD;
	sp<Depend> spDepend(new Depend(mustHave));
	spDepend->matchAttributes().push_back(ifHas);
	spDepend->dependFlag() = Depend::e_attribute;
	addDepend(spDepend);
	return true;
}

bool Scope::enforce(const std::vector<String> &ifHas, const String &mustHave)
{
	SAFEGUARD;
	sp<Depend> spDepend(new Depend(mustHave));
	spDepend->matchAttributes() = ifHas;
	spDepend->dependFlag() = Depend::e_attribute;
	addDepend(spDepend);
	return true;
}

bool Scope::populate(const String &ifIs, const String &mustHave)
{
	SAFEGUARD;
	sp<Depend> spDepend(new Depend(mustHave));
	spDepend->dependName() = ifIs;
	spDepend->dependFlag() = Depend::e_populate;
	addDepend(spDepend);
	return true;
}

bool Scope::share(const String &ifIs, const String &share1,
		const String &share2)
{
	SAFEGUARD;
	sp<Depend> spDepend(new Depend(share1));
	spDepend->shareName() = share2;
	spDepend->dependName() = ifIs;
	spDepend->dependFlag() = Depend::e_share;
	addDepend(spDepend);
	return true;
}

bool Scope::within(const String &ifIs, const String &child,
		const String &parent, const Offset &offset)
{
	SAFEGUARD;
	sp<Depend> spDepend(new Depend(child));
	spDepend->shareName() = parent;
	spDepend->dependName() = ifIs;
	spDepend->dependFlag() = Depend::e_within;
	spDepend->offset() = offset;
	addDepend(spDepend);
	return true;
}

void Scope::clonePopulate(const String &ifThisHas, const String &thenThisDoes)
{
	t_depends::iterator i_depend;
	for(i_depend = m_depend.begin();i_depend!=m_depend.end();i_depend++)
	{
		if((*i_depend)->dependFlag() & Depend::e_populate)
		{
			if((*i_depend)->dependName() == ifThisHas)
			{
				populate(thenThisDoes, (*i_depend)->attributeName());
			}
		}
	}
}

bool Scope::clear(const String &ifIs)
{
	SAFEGUARD;
	sp<Layout> spLayout;
	spLayout = lookupLayout(ifIs);
	if(spLayout.isValid() && spLayout->locked())
	{
		feX("fe::Scope::clear","cannot clear locked layout");
	}

	t_depends::iterator it_depend;
	for(it_depend = m_depend.begin();it_depend!=m_depend.end();/*it_depend++*/)
	{
		if((*it_depend)->dependFlag() & Depend::e_populate)
		{
			if((*it_depend)->dependName() == ifIs)
			{
				if((*it_depend)->dependFlag() & Depend::e_available)
				{
					sp<Depend> spDepend(new Depend());
					spDepend = *it_depend;
					spDepend->dependFlag() =
						(*it_depend)->dependFlag() & ~Depend::e_populate;
					m_depend.push_back(spDepend);
				}

				t_depends::iterator it_remove;
				it_remove = it_depend;
				it_depend++;
				m_depend.remove(*it_remove);
			}
			else
			{
				it_depend++;
			}
		}
		else
		{
			it_depend++;
		}
	}
	return true;
}

sp<Attribute> Scope::addDepend(sp<Depend> &depend)
{
	SAFEGUARD;
	sp<Attribute> spAttribute;
	bool duplicate = false;
	for(t_depends::iterator i_dep = m_depend.begin();
			i_dep != m_depend.end();
			i_dep++)
	{
		if(*depend == **i_dep)
		{
			duplicate = true;
		}
	}
	if(duplicate)
	{
		if(depend->dependFlag() & Depend::e_available)
		{
			UWORD attributeIndex;
			spAttribute = findAttribute(depend->attributeName(),
					attributeIndex);
		}
		return spAttribute;
	}
	m_depend.push_back(depend);

	if(depend->dependFlag() & Depend::e_attribute)
	{
		depAttributeCheck(depend);
	}
	if(depend->dependFlag() & Depend::e_populate)
	{
		depRecordCheck(depend);
	}
	if(depend->dependFlag() & Depend::e_available)
	{
		spAttribute = dynamicAddAttribute(depend->attributeType(),
				depend->attributeName());
	}
	return spAttribute;
}

void Scope::depAttributeCheck(sp<Depend> &depend)
{
	SAFEGUARD;
	UWORD layoutCount = m_layouts.size();
	for(UWORD index = 0; index < layoutCount; index++)
	{
		if(m_layouts[index]->locked())
		{
			for(unsigned int i = 0; i < depend->matchAttributes().size(); i++)
			{
				if(checkAttribute(depend->matchAttributes()[i],
					m_layouts[index]))
				{
					if(!checkAttribute(depend->attributeName(),
						m_layouts[index]))
					{
						feX("fe::Scope::depAttributeCheck",
							"enforce attribute failure (already locked): "
							"record type: %s MustHave: %s IfHas: %s",
							m_layouts[index]->name().c_str(),
							depend->attributeName().c_str(),
							depend->matchAttributes()[i].c_str());
					}
				}
			}
		}
	}
}

void Scope::depRecordCheck(sp<Depend> &depend)
{
	SAFEGUARD;
	UWORD layoutCount = m_layouts.size();
	for(UWORD index = 0; index < layoutCount; index++)
	{
		if(m_layouts[index]->locked())
		{
			if(m_layouts[index]->name() == depend->dependName())
			{
				if(!checkAttribute(depend->attributeName(),m_layouts[index]))
				{
					feX("fe::Scope::depRecordCheck",
						"enforce attribute failure (already locked): "
						"record type: %s MustHave: %s IfIs: %s",
						m_layouts[index]->name().c_str(),
						depend->attributeName().c_str(),
						depend->dependName().c_str());
				}
			}
		}
	}
}

sp<Attribute>& Scope::findAttribute(const String &name)
{
	SAFEGUARD;
	UWORD dummy;
	return findAttribute(name, dummy);
}

sp<Attribute>& Scope::findAttribute(const String &name, UWORD &index)
{
	SAFEGUARD;
	for(index = 0;index < m_attributes.size(); index++)
	{
		if(name == m_attributes[index]->name())
		{
			return m_attributes[index];
		}
	}

	static sp<Attribute> spAttribute;
	spAttribute=NULL;
	return spAttribute;
}

/// @todo cache m_index in Attribute itself for O(1) lookup
UWORD Scope::attributeIndex(sp<Attribute> spAttribute)
{
	SAFEGUARD;
	for(UWORD index = 0;index < m_attributes.size(); index++)
	{
		if(m_attributes[index] == spAttribute)
		{
			return index;
		}
	}
	feX("fe::Scope::attributeIndex","attribute scope mismatch");
	return 0;
}

sp<Attribute> Scope::attribute(UWORD index)
{
	SAFEGUARD;
	return m_attributes[index];
}

void Scope::peek(Peeker &peeker)
{
	SAFEGUARD;
	peeker.nl();
	peeker.bold();
	peeker.cat("Scope:");
	peeker.normal();
	peeker.in();

	// attributes
	peeker.nl();
	peeker.cat("Attributes:");
	peeker.in();
	for(t_attributes::iterator it = m_attributes.begin();
			it != m_attributes.end(); it++)
	{
		peeker.nl();
		peeker(**it);
		peeker.cat(":");
		std::list<String> names;
		m_spTypeMaster->reverseLookup((*it)->type(),names);
		for(std::list<String>::iterator n_it = names.begin();
			n_it != names.end(); n_it++)
		{
			peeker.cat(" ");
			peeker.cat(*n_it);
		}
	}
	peeker.out();

	// layouts
	peeker.nl();
	peeker.cat("Layouts:");
	peeker.in();
	for(t_layouts::iterator it = m_layouts.begin(); it != m_layouts.end(); it++)
	{
		peeker(**it);
	}
	peeker.out();

#if 0
	// depends
	peeker.nl();
	peeker.cat("Depends:");
	peeker.in();
	for(t_depends::iterator it = m_depend.begin(); it != m_depend.end(); it++)
	{
		peeker.nl();
		peeker(**it);
	}
	peeker.out();
#endif

	peeker.out();
}

Accessor<int> &Scope::refCount(void)
{
	SAFEGUARD;
	FEASSERT(m_paCount);
	return *m_paCount;
}

sp<StoreI> Scope::createStore(sp<Layout> spLayout) const
{
	SAFEGUARD;
	sp<StoreI> spStore;
	if(m_hpRegistry.isValid() && m_storeName != "")
	{
		spStore = m_hpRegistry->create(m_storeName);
		if(!spStore.isValid())
		{
			spStore = new SegmentStore;
		}
	}
	else
	{
		spStore = new SegmentStore;
	}
	spStore->setLayout(spLayout);

	return spStore;
}

void Scope::specifyStore(sp<Registry> spRegistry, const String &storeName)
{
	SAFEGUARD;
	m_hpRegistry = spRegistry;
	m_storeName = storeName;
}

#if FE_SCOPE_SUPPORT_WATCHERS
void Scope::addWatcher(sp<WatcherI> spWatch)
{
	m_watchers.push_back(spWatch);
}

bool Scope::removeWatcher(sp<WatcherI> spWatch)
{
	for(t_watchers::iterator it = m_watchers.begin();
		it != m_watchers.end(); it++)
	{
		if(*it == spWatch)
		{
			m_watchers.erase(it);
			return true;
		}
	}
	return false;
}

void Scope::watch(const Record &record)
{
	for(t_watchers::iterator it = m_watchers.begin();
		it != m_watchers.end(); it++)
	{
		(*it)->add(record);
	}
}

void Scope::unwatch(const Record &record)
{
	for(t_watchers::iterator it = m_watchers.begin();
		it != m_watchers.end(); it++)
	{
		(*it)->remove(record);
	}
}
#endif

OffsetTable::OffsetTable(void)
{
	m_table = NULL;
	m_size = 0;
#if FE_COUNTED_TRACK
	setName("OffsetTable");
#endif
}

OffsetTable::~OffsetTable(void)
{
	if(m_table) { delete [] m_table; }
}

void OffsetTable::resize(unsigned int a_size)
{
	Offset *pOffset = new Offset[a_size];
	for(unsigned int i = 0; i < a_size; i++)
	{
		pOffset[i] = offsetNone;
	}
	if(m_table != NULL)
	{
		for(unsigned int i = 0; i < m_size; i++)
		{
			pOffset[i] = m_table[i];
		}
		delete [] m_table;
	}
	m_table = pOffset;
	m_size = a_size;
	for(UWORD index = 0; index < m_layouts.size(); index++)
	{
		m_layouts[index]->setOffsetTable(m_table);
	}
}

void OffsetTable::attach(sp<Layout> a_layout)
{
	m_layouts.push_back(a_layout);
	a_layout->setOffsetTable(m_table);
}

} /* namespace */

