/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#ifndef __data_WeakRecord_h__
#define __data_WeakRecord_h__

namespace fe
{

/**	@brief Non-persistant reference to an instance of a Layout

	@ingroup data
*/
class FE_DL_EXPORT WeakRecord
{
	friend class Scope;
	public:
					WeakRecord(I32 ignored=0);
//					WeakRecord(void *datablock);
					WeakRecord(const Record &other);
					WeakRecord(const WeakRecord &other);
virtual				~WeakRecord(void);

					operator Record(void);

		WeakRecord	&operator=(const WeakRecord &other);

		bool		operator==(const WeakRecord &other) const;
		bool		operator!=(const WeakRecord &other) const;

					/** @brief Return true if the record points to the
						original valid state block

						A record without the ":SN" field can never
						be confirmed as valid.
						You may be able access it if you are sure it
						still exists by other means. */
		bool		isValid(void) const;
					/** Set the state block. */
		void		set(void *datablock);
					/** Return the state block. */
		void		*data(void) const;
					/** Return the Layout. */
		sp<Layout>	layout(void) const;

					/** Return a raw pointer to the Layout.
						Mainly intended for Accessor for speed. */
		Layout		*rawLayout(void) const;

					/** @brief complain if the record doesn't have
						a serial number */
		void		checkSerialNumber(void) const;

	private:

		void		cacheSerialNumber(void);
		I32			readSerialNumber(void) const;

		hp<Layout>	m_hpLayout;
		void*		m_pStateBlock;
		I32			m_serialNumber;
		U32			m_serialOffset;
};

class WeakRecordInfo : public BaseType::Info
{
	public:
virtual	String	print(void *instance);
virtual	IWORD	output(std::ostream &ostrm, void *instance, t_serialMode mode);
virtual	void	input(std::istream &istrm, void *instance, t_serialMode mode);
virtual	IWORD	iosize(void);
virtual	bool	getConstruct(void);
virtual	void	construct(void *instance);
virtual	void	destruct(void *instance);
};

inline WeakRecord::WeakRecord(I32):
	m_pStateBlock(NULL),
	m_serialNumber(-1)
{
}

inline WeakRecord::WeakRecord(const Record& other)
{
	if(other.isValid())
	{
		m_hpLayout = other.layout();
		m_pStateBlock = other.data();

		cacheSerialNumber();
	}
	else
	{
		m_pStateBlock=NULL;
		m_serialNumber= -1;
	}
}

inline WeakRecord::WeakRecord(const WeakRecord &other)
{
	if(other.isValid())
	{
		m_hpLayout = other.m_hpLayout;
		m_pStateBlock = other.m_pStateBlock;
		m_serialOffset = other.m_serialOffset;
		m_serialNumber = other.m_serialNumber;
	}
	else
	{
		m_pStateBlock=NULL;
		m_serialNumber= -1;
	}
}

/*
inline WeakRecord::WeakRecord(void *datablock)
{
	m_pStateBlock = datablock;
	m_hpLayout = FE_SB_TO_HDR(m_pStateBlock)->m_pStore->getLayout();
}
*/

inline WeakRecord::~WeakRecord(void)
{
}

inline WeakRecord::operator Record(void)
{
	Record record;
	if(m_pStateBlock)
	{
		record.set(m_pStateBlock);
	}
	return record;
}

inline I32 WeakRecord::readSerialNumber(void) const
{
	return (m_serialOffset==offsetNone)?
		-1: *((I32 *)((char *)(m_pStateBlock) + m_serialOffset));
}

inline void WeakRecord::cacheSerialNumber(void)
{
	U32 index = m_hpLayout->scope()->serialIndex();
	m_serialOffset = m_hpLayout->offsetTable()[index];
	m_serialNumber = readSerialNumber();

#if FE_CODEGEN <= FE_DEBUG
//	checkSerialNumber();
#endif
}

inline Layout *WeakRecord::rawLayout(void) const
{
	return m_hpLayout.raw();
}

inline WeakRecord &WeakRecord::operator=(const WeakRecord &other)
{
	if(this != &other)
	{
		m_hpLayout = other.m_hpLayout;
		m_pStateBlock = other.m_pStateBlock;
		m_serialOffset = other.m_serialOffset;
		m_serialNumber = other.m_serialNumber;
	}
	return *this;
}

inline bool WeakRecord::operator==(const WeakRecord &other) const
{
	return (m_pStateBlock == other.m_pStateBlock);
}

inline bool WeakRecord::operator!=(const WeakRecord &other) const
{
	return (m_pStateBlock != other.m_pStateBlock);
}

inline bool WeakRecord::isValid(void) const
{
	if(m_pStateBlock)
	{
		if(m_serialNumber<0 || m_serialNumber==readSerialNumber())
		{
			return TRUE;
		}

		const_cast<WeakRecord*>(this)->m_pStateBlock=NULL;
		const_cast<WeakRecord*>(this)->m_serialNumber= -1;
	}

	return FALSE;
}

inline void WeakRecord::set(void *datablock)
{
	m_pStateBlock = datablock;
	m_hpLayout = FE_SB_TO_HDR(m_pStateBlock)->m_pStore->getLayout();

	Record record;
	record.set(datablock);
	cacheSerialNumber();
}

inline void *WeakRecord::data(void) const
{
	return m_pStateBlock;
}

inline sp<Layout> WeakRecord::layout(void) const
{
	return m_hpLayout;
}

} /* namespace */

#endif /* __data_WeakRecord_h__ */

