/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "platform.pmh"

#define FE_TICK_DEBUG	FALSE

namespace fe
{

FE_DL_PUBLIC unsigned long SystemTicker::ms_ticksPerMS = 1;
FE_DL_PUBLIC float SystemTicker::ms_usPerTick = 1;

void milliSleep(long milliseconds)
{
#if FE_OS==FE_WIN32
	const UINT uPeriod=1;
	MMRESULT result=timeBeginPeriod(uPeriod);
	if(result!=TIMERR_NOERROR)
	{
		feLog("timeBeginPeriod failed\n");
	}
	Sleep(milliseconds);
	result=timeEndPeriod(uPeriod);
	if(result!=TIMERR_NOERROR)
	{
		feLog("timeEndPeriod failed\n");
	}
#elif FE_OS==FE_LINUX
	static struct timeval timeout;

	timeout.tv_sec=milliseconds/1000;
	timeout.tv_usec=(milliseconds-timeout.tv_sec*1000)*1000;

	if(select(0,(fd_set *)NULL,(fd_set *)NULL,(fd_set *)NULL,&timeout) < 0)
	{
		feX("fe::milliSleep","failed");
	}
#elif FE_OS==FE_OSX
	static struct timeval timeout;

	timeout.tv_sec=milliseconds/1000;
	timeout.tv_usec=(milliseconds-timeout.tv_sec*1000)*1000;

	if(select(0,(fd_set *)NULL,(fd_set *)NULL,(fd_set *)NULL,&timeout) < 0)
	{
		feX("fe::milliSleep","failed");
	}
#else
	feX("fe::milliSleep", "not implemented");
#endif
}

SystemTicker::SystemTicker(void)
{
	m_lastexcl = timeMS();
	m_lastincl = timeMS();
	m_lasttick = systemTick();
	m_name = "TICK";
}

SystemTicker::SystemTicker(const char *name)
{
	m_lastexcl = timeMS();
	m_lastincl = timeMS();
	m_lasttick = systemTick();
	m_name = name;
}

unsigned long SystemTicker::timeMS(void)
{
#if FE_OS==FE_LINUX
	static struct timeval timer;
	if( gettimeofday(&timer,NULL) )
	{
		feX("fe::SystemTicker::timeMS", "failed");
	}
	else
	{
		return ((timer.tv_sec & 0x000FFFFF)*1000+(timer.tv_usec/1000));
	}
#elif FE_OS==FE_OSX
	static struct timeval timer;
	if( gettimeofday(&timer,NULL) )
	{
		feX("fe::SystemTicker::timeMS", "failed");
	}
	else
	{
		return ((timer.tv_sec & 0x000FFFFF)*1000+(timer.tv_usec/1000));
	}
#elif FE_OS==FE_WIN32
	return timeGetTime();
#else
	feX("fe::SystemTicker::timeMS", "unimplemented");
#endif
}

void SystemTicker::calibrate(void)
{
#if FE_OS!=FE_OSX
	unsigned long ms, nms, tick;

	// First, force us to end and therefore start a new timestep
	// so that it is less likely the calibration will be interrupted
	// between GetTimeMS and GetSystemTick. (assuming milliSleep
	// is not a spin sleep)
	milliSleep(10);
	ms = timeMS();
	tick = systemTick();
	if(0 == tick)
	{
		feX("fe::SystemTicker::calibrate", "systemTick failed");
	}

	// Now pause for a while
	milliSleep(100);

	nms = timeMS();
	unsigned long tick2=systemTick();

#if FE_TICK_DEBUG
	feLog("SystemTicker::calibrate (%u - %u)/(%u - %u)\n",tick2,tick,nms,ms);
#endif
	tick=tickDifference(tick,tick2);

	ms = nms - ms;
	/// @todo sanity check for rollover

	if(ms)
	{
		ms_ticksPerMS = (unsigned long)(double(tick)/double(ms));
#if FE_TICK_DEBUG
		feLog("  %u = %u / %u\n",ms_ticksPerMS,tick,ms);
#endif

		ms_usPerTick = 1e3*ms/double(tick);
#if FE_TICK_DEBUG
		feLog("  %.6G = 1e3 * %u / %u\n",ms_usPerTick,ms,tick);
#endif
	}
	else
	{
		feX("fe::SystemTicker::calibrate", "difference failed");
	}
#endif
}

void SystemTicker::log(const char *fmt, ...)
{
	String text;
	unsigned long now;
	unsigned long tick;
	tick = systemTick();
	now = timeMS();

	int size=0;
	while(size>=0)
	{
		va_list ap;
		va_start(ap, fmt);

		text.vsPrintf(fmt,ap,size);

		va_end(ap);
	}

	//LOG("[33m%s:[32m %5u [37m%s[0m\n",m_name, now-m_last,(const char *)text);


//	fprintf(stderr,"stderr [33m%s:[32m [1m%4u[0m %4u %10u"
//			" [37m%s[0m\n"

	feLog("[33m%-25s:[32m [1m%4u[0m %4u %12u [37m%s[0m\n"
		,m_name.c_str()
		,(unsigned int)(now-m_lastexcl)
		,(unsigned int)(now-m_lastincl)
		,(unsigned int)(tick-m_lasttick)
		,text.c_str());
	m_lastincl = now;
	m_lastexcl = timeMS();
	m_lasttick = systemTick();
}

void SystemTicker::log(void)
{
	m_lastincl = timeMS();
	m_lastexcl = timeMS();
	m_lasttick = systemTick();
}

}



