/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "platform.pmh"

#if FE_COMPILER==FE_GNU
#include <cxxabi.h>
#if FE_OS!=FE_OSX
#include <execinfo.h>
#endif
#define GNU_SOURCE
#include <dlfcn.h>

void feLogBacktrace(U32 skip)
{
#if FE_OS!=FE_OSX
	fe::String value;
	if(!fe::System::getEnvironmentVariable("FE_E_BACKTRACE", value) ||
			!atoi(value.c_str()))
	{
		return;
	}

	const U32 backtrace_size = 256;
	const U32 buffer_length=256;
	const I32 skip_bottom=2;	// before main()

	void* stack[backtrace_size];
	I32 count=backtrace(stack,backtrace_size);
	char** symbols=backtrace_symbols(stack,count);

	feLog("[35mBACKTRACE[0m\n");
	feLog("Backtrace:\n");

//	backtrace_symbols_fd(stack, count, 2);

	Dl_info info;
	for(I32 m=skip;m<count-skip_bottom;m++)
	{
		if(dladdr(stack[m],&info) != 0)
		{
			char sym_name[buffer_length];
			sym_name[0]=0;

			char* start=strchr(symbols[m],'(');
			if(start)
			{
				I32 symlen=strchr(start,')')-start;
				if(symlen>0)
				{
					strncpy(sym_name,start+1,symlen-1);
					sym_name[symlen-1]=0;
				}
			}

			char buffer[buffer_length];
			UWORD length=buffer_length;
			I32 status = 0;

			char* demangle=
					abi::__cxa_demangle(info.dli_sname,buffer,&length,&status);
/*
			feLog("%s\n%s\n%s\n%s\n\n",
					demangle,
					sym_name,
					info.dli_fname,
					symbols[m]);
*/
			feLog("%d: [37m[1m%s\n",
					m-skip,demangle? demangle: sym_name);
			feLog("[0m[2m");

#if FALSE
			// doesn't work on DL's and not very accurate either
			if(system(NULL))
			{
				char command[buffer_length];
				sprintf(command, "addr2line -e %s %p",
						info.dli_fname,stack[m]);
				system(command);
			}
#endif
			feLog("%s\n",info.dli_fname);
			feLog("[0m");
		}
	}

	feLog("[35mEND BACKTRACE[0m\n");

	if(symbols)
	{
		free(symbols);
	}

	feLog("[35mSTARTING GDB[0m\n");

	if(system(NULL) && dladdr(stack[count-1],&info) != 0)
	{
//		char* exename=info.dli_fname;
		char exename[buffer_length];
		exename[readlink("/proc/self/exe", exename, sizeof(exename)-1)]='\0';

		pid_t pid=getpid();
		char command[buffer_length];
		if(fe::System::getEnvironmentVariable("FE_E_ATTACH", value) &&
				atoi(value.c_str()))
		{
			sprintf(command,"gdb -quiet %s %d",exename,pid);
		}
		else
		{
			sprintf(command,
					"echo 'bt\ndetach\nquit\n'| gdb -batch -x /dev/stdin %s %d",
					exename,pid);
		}
//		feLog("Running: %s\n",command);
		system(command);
	}

	feLog("[35mEND GDB[0m\n");
#endif
}

#include <signal.h>

void feSegvHandler(int signal)
{
	feLog("[31mSEGMENTATION FAULT[0m\n");
	feLogBacktrace();
	exit(139);
}

void feRegisterSegvHandler(void)
{
	struct sigaction action;
	memset(&action,0,sizeof(action));
	action.sa_handler=feSegvHandler;
	sigemptyset(&action.sa_mask);
	action.sa_flags=0;

	int fail=sigaction(SIGSEGV,&action,NULL);
	if(fail)
	{
		feLog("feRegisterSegvHandler failed\n");
	}
}
#else
void feLogBacktrace(U32)
{
	feLog("[35mBACKTRACE[0m: not implemented\n");
}

void feRegisterSegvHandler(void)
{
	feLog("feRegisterSegvHandler not implemented\n");
}
#endif

namespace fe
{

void vsPrintf(std::string &target, const char* fmt, va_list ap,int& size)
{
	if(size<0)
	{
		feX("(debug)vsPrintf size<0 indicates unheeded discontinuation");
	}
	if(!size)
	{
		// start at 80 bytes, seems reasonable (1 tty line)
		size=80;
	}
	char *tmpbuf = new char[size];
	FEASSERT(NULL != tmpbuf);

	int n=FEVSNPRINTF((char *)tmpbuf, size, fmt, ap);

	if(n> -1 && n<size)
	{
		size= -1;
		target = tmpbuf;
		delete [] tmpbuf;
		return;
	}
	if(n> -1)
	{
		size=n+1;
	}
	else
	{
		size*=2;
	}

	delete [] tmpbuf;
	target="";
}
#if FALSE
void vsPrintf(std::string &target, const char *fmt, va_list ap)
{

	int sz = 80;
	char *tmp = new char[sz];
	FEASSERT(NULL != tmp);
	while(true)
	{
		int n=FEVSNPRINTF((char *)tmp, sz, fmt, ap);

		if(n> -1 && n<sz) break;
		if(n> -1) sz=n+1;
		else sz*=2;

		delete [] tmp;
		tmp = new char[sz];
		FEASSERT(NULL != tmp);

		// NOTE calling vsnprintf again without va_end,va_start is undefined
		feX("(debug)vsPrintf warning, expanding buffer\n");
	}
	target = tmp;
	delete [] tmp;
}
#endif

ImmediateOutputLogger debugLogger;

void ImmediateOutputLogger::operator()(const char *format, ...) const
{
	std::string output;

	int size=0;
	while(size>=0)
	{
		va_list ap;
		va_start(ap, format);

		vsPrintf(output,format,ap,size);

		va_end(ap);
	}

#if FE_OS==FE_WIN32
	OutputDebugString(output.c_str());
#endif

	std::cout << output.c_str();
	std::cout.flush();
}

void ImmediateOutputLogger::operator()(const String &string) const
{
	operator()("%s",string.c_str());
}

}

fe::Logger& feLog = fe::debugLogger;

