/*	Copyright (C) 2003-2008 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "plugin/plugin.pmh"

#define	FE_COMPONENT_UNLOAD	TRUE	//* turn off to retain symbols (valgrind)

namespace fe
{

Component::Component(void):
	m_singleton(FALSE),
	m_stray(FALSE)
{
}

Component::~Component(void)
{
	//* check for unnamed components (not usually turned on)
//	FEASSERT(!name().empty());

//	feLog("~Component() \"%s\"\n",m_name.c_str());

	disjoin();
	setStray(FALSE);
}

Component::Hub::Hub(void):
	m_name("Component::Hub"),
	m_singletonCount(0)
{
}

Component::Hub::~Hub(void)
{
	abandonHandle();
//	feLog("~Hub %p count=%d \"%s\"\n",this,count(),name().c_str());
}

void Component::Hub::attach(Component* pComponent,BWORD singleton)
{
	m_handleList.append(sp<Component>(pComponent));

	pComponent->trackReference(static_cast<Counted*>(this));

	if(singleton)
	{
		m_singletonCount++;
	}
}

void Component::Hub::detach(Component* pComponent,BWORD singleton)
{
	pComponent->untrackReference(static_cast<Counted*>(this));

#if FE_CODEGEN <= FE_DEBUG
	BWORD removed=
#endif
	m_handleList.remove(sp<Component>(pComponent));
	FEASSERT(removed);

	if(singleton)
	{
		m_singletonCount--;
	}
}

void Component::setStray(BWORD stray)
{
//	feLog("Component::setStray(%d) %d %s\n",stray,m_stray,name().c_str());

	if(m_spLibrary.isValid())
	{
		if(stray && !m_stray)
		{
			m_spLibrary->adjustStrayCount(1);
		}
		else if(!stray && m_stray)
		{
			m_spLibrary->adjustStrayCount(-1);
		}
	}

	m_stray=stray;
}

void Component::adjoin(sp<Component> spComponent)
{
#if FE_COUNTED_MT
	//* not sure it's necessary here, but it seems like a good idea
	boost::recursive_mutex::scoped_lock guard(Counted::ms_counted_mutex);
#endif

	if(!spComponent.isValid())
	{
		feLog("Component::adjoin invalid component");
		return;
	}

	if(m_hpHub.isValid())
	{
		String errmsg;
		errmsg.sPrintf("Component::adjoin \"%s\" already adjoined\n",
				name().c_str());
		feX(errmsg.c_str());
	}

	Component* pOther=spComponent.raw();
	if(!pOther->m_hpHub.isValid())
	{
		pOther->assignHub(new Hub());
		pOther->m_hpHub->setName("Component::Hub "+pOther->name());
	}

	assignHub((Hub*)pOther->m_hpHub.raw());
}

void Component::disjoin(void)
{
	if(m_hpHub.isValid())
	{
		hp<Hub> hpHub=m_hpHub;
		m_hpHub=(Hub*)NULL;

		I32 references=count();

		setStray(FALSE);

		//* NOTE deletion may occur here
		hpHub->detach(this,isSingleton());

		if(hpHub.isValid())
		{
			if(references>1)
			{
//				feLog("Component::disjoin release Hub from \"%s\"\n",
//						name().c_str());
				hpHub->release();
			}
		}
	}
}

void Component::assignHub(Hub* pHub)
{
	if(count()>1)
	{
		pHub->acquire();
	}
	if(count()<2+isSingleton())
	{
		setStray(TRUE);
	}

	m_hpHub=pHub;
	pHub->attach(this,isSingleton());
}

void Component::setLibrary(Library* pLibrary)
{
	//* assure library lifetime while this component exists
	FEASSERT(!m_spLibrary.isValid());
	if(!m_spLibrary.isValid())
		m_spLibrary=pLibrary;
}

hp<Registry> Component::registry(void) const
{
	return m_spLibrary.isValid()? m_spLibrary->registry(): hp<Registry>();
}

sp<Component> Component::create(const String &implementation) const
{
	sp<Component> spComponent = registry()->create(implementation);
	return spComponent;
}

void Component::release(void)
{
	int refCount;

	{
//		feLog("+%p Component::release %d %s\n",(U32)this,count(),
//				name().c_str());

#if FE_COUNTED_MT && FE_COMPONENT_TRACK
		boost::recursive_mutex::scoped_lock guard(Counted::ms_counted_mutex);
#endif

#if FE_COMPONENT_TRACK
		if(count() && registry().isValid())
			registry()->tracker().release(this,name(),count()-1);
#endif

		refCount=releaseInternal();
		if(m_hpHub.isValid())
		{
			if(refCount==1+isSingleton())
			{
				setStray(TRUE);
			}

			if(refCount==1)
			{
//				feLog("Component::release release Hub from %s\n",
//						name().c_str());

				m_hpHub->release();

				//* this component might now be deleted
				return;
			}
		}
	}

	if(!refCount)
	{
		abandonHandle();
		fe::internal::checkUnload(m_spLibrary,this);
	}
}

namespace internal
{
// static
void checkUnload(sp<Library>& spLibrary,Component* pComponent)
{
	sp<Library> spChainLibrary;
	DL_Loader* pLoader=NULL;
	if(spLibrary.isValid() && spLibrary->adjustedReferences()==1)
	{
		pLoader=spLibrary->getLoader();
		spChainLibrary=spLibrary->getChainLibrary();
	}

	if(pComponent)
		delete pComponent;
	else
		spLibrary=NULL;

	while(pLoader)
	{
		feLog("Component::checkUnload closeDL \"%s\"\n",
				pLoader->name().c_str());

#if	FE_COMPONENT_UNLOAD
		pLoader->closeDL();
#endif

		delete pLoader;
		pLoader=NULL;

		if(spChainLibrary.isValid() && spChainLibrary->adjustedReferences()==1)
		{
			pLoader=spChainLibrary->getLoader();
			spChainLibrary=spChainLibrary->getChainLibrary();
		}
	}
}

} // namespace internal

} // namespace fe
